// 
//  Copyright (C) 2002,2003,2007,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "parameter.h"

#include "PegApp.hh"
#include "PegParData.hh"
#include "PegCmdButtons.hh"
#include "PegParEntry.hh"
#include "PegParEntryString.hh"
#include "PegParEntryInteger.hh"
#include "PegParEntryReal.hh"
#include "PegParEntryBoolean.hh"

#include "Parameter.hh"


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
PegParData::PegParData(string fname, int argc, const char **argv, GtkWidget* parent)
{
   if (fname.length())
   {
      _fname = fname; 
      if (argc > 1)
         _param = new Parameter(_fname, argc-1, (char**)&argv[1]);
      else 
         _param = new Parameter(_fname, argc, (char**)argv);
      _fpath = _param->GetParamPath(); 
      _parCnt = _param->GetLength(); 
   }
   else
   {
      _param = NULL; 
      _parCnt = 0; 
   }
   _parent = parent; 
   _lastChanges = "\0"; 
   _showHidden = true;
   if (theApp)   theApp->get_preference("display.hidden", _showHidden); 
}


// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
PegParData::~PegParData()
{
   if (_layout)
   {
      gtk_widget_destroy(_layout); 
      _layout = NULL; 
   }
   if (_param)
   {
      delete _param;
      _param = NULL; 
   }
   list<string>::iterator it;
   int sz = _list.size(); 
   for(int rr = sz; rr--; )
      _list.erase(_list.begin()); 
}


// --------------------------------------------------------
// load data and set up parameter editing pane
// --------------------------------------------------------
void PegParData::Setup()
{
   GtkWidget* prev_w = NULL;
   string loaded = "\0"; 
   int row = 1; 

   if (_param)
      loaded = _param->GetParamPath();  

   if (loaded.length())
   {
      PegParEntry::_setColors = 1;

      ParamItem item; 
      int   hiddenCnt = 0; 

      // determine number of hidden params
      for (int rr = 0; rr < _parCnt; rr++)
      { 
	 _param->GetPosition(rr, item);
         if (item.IsHiddenMode())   hiddenCnt++;
      }

      // create appropriately sized table 
      _layout = gtk_table_new(_parCnt, 12, false);

      // load table
      for (int rr = 0; rr < _parCnt; rr++)
      { 
         _param->GetPosition(rr, item);
         switch(item.GetType())
         {
            case ParamItem::REAL_TYPE:
            {
               string nme = item.GetName();
               ParamReal ritem;
               _param->GetParameterByName(nme, ritem);
               PegParEntry* ps = new PegParEntryReal(&ritem);
               if (_showHidden || !item.IsHiddenMode())
               {
		  GtkWidget* ww = ps->getLabelBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
           	     1,4, row, row+1);
		  ww = ps->getEntryBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
		     5,10, row, row+1);
                  ww = ps->getIndBox(); 
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww,
                     5,10, row, row+1);
                  row++; 
		  ((PegParEntryReal*)ps)->manage();
               }
	       else hiddenCnt++; 
               _list.push_back(ps);
            }
            break; 
            case ParamItem::INTEGER_TYPE:
            {
               string nme = item.GetName();
               ParamInteger iitem;
               _param->GetParameterByName(nme, iitem);
               PegParEntry* ps = new PegParEntryInteger(&iitem);
               if (_showHidden || !item.IsHiddenMode())
               {
		  GtkWidget* ww = ps->getLabelBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
           	     1,4, row, row+1);
		  ww = ps->getEntryBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
		     5,10, row, row+1);
                  ww = ps->getIndBox(); 
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww,
                     5,10, row, row+1);
                  row++; 
		  ((PegParEntryInteger*)ps)->manage();
               }
	       else hiddenCnt++; 
               _list.push_back(ps);
            }
            break; 
            case ParamItem::FILENAME_TYPE:  // fallthrough intended
            case ParamItem::STRING_TYPE:
            {
               string nme = item.GetName(); 
               ParamString sitem;
               _param->GetParameterByName(nme, sitem);
               PegParEntry* ps = new PegParEntryString(&sitem);
               if (_showHidden || !item.IsHiddenMode())
               {
		  GtkWidget* ww = ps->getLabelBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
           	     1,4, row, row+1);
		  ww = ps->getEntryBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
		     5,10, row, row+1);
                  ww = ps->getIndBox(); 
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww,
                     5,10, row, row+1);
                  row++; 
		  ((PegParEntryString*)ps)->manage();
               }
	       else hiddenCnt++; 
               _list.push_back(ps);
            }
            break; 
            case ParamItem::BOOLEAN_TYPE:
            {
               string nme = item.GetName();
               ParamBoolean bitem;
               _param->GetParameterByName(nme.c_str(), bitem);
               PegParEntry* ps = new PegParEntryBoolean(&bitem);
               if (_showHidden || !item.IsHiddenMode())
               {
		  GtkWidget* ww = ps->getLabelBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
           	     1,4, row, row+1);
		  ww = ps->getEntryBox();
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww, 
		     5,10, row, row+1);
                  ww = ps->getIndBox(); 
		  gtk_table_attach_defaults(GTK_TABLE(_layout), ww,
                     5,10, row, row+1);
                  row++; 
		  ((PegParEntryBoolean*)ps)->manage();
               }
	       else hiddenCnt++; 
               _list.push_back(ps);
            }
            break; 
            default:
            break; 
         }
      }

      // widgets are spaced at bottom of entries so adjust the last widget's
      // bottom offset so that it is not squished to bottom of form 
      if (prev_w)
      {
         int offset; 

         // get parameter specified offset if possible, else use default
//WLM         if (theParameters)
//WLM            offset = theParameters->GetSpacing();
//WLM          else
            offset = 10;

//WLM         XtVaSetValues(prev_w, 
//WLM            XmNbottomAttachment, XmATTACH_FORM, 
//WLM            XmNbottomOffset, offset, 
//WLM            NULL);
      }

      if (hiddenCnt && (hiddenCnt == _parCnt)) 
      {
         string msg = "All parameters for ";
         msg.append(_fname);
         msg.append(" are Hidden and the display hidden parameters option is diabled.");
         theApp->status_message(msg);
      }

      // update the recently accessed file list 
      PegCmdButtons* btns = theApp->getCmdButtons();
      string fnm = _param->GetParamPath();
      theApp->update_filelist(fnm); 
      if (btns)
      {
         if (!fnm.length())   fnm = _fname; 
         btns->setFileLabel(fnm);
      } 
   } 
   else 
   { 
      _layout = gtk_table_new(2, 12, false);
      string msg = _fname;
      if (msg.length())
         msg.append(" or its corresponding parameter file could not be opened or does not exist.");
      else
         msg = "No file loaded. Please specify a file via the File pulldown menu.";
      theApp->status_message(msg);
   }

   gtk_table_set_row_spacings(GTK_TABLE(_layout), 10);
   gtk_table_set_col_spacings(GTK_TABLE(_layout), 10);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(_parent), _layout);
   gtk_widget_show(_layout); 
}



// --------------------------------------------------------
// validate data (return 0 if no errors)
// --------------------------------------------------------
// return of bad parameter values (0 if no errors)  
int PegParData::Validate()
{
   if (!_param) return -1; 
   int badCnt = 0; 
   int error; 
   PegParEntry* entry = NULL; 

   list<PegParEntry*>::iterator it;
   for (it = _list.begin(); it != _list.end(); it++)
   {
      if ((entry = (*it)) == NULL) return -1;  
      error = 0; 
      switch(entry->getEntryType())
      {
         case PegParEntry::STRING:      // fallthrough intended 
         case PegParEntry::FILENAME:    // fallthrough intended 
         case PegParEntry::PSET:
            error = _param->SetValueAtPosition(entry->getPosition(), 
               ((PegParEntryString*)entry)->get());
         break; 
         case PegParEntry::INTEGER:
            error = _param->SetValueAtPosition(entry->getPosition(), 
               ((PegParEntryInteger*)entry)->get());
         break; 
         case PegParEntry::REAL:
            error = _param->SetValueAtPosition(entry->getPosition(), 
               ((PegParEntryReal*)entry)->get());
         break; 
         case PegParEntry::BOOLEAN:
            error = _param->SetValueAtPosition(entry->getPosition(), 
               ((PegParEntryBoolean*)entry)->get());
         break; 
         default:
            error = _param->SetValueAtPosition(entry->getPosition(),
               entry->get());
         break; 
      }
      if (error == 1)
         entry->clear_error_color(); 
      else 
      {
         entry->set_error_color(); 
         badCnt++;
      }
   }  // end for 
   return (badCnt); 
}


// --------------------------------------------------------
// write out parameters (return 0 if no error)
// --------------------------------------------------------
int PegParData::Write(char *fname, int overwrite)
{
   int err = 0; 

   if (_param)
   {
      string msg; 
      if (fname)
      {
         // write output file to specified file (set overwrite on) 
	 if (_param->SaveAs(fname, overwrite, 1) == 1)
         {
            msg = fname;
            msg.append(" has been updated.");
            theApp->status_message(msg);
         }
         else
         {
            msg = "An error occured while attempting to update ";
            msg.append(fname);
            theApp->status_message(msg);
	    err = -2; 
         }
      }
      else 
      {
         if (_param->GetReadOnly())
         {
            err = -1; 
            msg = "Unable to write changes to ";
            msg.append(_param->GetParamPath()); 
            msg.append(" (read-only file).");
            theApp->status_message(msg);
         } 
         else 
         {
	    if (_param->Write(_fname, 1) == 1)
            {
               msg = _param->GetParamPath();
               msg.append(" has been updated.");
               theApp->status_message(msg);
            }
            else
            {
               msg = "An error occured while attempting to update ";
               msg.append(_param->GetParamPath()); 
               theApp->status_message(msg);
	       err = -2; 
            }
         } 
      } 
      if (err == 0) _lastChanges = _param->GetChanges();
   }
   else 
   {
      err = -1; 
   } 
   return err; 
}


// --------------------------------------------------------
// determine if unsaved changes exist (0 if no changes)
// --------------------------------------------------------
int PegParData::ChangesMade()
{
   int changes = 0;

   if (_param)
   {
      Validate(); 
      string val = _param->GetChanges(); 
      changes = _cxcparam_strcmp_loose(val, _lastChanges);
   }
   return changes; 
}


// --------------------------------------------------------
// Reset entries in editing pane- label props  & eventually hidden param vis
// --------------------------------------------------------
void PegParData::ResetEntries()
{
   PegParEntry* entry = NULL; 
   PegParEntry::_setColors = 1;  // PegParEntry uses a static initializer flag
   for (list<PegParEntry*>::iterator it = _list.begin(); it != _list.end(); it++)
      if ((entry = (*it)) != NULL)   entry->reset_label();
   if (theApp)
   {
      bool showhid;
      theApp->get_preference("display.hidden", showhid);
      if (showhid != _showHidden) 
         theApp->status_message("The display.hidden parameter value has changed, reopen the file to see changes."); 
   }
} 


string PegParData::GetFilePath()
{
   string fp = _fpath;
   if (_param)  fp = _param->GetParamPath(); 
   return fp; 
}


int PegParData::SetParamValue(string pname, string value)
{
   int retval = 1; 
   PegParEntry *entry = NULL;
   list<PegParEntry*>::iterator it;
   for (it = _list.begin(); it != _list.end(); it++)
   {
      if ((entry = (*it)) == NULL) return -1;  
      if (entry->getName() == pname)
      {
	 retval = 0; 
         switch(entry->getEntryType())
	 {
	    case PegParEntry::STRING:      // fallthrough intended
	    case PegParEntry::FILENAME:    // fallthrough intended
	    case PegParEntry::PSET:
	       ((PegParEntryString*)entry)->set(value.c_str());
            break;
            case PegParEntry::INTEGER:
               ((PegParEntryInteger*)entry)->set(value.c_str());
            break;
            case PegParEntry::REAL:
               ((PegParEntryReal*)entry)->set(value.c_str());
            break;
            case PegParEntry::BOOLEAN:
               ((PegParEntryBoolean*)entry)->set(value.c_str());
            break;
            default:
	       retval = 1; 
            break;
         }
         break;
      }
   } 
   return retval; 
}

string PegParData::GetParamValue(string pname)
{
   string retval = ""; 
   PegParEntry *entry = NULL;
   list<PegParEntry*>::iterator it;
   for (it = _list.begin(); it != _list.end(); it++)
   {
      if ((entry = (*it)) == NULL) return retval;  
      if (entry->getName() == pname)
      {
         char *val; 
         switch(entry->getEntryType())
	 {
	    case PegParEntry::STRING:      // fallthrough intended
	    case PegParEntry::FILENAME:    // fallthrough intended
	    case PegParEntry::PSET:
	       val = ((PegParEntryString*)entry)->get();
            break;
            case PegParEntry::INTEGER:
               val = ((PegParEntryInteger*)entry)->get();
            break;
            case PegParEntry::REAL:
               val = ((PegParEntryReal*)entry)->get();
            break;
            case PegParEntry::BOOLEAN:
               val = ((PegParEntryBoolean*)entry)->get();
            break;
            default:
            break;
         }
	 if (val)  retval = val; 
         break;
      }
   } 
   return retval; 
}




