// 
//  Copyright (C) 2002,2007,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <limits.h>

#include "PegApp.hh"
#include "PegParData.hh"
#include "Parameter.hh"
#include "PegParEntryInteger.hh"
#include "PegSpinAdjustDialog.hh"
#include "PegUtils.hh"


// --------------------------------------------------------
// Constructor
// --------------------------------------------------------
PegParEntryInteger::PegParEntryInteger(ParamInteger* item):
   PegParEntry((ParamItem*)item)
{
   int pmin, pmax, pval, step;
   _entry_type = PegParEntry::INTEGER; 
   _item = new ParamInteger(*item);

   _scale = 1; 

   _inedit = 0; 
   pval = _item->GetValue(); 
   pmin = _item->GetMinValue(); 
   pmax = _item->GetMaxValue(); 

   if (pmin == INDEF_INT)  pmin = INT_MIN;
   if (pmax == INDEF_INT)  pmax = INT_MAX;
   if (pmin == INT_MIN || pmax == INT_MAX)
   {
      if (pval != INDEF_INT)   step = pval / 10;
         else                  step = 1;
   }
   else step = (pmax - pmin) / 10;
   if (step <1)   step = 1;

   _adj = (GtkAdjustment *)
        gtk_adjustment_new(pval, pmin, pmax, 1.0, step, 0.0);


   _spin_w   = gtk_spin_button_new(_adj, 1.0, 0);
   gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(_spin_w), 0);
   gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(_spin_w),
      GTK_UPDATE_IF_VALID);

  if (_tip_grp)
  {
     string prompt = item->GetPrompt();
     gtk_tooltips_set_tip(_tip_grp, _spin_w, prompt.c_str(), NULL);
     gtk_tooltips_set_tip(_tip_grp, _ind_txt, prompt.c_str(), NULL);
  }

   gtk_widget_add_events(_spin_w, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(_spin_w,"button-press-event", G_CALLBACK(ButtonPressCB),
      (void*)this);
   g_signal_connect(_spin_w,"output", G_CALLBACK(change_cb),
      (void*)this);
   // The 'changed' signal operate on a per key change- bad for _ind_txt since
   // it evaluates the indirection value as it goes, so use 'activate' and
   // 'focus-out' signals instead.
   g_signal_connect(_ind_txt,"activate", G_CALLBACK(IndirEditCB), (void*)this);
   g_signal_connect(_ind_txt, "focus-out-event", G_CALLBACK(FocusOutCB),
                    (void*)this);

   int subcnt = _item->GetSubsetCount();
   if (subcnt)
   {
      _enumer = gtk_combo_box_new_text();

      // add 2 to subcnt- 1 for INDEF + 1 for NULL
      string intval; 
      int* tmplbls = _item->GetSubset();
      int curval = _item->GetCurrent();
      int rr, pos = -1;
      for (rr = 0; rr < subcnt; rr++) 
      {
         intval = _int_to_string(tmplbls[rr]);
         if (tmplbls[rr] == curval)   pos = rr; 
         gtk_combo_box_append_text(GTK_COMBO_BOX(_enumer), intval.c_str());
      }
      gtk_combo_box_append_text(GTK_COMBO_BOX(_enumer), "INDEF");
      if (INDEF_INT == curval)   pos = subcnt;
      if (pos == -1)
      {
         char buff[256];
         sprintf(buff, "The value %1d for enumerated integer %s is invalid- defaulting to INDEF",
            curval, _name.c_str());
         theApp->status_message(buff);
         pos = subcnt;
      }
      gtk_combo_box_set_active(GTK_COMBO_BOX(_enumer), pos);
      _disp_type = PegParEntry::ENUM; 
   }
   else 
   {
      // set up range checking info 
      _min = INT_MIN;
      _max = INT_MAX; 
      switch(_item->GetRangeCheck())
      {
         case MIN_MAX_CHECK:
            _min = _item->GetMinValue();
            _max = _item->GetMaxValue();
         break; 
         case MIN_ONLY_CHECK:
            _min = _item->GetMinValue();
         break; 
         case MAX_ONLY_CHECK:
            _max = _item->GetMaxValue();
         break; 
         case DEFAULT_CHECK:    // fallthrough intended 
         case SUBSET_CHECK:     // fallthrough intended 
         case NO_RANGE_CHECK:   // fallthrough intended 
         default:
         break; 
      }
     _disp_type = PegParEntry::DEFAULT;
   }

   string tmp = "\0";
   char *rcur = _item->GetRedirectCurrent();

   initEntry();

   if (rcur[0] != '\0')  
   {
      _default = _disp_type;
      _disp_type = PegParEntry::INDIR;
      tmp.append(rcur);
   }
   else
   {
      int val = _item->GetCurrent();
      if (val == INDEF_INT)
         tmp = "INDEF";
      else
	 tmp = _int_to_string(val); 
   }

   if (_disp_type == PegParEntry::ENUM && _enumer)
      gtk_box_pack_start(GTK_BOX(_entry), _enumer, false, false, 0);
   else if (_spin_w)
   {
      gtk_box_pack_start(GTK_BOX(_entry), _spin_w, false, false, 0);
   }

   set(tmp.c_str());
}


// --------------------------------------------------------
// Destructor
// --------------------------------------------------------
PegParEntryInteger::~PegParEntryInteger()
{
   if (_item)
   {
      delete _item;
      _item = NULL; 
   }
}


// --------------------------------------------------------
// Set value of entry
// --------------------------------------------------------
void PegParEntryInteger::set(const char* val)
{
   if (_inedit)  return; // if in an edit- discard this request
   _inedit = 1; 
   if (val)
   {
      // make sure that indirections are displayed as indirections
      // and entries as entries 
      if (((val[0] == ')') && (_disp_type != PegParEntry::INDIR)) ||
          ((val[0] != ')') && (_disp_type == PegParEntry::INDIR)))
         toggleDispType();

      if (_disp_type == PegParEntry::DEFAULT)
      {
	 if (_cxcparam_strcmp_loose(val,"INDEF") == 0)
	    gtk_entry_set_text(GTK_ENTRY(_spin_w), (char*)val); 
	 else
	 {
            Parameter *par = thePData->GetParameter();
            if (par->SetValueAtPosition(_item->GetPosition(), val) != 1)
	    {
	       string msg = "Invalid value entered: ";
	       msg.append(val);
	       switch(_item->GetRangeCheck())
	       {
                 case SUBSET_CHECK:
                    msg.append(" - value must be: ");
                    msg.append(_int_to_string(_item->GetMinValue()));
	         break;
	         case MIN_MAX_CHECK:
	            msg.append(" - value must be between : ");
	            msg.append(_int_to_string(_item->GetMinValue()));
	            msg.append(" and ");
	            msg.append(_int_to_string(_item->GetMaxValue()));
	         break;
	         case MIN_ONLY_CHECK:
	            msg.append(" - value must be greater than : ");
	            msg.append(_int_to_string(_item->GetMinValue()));
	         break;
	         case MAX_ONLY_CHECK:
	            msg.append(" - value must be less than : ");
	            msg.append(_int_to_string(_item->GetMaxValue()));
	         break;
	         default:
	            // no range to output
	         break;
	       }
	       gtk_entry_set_text(GTK_ENTRY(_spin_w), (char*)val); 
	       set_error_color(); 
               theApp->status_message(msg);
	    }
            else 
            {
	       long v2;
	       _cxcparam_string_to_long(val, v2);
               gtk_entry_set_text(GTK_ENTRY(_spin_w), (char*)val); 
	       gtk_spin_button_set_value(GTK_SPIN_BUTTON(_spin_w), v2);
	       clear_error_color(); 
	    }
         }
      }
      else if (_disp_type == PegParEntry::INDIR)
      {
         string cur = gtk_entry_get_text(GTK_ENTRY(_ind_txt));
	 if (cur != val)  gtk_entry_set_text(GTK_ENTRY(_ind_txt), (char*)val);
         if (thePData && _item)
         {
            Parameter *par = NULL;

            if ((par = thePData->GetParameter()) != NULL)
            {
               string eval = par->EvalIndir((char*)_name.c_str(), (char*)val);
	       gtk_entry_set_text(GTK_ENTRY(_ind_val), eval.c_str());
            }
         }
      }
      else 
      {
         int subcnt = _item->GetSubsetCount();
         int pos = -1;
         string sval = "INDEF";

         int* tmplbls = _item->GetSubset();
         string getv = gtk_combo_box_get_active_text(GTK_COMBO_BOX(_enumer));
         
         if (_cxcparam_strcmp_loose(sval, val) == 0)
            pos = subcnt;   // set value to INDEF
         else 
         {
            for (int rr = subcnt; rr--; )
            {
	       sval = _int_to_string(tmplbls[rr]);
               if (_cxcparam_strcmp_loose(sval,val) == 0)
               {
                  pos = rr;
                  break;
               }
            }
         }
         if ((pos != -1) && (_cxcparam_strcmp_loose(getv,val) != 0))
	    gtk_combo_box_set_active(GTK_COMBO_BOX(_enumer), pos);
      } 

      if (thePData)
      {
         Parameter *par = thePData->GetParameter();
         if (par->SetValueAtPosition(_item->GetPosition(), val) != 1)
	    set_error_color(); 
	 else 
	    clear_error_color(); 
      }
   } 
   _inedit = 0; 
   _lastval = gtk_entry_get_text(GTK_ENTRY(_spin_w)); 
}


// --------------------------------------------------------
// Get the value of the entry
// --------------------------------------------------------
char* PegParEntryInteger::get()
{
   static string val;

   switch(_disp_type)
   {
      case PegParEntry::DEFAULT:
	 val = gtk_entry_get_text(GTK_ENTRY(_spin_w));
      break;
      case PegParEntry::INDIR:
	 val = gtk_entry_get_text(GTK_ENTRY(_ind_txt));
      break;
      case PegParEntry::ENUM:
         val = gtk_combo_box_get_active_text(GTK_COMBO_BOX(_enumer));
      break;
      default:
	 val = _int_to_string(_item->GetValue());
      break;
   }
   return (char*)val.c_str();
}


// handle middle button press on spin button to adjust scale 
int PegParEntryInteger::ButtonPressCB(GObject *object, GdkEvent *event,
   gpointer user_data)
{
   GdkEventButton *evt = (GdkEventButton*)event;
   if (evt->button == 2)
   {
      PegParEntryInteger *obj = (PegParEntryInteger*)user_data;
      obj->set_adjustments(); 
      return 1;    // don't fallthough to default spin button handler 
   }
   else
      return 0;    // fallthrough to default spin button handler
}


// callback to handle leaving focus event from text|indirection field
int PegParEntryInteger::FocusOutCB(GObject *object, GdkEvent *event,
   gpointer user_data)
{
   PegParEntryInteger *obj = (PegParEntryInteger*)user_data;
   if (obj && obj->in_edit())   return 0; 
   GtkWidget* ww = GTK_WIDGET(object);
   if (obj)
      if (ww == obj->GetIndirField())         return obj->update_indir();
         else if (ww == obj->GetTextField())  return obj->update_text();
   return 0;
}

// callback to handle activate signal ('return') from text field
int PegParEntryInteger::TextEditCB(GObject *object, gpointer user_data)
{
   PegParEntryInteger *obj = (PegParEntryInteger*)user_data;
   if (obj) return obj->update_text();
      else return 0;
}

// callback to handle activate signal ('return') from indirection field
int PegParEntryInteger::IndirEditCB(GObject *object, gpointer user_data)
{
   PegParEntryInteger *obj = (PegParEntryInteger*)user_data;
   if (obj->in_edit())   return 0;
   if (obj) return obj->update_indir();
      else return 0;
}


int PegParEntryInteger::change_cb(GObject *object, gpointer user_data)
{
   PegParEntryInteger *my = (PegParEntryInteger*)user_data;

   static string prval;
   GtkWidget *obj = GTK_WIDGET(object);
   string  sval = gtk_entry_get_text(GTK_ENTRY(obj));

   prval = my->get_lastval(); 

   if (my->in_edit()) return 1; 

   if (sval != prval)
   {
      prval = sval; 
      if (my)  my->set(sval.c_str()); 
      return 1; 
   }
   else if (sval.find(")") == 0)
   {
      if (my)  my->set(sval.c_str()); 
      return 1; 
   }
   else if (_cxcparam_strcmp_loose(prval,"INDEF") == 0)
   {
      // disable spin button changes when value is INDEF
      return 1; 
   }
   else if (my->check_valid_input(sval) != 0)
   {
      // This ensures that we don't reset an invalid value to the last 
      // valid value on a leave focus- we want to force the user to fix
      // invalid values rather than silently switch back
      return 1; 
   }

   return 0;
}


int PegParEntryInteger::update_text()
{
   Parameter *par = NULL;

   if (_inedit) return 0;  

   if ((par = thePData->GetParameter()) != NULL)
   {
      int vv = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(_spin_w));
      string val = _int_to_string(vv); 

      if (par->SetValueAtPosition(_item->GetPosition(), val.c_str()) != 1)
      {
         string msg = "Invalid value entered: ";
         msg.append(val);
	 set_error_color(); 
         switch(_item->GetRangeCheck())
         {
            case SUBSET_CHECK:
               msg.append(" - value must be: ");
               msg.append(_int_to_string(_item->GetMinValue()));
            break; 
            case MIN_MAX_CHECK:
               msg.append(" - value must be between : ");
               msg.append(_int_to_string(_item->GetMinValue()));
               msg.append(" and ");
               msg.append(_int_to_string(_item->GetMaxValue()));
            break; 
            case MIN_ONLY_CHECK:
               msg.append(" - value must be greater than : ");
               msg.append(_int_to_string(_item->GetMinValue()));
            break; 
            case MAX_ONLY_CHECK:
               msg.append(" - value must be less than : ");
               msg.append(_int_to_string(_item->GetMaxValue()));
            break; 
            default:
               // no range to output
            break; 
         }
         theApp->status_message(msg);
      }
      else
	 clear_error_color(); 
      set(val.c_str());
   }
   return 0; 
}


int PegParEntryInteger::update_indir()
{

   if (_inedit)  return 0; 

   Parameter *par = NULL;
   string val = gtk_entry_get_text(GTK_ENTRY(_ind_txt));

   if ((par = thePData->GetParameter()) != NULL)
   {
      if (par->SetValueAtPosition(_item->GetPosition(), val) != 1)
      {
	 set_error_color(); 
      }
      else
      {
	 clear_error_color(); 
         string evl = par->EvalIndir((char*)_name.c_str(), (char*)val.c_str());
         if (evl.length())
            gtk_entry_set_text(GTK_ENTRY(_ind_val), evl.c_str());
         else
            gtk_entry_set_text(GTK_ENTRY(_ind_val), "\0");
      }
   }
   set(val.c_str());
   return 0;
}

// --------------------------------------------------------
// display the appropriate widgets for indirection/default
// --------------------------------------------------------
void PegParEntryInteger::manage()
{
   string tmp = get();

   gtk_widget_show(_lbl_w);
   switch(_disp_type)
   {
      case PegParEntry::INDIR:
	 gtk_widget_show_all(_ind_box);
	 gtk_widget_hide_all(_entry_box);
         if (tmp.length() && tmp.find(')') == 0)
            set(tmp.c_str());
	 else
         {
            gtk_entry_set_text(GTK_ENTRY(_ind_txt), "\0");
            gtk_entry_set_text(GTK_ENTRY(_ind_val), "\0");
         }
      break;
      case PegParEntry::DEFAULT:  // fallthrough intended
      default:
         gtk_widget_hide_all(_ind_box);
	 gtk_widget_show_all(_entry_box);
      break;
   }
}

void PegParEntryInteger::set_error_color()   
{
   if (_disp_type == PegParEntry::DEFAULT)
      gtk_widget_modify_bg(_spin_w, GTK_STATE_NORMAL, &_errcolor);
   else 
      gtk_widget_modify_bg(_ind_txt, GTK_STATE_NORMAL, &_errcolor);
}


void PegParEntryInteger::clear_error_color() 
{
   if (_disp_type == PegParEntry::DEFAULT)
      gtk_widget_modify_bg(_spin_w, GTK_STATE_NORMAL, NULL);
   else 
      gtk_widget_modify_bg(_ind_txt, GTK_STATE_NORMAL, NULL);
}

void PegParEntryInteger::set_adjustments()
{
   string title = "Adjustments for " + _name; 
   PegSpinAdjustDialog* dlg = new PegSpinAdjustDialog(title, _adj); 
   if (dlg) 
   {
      if (!dlg->doit())
      {
         // adjustments made 
         GtkAdjustment* adj2 = (GtkAdjustment *)
            gtk_adjustment_new(_item->GetValue(), _item->GetMinValue(),
	       _item->GetMaxValue(), _adj->step_increment,
	       _adj->step_increment, 0.0);
         gtk_spin_button_set_adjustment(GTK_SPIN_BUTTON(_spin_w), adj2); 
         _adj = adj2;
      }
      delete dlg; 
   }
}



bool PegParEntryInteger::check_valid_input(string val)
{
   Parameter *par = thePData->GetParameter();
   return (par->SetValueAtPosition(_item->GetPosition(), val) != 1);
}


    
