// 
//  Copyright (C) 2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PegPrefs.hh"
#include <gdk/gdk.h>
#include <fstream>
#include <iostream>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "PegUtils.hh"

#define PREF_MAX_LEN  1024


PegPrefs::PegPrefs()
{
   initialize(); 
}

PegPrefs::~PegPrefs()
{
}


void PegPrefs::initialize()
{
   _clobber = true;
   _autocolor = "";
   _hiddencolor = ""; 
   _automarkup = "i";
   _hiddenmarkup = "";
   _tips = true; 
   _showhidden = true; 
   _prompts = false; 
   load_preferences(); 
}

int PegPrefs::load_preferences(string fname)
{
   char   linebuffer[PREF_MAX_LEN];
   int    errs_found = 0;

   if (fname.length())
      filename = fname;
   else
   {
      filename = getenv("HOME");
      if (filename.length()) filename.append("/");
      filename.append(".pegrc");
   }

   ifstream *infile =  new ifstream(filename.c_str(), ios::in);

   if (!infile->good())  return 1;   // file failed to open

   while (infile->getline(linebuffer, PREF_MAX_LEN))
      errs_found += parse_line(linebuffer, ':');

   delete infile;

   return 0;
}


int PegPrefs::save_preferences(string in_fname)
{
    string fname;

    // set file name  
   if (in_fname.length() == 0)
      fname = filename;         // no filename provided, use stored filename 
   else 
      fname = in_fname;         // use provided filename

   // determine access status 
   if (fname.length() == 0) 
      return 2;                 // no file name specified;
   else if (!access(fname.c_str(), F_OK))
   {  
      if (access(fname.c_str(), W_OK) != 0)
         return 3;              // don't have write access 
   }  

   ofstream out(fname.c_str());
   if (!out)   return 4;       // file create failed
   
   out << "   # peg preferences file " << endl;
   out << endl; 
   if (_tips)         out << "display.tooltips : true" << endl;
      else            out << "display.tooltips : false" << endl;
   if (_prompts)      out << "display.prompts : true" << endl;
      else            out << "display.prompts : false" << endl;
   if (_showhidden)   out << "display.hidden : true" << endl;
      else            out << "display.hidden : false" << endl;
   if (_clobber)      out << "clobber : true" << endl;
      else            out << "clobber : false" << endl;
   out << endl;
   out << "   # auto.color|hidden.color may be a named color 'red' or #rrggbb '#ff0000'" << endl;
   out << "   # if not set (recommended), the default theme label color is used" << endl;
   out << endl; 
   out << "auto.color    : " << _autocolor << endl;
   out << "hidden.color  : " << _hiddencolor << endl;
   out << endl; 
   out << "   # auto.markup|hidden.markup may be a combination of the letters b|u|i for   " <<endl; 
   out << "   # bold, underline, or italic. If omitted, the default normal theme is used" <<endl; 
   out << endl; 
   out << "auto.markup   : " << _automarkup << endl;
   out << "hidden.markup : " << _hiddenmarkup << endl;
   out << endl;

   out.close();
   return 0;
}


int PegPrefs::parse_line(const char *in_txt, char del) 
{
   string::size_type sep;
   const char *pos = in_txt;
   string prop, val; 

   if (!in_txt)    return 1;  // no text 

   while(isspace(*pos)) pos++;  

   if (strlen(pos) == 0)   return 0;   // blank line  
   if (pos[0] == '#')      return 0;   // comment line 

   prop = pos; 
   sep = prop.find(del); 
   if (sep == string::npos)          return 2;   // no delimiter 
   val = prop.substr(sep+1, prop.length());
   prop = prop.substr(0, sep); 

   if (!val.length())    return 3;   // no value field
   if (!prop.length())    return 4;   // no prop field

   string::size_type first = val.find_first_not_of(" \t");
   if (first == string::npos)   val = "";
      else val = val.substr(first, val.find_last_not_of(" \t")-first+1);
   first = prop.find_first_not_of(" \t");
   if (first == string::npos)   prop = "";
      else prop = prop.substr(first, prop.find_last_not_of(" \t")-first+1);

   bool bv; 

   if (prop == "clobber")
   {
      if (!_string_to_bool(val, bv))     _clobber = bv;
      else    return 5; // value not a bool  
   }
   else if (prop == "display.tooltips")
   {
      if (!_string_to_bool(val, bv))     _tips = bv;
      else    return 5; // value not a bool  
   }
   else if (prop == "display.prompts")
   {
      if (!_string_to_bool(val, bv))     _prompts = bv;
      else    return 5; // value not a bool  
   }
   else if (prop == "display.hidden")
   {
      if (!_string_to_bool(val, bv))     _showhidden = bv;
      else    return 5; // value not a bool  
   }
   else if (prop == "auto.color")
      set_autocolor(val); 
   else if (prop == "hidden.color")
      set_hiddencolor(val); 
   else if (prop == "auto.markup")
      set_automarkup(val); 
   else if (prop == "hidden.markup")
      set_hiddenmarkup(val); 
   else 
      return 6;  // unknown prop
   return 0; 
}


int PegPrefs::get_preference(string prop, bool &val)
{
   if (prop == "clobber")
      val = _clobber;
   else if (prop == "display.tooltips")
      val = _tips;
   else if (prop == "display.prompts")
      val = _prompts;
   else if (prop == "display.hidden")
      val = _showhidden;
   else 
      return 1;   // value not an expected boolean 
   return 0;
}

int PegPrefs::get_preference(string prop, string &val)
{
   if (prop == "auto.color")
      val = _autocolor; 
   else if (prop == "auto.markup")
      val = _automarkup; 
   else if (prop == "hidden.color")
      val = _hiddencolor; 
   else if (prop == "hidden.markup")
      val = _hiddenmarkup; 
   else 
      return 1;   // value not an expected boolean 
   return 0; 
}

int PegPrefs::set_autocolor(string val)
{
   GdkColor clr;
   if (val.length() && gdk_color_parse(val.c_str(), &clr) == 0)  return 1;
   _autocolor = val;
   return 0; 
}

int PegPrefs::set_hiddencolor(string val)
{
   GdkColor clr;
   if (val.length() && gdk_color_parse(val.c_str(), &clr) == 0)  return 1;
   _hiddencolor = val;
   return 0; 
}

int PegPrefs::set_automarkup(string val)
{
   string tmp;
   if (val.find('B') != string::npos || val.find('b') != string::npos)  
      tmp.append("b"); 
   if (val.find('U') != string::npos || val.find('u') != string::npos)  
      tmp.append("u"); 
   if (val.find('I') != string::npos || val.find('i') != string::npos)  
      tmp.append("i"); 
   if (tmp.length() != val.length()) return 1; 
   _automarkup = tmp; 
   return 0; 
}

int PegPrefs::set_hiddenmarkup(string val)
{
   string tmp;
   if (val.find('B') != string::npos || val.find('b') != string::npos)  
      tmp.append("b"); 
   if (val.find('U') != string::npos || val.find('u') != string::npos)  
      tmp.append("u"); 
   if (val.find('I') != string::npos || val.find('i') != string::npos)  
      tmp.append("i"); 
   if (tmp.length() != val.length()) return 1; 
   _hiddenmarkup = tmp; 
   return 0; 
}



