// 
//  Copyright (C) 2007,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismApp.hh"
#include "PrismMenu.hh"
#include "PrismStatusArea.hh"
#include "PrismFileList.hh"
#include "PrismPrefs.hh"
#include "PrismPrefsDialog.hh"
#include "PrismUtils.hh"
#include "CiaoGuiIPCServer.hh"
#include "CiaoGuiIPCMessage.hh"

#include <numpy/noprefix.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

using namespace std;

PrismApp::PrismApp()
{
   _init = 0; 
   _fileList = NULL; 
   _showtips = TRUE; 
   menu = NULL; 
   check_dependencies(); 
   _prefs = new PrismPrefs(); 
   _tips = gtk_tooltips_new();
   reset_tips();
}

void PrismApp::initialize(string toolfile, string xpaname, char** filenames)
{
   if (!_init)
   {
      string filename = "";
      _pyModule = NULL;
      _prefsDialog = NULL; 
      statusArea = NULL; 
      def_height = 500;
      def_width  = 700; 
      int ndx = 0;
      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_default_size(GTK_WINDOW(window), def_width, def_height);
      vbox = gtk_vbox_new(false, 1);
      gtk_container_add (GTK_CONTAINER(window), vbox);

      gtk_container_set_border_width (GTK_CONTAINER (window), 0);

      // handle exit via <alt><f4> or close button in titlebar 
      g_signal_connect(window, "destroy", G_CALLBACK(exit_cb), &window);

      // create panes for contents and scrolled msg status area
      vpane = gtk_vpaned_new ();
      gtk_box_pack_end(GTK_BOX(vbox), vpane, 1, 1, 0);

      // add status area
      statusArea = new PrismStatusArea(); 
      gtk_paned_pack2(GTK_PANED(vpane), statusArea->base_widget(), TRUE, TRUE);
      statusArea->show();

      // set status area prefs
      if (_prefs && statusArea)
      {
         string clr = _prefs->error_text_color();
         if (clr.length())
	    statusArea->set_error_color(clr);
	 else
	    statusArea->set_color(0); 
         statusArea->set_bold(_prefs->error_text_bold()); 
         statusArea->set_italic(_prefs->error_text_italic()); 
      }

      // add menu after status area so Analysis Tool Menu can report status
      menu  = new PrismMenu(toolfile, _tips); 
      gtk_box_pack_start(GTK_BOX(vbox), menu->base_widget(), 0, 0, 0);
      menu->show(); 

      if (!_found_chips) status_message("Unable to locate chipsServer executable... plotting and histograms disabled."); 
      if (!_found_ds9) status_message("Unable to locate ds9 executable... imaging disabled."); 
      if (!_found_xpa) status_message("Unable to locate xpaset executable... imaging disabled."); 

      // add file frame
      _fileList = new PrismFileList();
      gtk_paned_pack1(GTK_PANED(vpane), _fileList->base_widget(), TRUE, TRUE);
      _fileList->show(); 

      if (filenames && filenames[ndx]) filename = filenames[ndx++]; 

      if (_fileList->add_file(filename) == 0)
         if (menu)  menu->update_filelist(filename); 

      while (filenames && filenames[ndx]) {
         filename = filenames[ndx++];
         load_file(filename, 1); 
      } 
      set_current_tab(0); 

      int loc = (int)(def_height * 0.8); 
      gtk_paned_set_position(GTK_PANED(vpane), loc);

      gtk_widget_show_all(vpane);
      gtk_widget_show_all(window);

      _img_dialog = true; 

      if (xpaname.length())
         _xpaname = xpaname;
      else
         _xpaname = "prism"; 

      add_ipc();

      _init = 1; 
  }
}

PrismApp::~PrismApp()
{
   cleanup();
   exit(0);
}

int PrismApp::status_message(string val, int type)
{
   if (!statusArea)  return -1;
   return statusArea->add_message(val, type);
}

void PrismApp::exit_cb()
{
   theApp->cleanup();
   exit(0);
}

int PrismApp::load_file(string filename, int new_tab, int editable)
{
   if (_fileList)
   {
      bool confdlg;
      if (editable && _prefs) 
      {
         confdlg = _prefs->use_edit_confirm_dialog(); 
         _prefs->set_edit_confirm_dialog(false);  
      }
      if (_fileList->add_file(filename, new_tab) == 0)
         if (menu) menu->update_filelist(filename); 
      if (editable && _prefs)
      {
	 lock_current_tab(1); 
         _prefs->set_edit_confirm_dialog(confdlg); 
      }
   }
   return 0;
}

string PrismApp::get_filename(int tab, int filter)
{
   // tab = -1 -> current tab    filter = 1 -> get filter 
   static string fname;
   if (_fileList)  fname = _fileList->get_filename(tab, filter); 
      else         fname = ""; 
   return fname; 
}

string PrismApp::get_current_filename()
{
   return theApp->get_filename(-1); 
}

void PrismApp::set_status_message(string val)
{
   theApp->status_message(val); 
}

int PrismApp::close_tab()
{
   if (_fileList)   _fileList->remove_tab(-1);
   return 0;
}

void PrismApp::show_tabs()
{
   if (_fileList)   _fileList->show_tabs();
}

void PrismApp::hide_tabs()
{
   if (_fileList)   _fileList->hide_tabs();
}

void PrismApp::reposition_tabs(int val)
{
   if (_fileList)   _fileList->rotate_tabs(val);
}

bool PrismApp::tabs_visible() 
{
   if (_fileList)  return _fileList->tabs_visible();
   return false; 
}

int PrismApp::get_num_tabs()
{
   if (_fileList)  return _fileList->get_num_tabs();
      else         return 0; 
}

int PrismApp::set_current_tab(int val)
{
   if (_fileList)  return _fileList->set_current_tab(val);
      else         return 1; 
}

int PrismApp::get_current_tab()
{
   if (_fileList)  return _fileList->get_current_tab();
      else         return -1;
}

int PrismApp::get_current_mode()
{
   if (_fileList)  return _fileList->get_current_mode(); 
      else         return -2; 
}

string PrismApp::get_status_messages()
{
   if (statusArea)  return statusArea->get_all_messages();  
      else  return ""; 
}

int PrismApp::get_status_msg_cnt()
{
   if (statusArea)  return statusArea->get_message_count();  
      else  return 0; 
}

void PrismApp::check_dependencies()
{
   FILE *fd;
   char buf[129];
   char *s;

   if ((fd = popen("\\which ds9  2>&1 > /dev/null ; echo $?", "r")) != NULL)
   {
      *buf = '\0';
      if (s=fgets(buf, 128, fd))   _found_ds9 = (s[0] == '0');
   }
   if ((fd = popen("\\which xpaset 2>&1 > /dev/null ; echo $?", "r")) != NULL)
   {
      *buf = '\0';
      if (s=fgets(buf, 128, fd))  _found_xpa = (s[0] == '0');
   }
   if ((fd = popen("\\which chipsServer 2>&1 > /dev/null ; echo $?", "r")) != NULL)
   {
      *buf = '\0';
      if (s=fgets(buf, 128, fd))  _found_chips = (s[0] == '0');
   }
}

void PrismApp::cleanup()
{
   char namebuf[256]; 
   int pd = (int)getpid();
   char* lgname = getlogin();
   if (lgname)  sprintf(namebuf,"/tmp/launch_chips.%s.%1d", lgname, pd);
      else      sprintf(namebuf,"/tmp/launch_chips.%1d", pd);
   if (access(namebuf, R_OK) == 0)   remove(namebuf);

   if (_prefs)
   {
      delete _prefs;
      _prefs = NULL;
   }
   if (_prefsDialog);
   {
      delete _prefsDialog;
      _prefsDialog = NULL;
   }
   if (_fileList)
   {
      delete _fileList;
      _fileList = NULL; 
   }
   if (window)
   {
      gtk_widget_destroy (window);
      window = NULL;
   }
   if (_pyModule)
   {
      Py_DECREF (_pyModule);
            _pyModule = static_cast <PyObject *> (0);
   }
}

int PrismApp::get_preference(string pref, bool &val)
{
   if (!_prefs)  return 1;    // no prefs object
   return _prefs->get_preference(pref, val);
}

int PrismApp::get_preference(string pref, long &val)
{
   if (!_prefs)  return 1;    // no prefs object
   return _prefs->get_preference(pref, val);
}

int PrismApp::get_preference(string pref, string &val)
{
   if (!_prefs)  return 1;    // no prefs object
   return _prefs->get_preference(pref, val);
}

string PrismApp::get_data_format(string cname, string ename, int typ)
{
   if (!_prefs)  return "";    // no prefs object
   return _prefs->get_format(cname, ename, typ);
}

void PrismApp::launch_preferences_dialog()
{
   if (!_prefs)
   {
      status_message("Unable to launch Preferences Dialog- no preferences object found.");
      return; 
   }
   if (!_prefsDialog)
   {
      _prefsDialog = new PrismPrefsDialog(_prefs);
      status_message("Preferences dialog will attempt to load $HOME/.prismrc"); 
   }
   if (_prefsDialog)    _prefsDialog->doit();

   reset_tips(); 
}

int PrismApp::visualize_current_tab(int type, int dialog)
{
   if (_fileList)   _fileList->visualize_tab(type, dialog);
   return 0;
}

int PrismApp::select_in_current_tab(int val, int as_col)
{
   if (_fileList)   _fileList->select_in_tab(val, as_col);
   return 0;
}

int PrismApp::reload_current_tab(int ext, long row)
{
   if (_fileList)   _fileList->reload_tab(ext, row);
   return 0;
}

int PrismApp::lock_current_tab(int mode)
{
   if (_fileList)   _fileList->lock_tab(mode);
   return 0;
}

bool PrismApp::check_column_expansion(string colname, int vdim, int adim)
{
   if (_prefs)  return _prefs->check_column_expansion(colname, vdim, adim);
   return false; 
}


int PrismApp::connect_plotter(string id)
{
   int rv = 0; 
   PyObject *py_retval;

   if (!_pyModule)
   {
      theApp->status_message("Preparing to connect to chips server... ");

      PyObject *pName;

      // Initialize python
      if (getenv ("CIAO_PYTHON_EXE"))
         Py_SetProgramName(getenv ("CIAO_PYTHON_EXE"));
      Py_Initialize();
      
      int zz=100;
      while (g_main_context_iteration(NULL, FALSE) &&zz--);   // refresh gui

      // Try to open module
      pName = PyString_FromString((char *)"pychips.all");
      _pyModule = PyImport_Import(pName);
      if (pName)  Py_DECREF(pName);
      if (_pyModule)
      {
         zz=100;
         while (g_main_context_iteration(NULL, FALSE) &&zz--);   // refresh gui

	 // request to connect to specific server
	 if (id.length())
	 {
            py_retval = call_py_functions (_pyModule, "connect", "%s",
	       id.c_str());
	    if (!py_retval)
	    {
	       rv = 1;
	       theApp->status_message("Failed to connect to chips server '" + id + "'");
	       ;
	    }
	    else 
	    {
               theApp->status_message("Connected to chips server '" + id + "'");
	       _chipsId = id; 
               Py_DECREF(py_retval);
	    }
	 }
         else 
	 {
            // create server and get the chips server id
            if ((py_retval = call_py_functions (_pyModule, "create_server")))
	    { 
               Py_DECREF(py_retval);
               py_retval = call_py_functions (_pyModule, "get_server_id");
	       if (!py_retval)
	          rv = 1;
	       else
	       {
	          _chipsId = PyString_AsString(py_retval);
                  theApp->status_message("Connected to chips server '" + _chipsId + "'");
                  Py_DECREF(py_retval);
	       }
	    }
	 }
      }
   }
   else
   {
      // if still attached to the same one - do nothing
      if (id.length() && id == _chipsId)   return 0; 
      if (!id.length() && _chipsId.length()) return 0; 

      // attempt to connect to the specified server
      py_retval = call_py_functions (_pyModule, "disconnect");
      if (py_retval)  Py_DECREF(py_retval);
      if (id.length())
         py_retval = call_py_functions (_pyModule, "connect", "%s",
            id.c_str());
      else 
         py_retval = call_py_functions (_pyModule, "create_server");
      if (!py_retval)
      {
         rv = 1;
	 if (id.length())
            theApp->status_message("Failed to connect to chips server '" + id + "'");
	 else
            theApp->status_message("Failed to connect to chips server");
      }
      else 
      {
         Py_DECREF(py_retval);
	 string tid; 
	 if (!id.length())
	 {
            py_retval = call_py_functions (_pyModule, "get_server_id");
	    if (py_retval)
	    {
	       tid = PyString_AsString(py_retval);
	       Py_DECREF(py_retval);
	    }
	 }
         else 
	    tid = id; 
         theApp->status_message("Changed chips server connection from '" + _chipsId + "' to '" + tid + "'");
         _chipsId = tid; 
      }
   }

   if (rv == 0 && _pyModule)
   {
      // load the prism histogram calls into pychips namespace
      string cmd;
      char *pth = getenv("ASCDS_INSTALL");
      if (pth)
      {
         cmd = "execfile ('";
         cmd.append(pth);
         cmd.append("/bin/prismHist.py')");
      }
      else   cmd = "execfile ('prismHist.py')";
      py_retval = PyRun_String(cmd.c_str(), Py_single_input, PyModule_GetDict(_pyModule), PyModule_GetDict(_pyModule));
      if (py_retval)  Py_DECREF(py_retval);
   }
   return rv; 
}


int PrismApp::print_current_plot()
{
   if (_pyModule)
   {
      PyObject *py_retval = call_py_functions (_pyModule, "print_window", "%s %s",
            "\0", "printdialog=true");
      if (py_retval)  Py_DECREF(py_retval);
   }
   return 0; 
}


// Create a function in python and call it.
// Input:
//    pModule = The python object module from which the function can be found
//    func_name = The name of the function to be called
//    format = The formats of the arguments to be passed into the python 
//      function. Acceptable formats include:
//          %s -- char * (ending in \0)
//          %S -- an array of char *
//          %d -- int
//          %D -- an array of int
//          %f -- double
//          %F -- an array of double
//          %p -- a Python Object
//          %a -- int indicating array size (NOTE this value MUST precede
//                any array argument. ie: (..., "%a %D", 3, {1,2,3})
PyObject *PrismApp::call_py_functions (PyObject *pModule, 
   const string &func_name, const string &format, ...)
{
   PyObject *pFunc;
   PyObject *pValue = static_cast <PyObject *>(0);
   PyObject *pArgs;
   va_list args;

   string::size_type pos;
   char type = ' ';
   int argcnt = 0;
   int array_size = 0;
   bool valid_format_found = true;
   int pArg_index = 0;

   // handle NULL module 
   if (!pModule)  pModule = _pyModule; 
   if (!pModule)   return pValue;

   // Get the function as pyobject from the module
   pFunc = PyObject_GetAttrString(pModule, (char *)func_name.c_str());
   
   // Make sure the object is callable
   if (pFunc == NULL || !PyCallable_Check(pFunc)) 
   {
      // Function, " + func_name + ", does not exist in the module.";
      return pValue;
   }
   
   // go through list of arguments twice. 1st time find out how many args
   // there are. 2nd time actually put the arguments in a py tuple
   for (int ii=0; ii<2; ii++)
   {
      if (ii == 1)
         va_start (args, format);
      pos = 0;

      while ((pos != string::npos) && (pos < format.length()))
      {
         pos = format.find ("%", pos);
         if ((pos != string::npos) && (pos < format.length()))
            type = format[++pos];
         else
            type = ' ';

         valid_format_found = true;
         switch (type)
         {
            case 'a':
               if (ii==1)
                  array_size = va_arg (args, int);
            break;
            case 's':
               if (ii==1)
               {
                  char *tmp = va_arg (args, char *);
                  pValue = PyString_FromString(tmp);
               }
            break;
            case 'd':
            {
               if (ii==1)
               {
                  long lval = static_cast <long> (va_arg (args, int));
                  pValue = PyInt_FromLong(lval);
               }
            }
            break;
            case 'D':
            {
               if ((ii==1) && (array_size > 0))
               {
                  int *ival = va_arg (args, int*);
                  long *lval = new long [array_size];
                   for (int ii=0; ii<array_size; ii++)
                     lval[ii] = static_cast <long> (ival[ii]);
                   int dims[1];
                  dims[0] = array_size;

                  pValue = PyArray_SimpleNewFromData(1, (npy_intp*)dims, NPY_LONG, lval);
                     delete []lval;
               }
               else if (ii == 1)
                  valid_format_found = false;
            }
            break;
            case 'f':
            {
               if (ii==1)
               {
                  double dval = va_arg (args, double);
                  pValue = PyFloat_FromDouble(dval);
               }
            }
            break;
            case 'F':
            {
               if ((ii==1) && (array_size > 0))
               {
                  double *dval = va_arg (args, double*);
                  int dims[1];
                  dims[0] = array_size;
                  pValue = PyArray_SimpleNewFromData(1, (npy_intp*)dims, NPY_DOUBLE, dval);
               }
               else if (ii==1)
                  valid_format_found = false;
            }
	    break; 
            case 'p':
            {
               if (ii==1)
                  pValue = va_arg (args, PyObject *);
            }
            break;
            default:
               valid_format_found = false;
            break;
         }

         if ((ii == 0) && valid_format_found)
            argcnt++;
         else if ((ii == 1) && valid_format_found)
         {
            // Add the pyobject value to the python argument list
            if (!pValue) 
            {
               Py_DECREF(pArgs);
               //Could not convert argument to python object;
               return pValue;
            }
            PyTuple_SetItem(pArgs, pArg_index++, pValue);
         }
      }

      // Create a python type as the argument list for the function
      if (ii == 0)
      {
         pArgs = PyTuple_New (argcnt);

         // if the are no arguments dont repeat loop
         if (argcnt == 0)
            break;
      }
   }

   // call the actual function
   pValue = PyObject_CallObject(pFunc, pArgs);

   Py_DECREF(pArgs);
   if (pValue) 
      Py_DECREF(pFunc);

   // return the results of the call
   return pValue;
}

void PrismApp::set_modified()
{
   if (_fileList)   _fileList->set_modified();
}

void PrismApp::reset_tips()
{
   if (_tips)
   {
      if (_prefs) _showtips = _prefs->display_tooltips();
      if (_showtips)
         gtk_tooltips_enable(GTK_TOOLTIPS(_tips)); 
      else 
         gtk_tooltips_disable(GTK_TOOLTIPS(_tips)); 
   }
}

int PrismApp::get_tab_position()
{
   int val = -1; 
   if (_fileList)  val = _fileList->get_tab_position(); 
   return val; 
}


CiaoGuiIPCMessage* PrismApp::VizMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   //  histogram | image | plot
   int typ, numParam = in_msg.GetTotalParams();
   string param1;
   in_msg.GetParam(0, param1);

   if (numParam > 1) 
   {
      returnMsg = new CiaoGuiIPCMessage("Error: Too many arguments... xpaset/get " + theApp->get_xpaname() + " image|histogram|plot\n");
      return returnMsg; 
   }
   if (param1 == "histogram")          typ =1;
      else if (param1 == "plot")       typ = 0;
         else if (param1 == "image")   typ = 2;
   theApp->visualize_current_tab(typ, 0); 
   return returnMsg; 
}


CiaoGuiIPCMessage* PrismApp::GetValueMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   //  get id
   int numParam = in_msg.GetTotalParams();
   string param1;
   in_msg.GetParam(1, param1);

   if (numParam > 2) 
      returnMsg = new CiaoGuiIPCMessage("Error: Too many arguments... xpaset/get " + theApp->get_xpaname() + " get id\n");
   else if (param1 == "id")
   {
      param1 = theApp->get_chipsId();
      if (param1.length())
         returnMsg = new CiaoGuiIPCMessage((char*)param1.c_str()); 
   }
   else 
      returnMsg = new CiaoGuiIPCMessage("Error: Unknown option... xpaset/get " + theApp->get_xpaname() + " get id\n");
   return returnMsg; 
}

CiaoGuiIPCMessage* PrismApp::SelectMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   //  select col|row #|all|none
   int typ, num = in_msg.GetTotalParams();
   string param1, param2;
   in_msg.GetParam(1, param1);
   in_msg.GetParam(2, param2);

   if (num != 3) 
   {
      returnMsg = new CiaoGuiIPCMessage("Error: Incorrect number of arguments... xpaset/get " + theApp->get_xpaname() + " select col|row #|all|none\n");
      return returnMsg; 
   }
   if (!_strcmp_cis(param1.c_str(),"row")) typ = 0;
      else if (!_strcmp_cis(param1.c_str(),"col")) typ = 1;
         else 
            return (new CiaoGuiIPCMessage("Error: Argument number 2 must be col|row... xpaset/get " + theApp->get_xpaname() + " select col|row #|all|none\n"));

   if (!_strcmp_cis(param2.c_str(),"all"))          num = -1;  // select all
      else if (!_strcmp_cis(param2.c_str(),"none")) num = -2;  // deselect all
	 else   in_msg.GetParam(2, num); 
   theApp->select_in_current_tab(num, typ); 
   return returnMsg; 
}

CiaoGuiIPCMessage* PrismApp::FileMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   
   //  file [-newtab] [filename]
   int numParam = in_msg.GetTotalParams();
   if (numParam > 3)
   {
      // too many parameters
      returnMsg = new CiaoGuiIPCMessage("Error: Too many arguments... xpaset/get " + theApp->get_xpaname() + " file [-newtab] [filename]\n");
   }
   else if (numParam == 3) 
   {
      // set filename request 
      int tb;
      string param1, param2, fname;
      in_msg.GetParam(1, param1);
      in_msg.GetParam(2, param2);
      if (param1 == "-newtab")
      {
         tb = 1;
         fname = param2;
         theApp->load_file(fname, tb);
      }
      else if (param2 == "-newtab")
      {
         tb = 1;
         fname = param1;
         theApp->load_file(fname, tb);
      }
      else
      {
         returnMsg = new CiaoGuiIPCMessage("Error: Invalid arguments... xpaset/get " + theApp->get_xpaname() + " file [-newtab] [filename]\n");
      }
   }
   else if (numParam == 2) 
   {
      // set filename request 
      int tb = 0;
      string param1, fname = "";
      in_msg.GetParam(1, param1);
      if (param1 == "-newtab")   tb = 1;
         else  fname = param1;
      theApp->load_file(fname, tb);
   }
   else
   {
      // get current filename request 
      string fname = theApp->get_filename(); 
      if (fname.length())
         returnMsg = new CiaoGuiIPCMessage((char*)fname.c_str()); 
      else 
         returnMsg = NULL; 
   }
   return returnMsg;
}

CiaoGuiIPCMessage* PrismApp::StatusMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   
   // status [-msgs | -cnt ]
   int numParam = in_msg.GetTotalParams();
   if (numParam !=2)
   {
      // incorrect number of parameters
      returnMsg = new CiaoGuiIPCMessage("Error: Incorrect number of parameters... xpaget " + theApp->get_xpaname() + " status {-msgs|-cnt}\n");
   }
   else 
   {
      string param1;
      in_msg.GetParam(1, param1);
      if (param1 == "-msgs")
      {
	 string ms = theApp->get_status_messages();
         returnMsg = new CiaoGuiIPCMessage((char*)ms.c_str());
      }
      else if (param1 == "-cnt") 
      {
         returnMsg = new CiaoGuiIPCMessage();
         returnMsg->AddParam(theApp->get_status_msg_cnt());
      }
      else 
         returnMsg = new CiaoGuiIPCMessage("Error: Unsupported option... xpaget " + theApp->get_xpaname() + " status {-msgs|-cnt}\n");
   }
   return returnMsg;
}


CiaoGuiIPCMessage* PrismApp::TabMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;
   
   // tab [-add|-remove|-lock|-unlock|-mode|-show|-hide|-left|-right|
   //      -top|-bottom|-position|-visible|-count|<index>] 

   int numParam = in_msg.GetTotalParams();
   if (numParam !=2)
   {
      // incorrect number of parameters
      returnMsg = new CiaoGuiIPCMessage("Error: Incorrect number of parameters... xpaset/get " + theApp->get_xpaname() + " tab expects 1 argument\n");
      return returnMsg;
   }
   string param1;
   in_msg.GetParam(1, param1);

   if (param1 == "-add")
      theApp->load_file("", 1); 
   else if (param1 == "-remove")
      theApp->close_tab(); 
   else if (param1 == "-lock")
      theApp->lock_current_tab(0);
   else if (param1 == "-unlock")
      theApp->lock_current_tab(1);
   else if (param1 == "-mode")
   {
      returnMsg = new CiaoGuiIPCMessage();
      switch(theApp->get_current_mode())
      {
        case -1: returnMsg->AddParam("read only"); break;  
        case 1:  returnMsg->AddParam("locked"); break;  
        case 2:  returnMsg->AddParam("unlocked"); break;  
        case 3:  returnMsg->AddParam("modified"); break;  
        default: returnMsg->AddParam("unknown"); break;  
      }
   }
   else if (param1 == "-show")
      theApp->show_tabs(); 
   else if (param1 == "-hide")
      theApp->hide_tabs(); 
   else if (param1 == "-left")
      theApp->reposition_tabs(0); 
   else if (param1 == "-right")
      theApp->reposition_tabs(1); 
   else if (param1 == "-top")
      theApp->reposition_tabs(2); 
   else if (param1 == "-bottom")
      theApp->reposition_tabs(3); 
   else if (param1 == "-position")
   {
      returnMsg = new CiaoGuiIPCMessage();
      switch(theApp->get_tab_position())
      {
        case 0:  returnMsg->AddParam("left"); break;  
        case 1:  returnMsg->AddParam("right"); break;  
        case 2:  returnMsg->AddParam("top"); break;  
        case 3:  returnMsg->AddParam("bottom"); break;  
        default: returnMsg->AddParam("unknown"); break;  
      }
   }
   else if (param1 == "-visible")
   {
      if (theApp->tabs_visible())
         returnMsg = new CiaoGuiIPCMessage("true");
      else   
         returnMsg = new CiaoGuiIPCMessage("false");
   }
   else if (param1 == "-count")
   {
      returnMsg = new CiaoGuiIPCMessage();
      returnMsg->AddParam(theApp->get_num_tabs());
   }
   else if (param1 == "-current")
   {
      returnMsg = new CiaoGuiIPCMessage();
      int pg = theApp->get_current_tab();
      if (pg < 0)  returnMsg->AddParam("no current tab"); 
         else      returnMsg->AddParam(pg); 
   }
   else
   {
      short pos; 
      if (_string_to_short(param1, pos))
      {
         // conversion failed 
         returnMsg = new CiaoGuiIPCMessage("Error: Expected a tab flag or numeric tab position\n");
         return returnMsg;
      } 
      theApp->set_current_tab(pos); 
   }
   return returnMsg;
}

void PrismApp::add_ipc()
{
   ipc = new CiaoGuiIPCServer((char*)_xpaname.c_str(), 1);
   if (ipc)
   {
      ipc->AddCallback("file", &PrismApp::FileMessageCB,
         "file [[-newtab] <filename>] : loads filename, or retrieves file if no args specified");
      ipc->AddCallback("get", &PrismApp::GetValueMessageCB,
         "get id          : retrieves the chips server id used for plots/histograms");
      ipc->AddCallback("histogram", &PrismApp::VizMessageCB,
         "histogram       : generate a histogram of the selected column");
      ipc->AddCallback("image", &PrismApp::VizMessageCB, 
         "image           : generate an image of the current file");
      ipc->AddCallback("plot", &PrismApp::VizMessageCB,
         "plot            : generate plot of selected columns");
      ipc->AddCallback("status", &PrismApp::StatusMessageCB,
         "status -msgs|-cnt: retrieves messages or the number of messages in the status area");
      ipc->AddCallback("select", &PrismApp::SelectMessageCB,
         "select col|row all|none|#: selects/deselects all, none, or the specified row/col");
      ipc->AddCallback("tab", &PrismApp::TabMessageCB, 
         "tab -add|-remove|-lock|-unlock|-mode|-show|-hide|-left|-right|-top|-bottom|\n     -position|-count|-current|<current_index> : tab operations");
      ipc->GtkLoopAdd();
   }
}

void PrismApp::change_error_color(string cname)
{
   if (statusArea)  statusArea->set_error_color(cname);
}

void PrismApp::set_error_color(bool val)
{
   if (statusArea)  statusArea->set_color(val);
}

void PrismApp::set_error_bold(bool val)
{
   if (statusArea)  statusArea->set_bold(val);
}

void PrismApp::set_error_italic(bool val)
{
   if (statusArea)  statusArea->set_italic(val);
}



