// 
//  Copyright (C) 2007,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismApp.hh"
#include "PrismAppendColumnCmd.hh"
#include "PrismAppendRowCmd.hh"
#include "PrismDataArea.hh"
#include "PrismDataBlock.hh"
#include "PrismDataColumn.hh"
#include "PrismDataSet.hh"
#include "PrismDataTable.hh"
#include "PrismDataToolBar.hh"
#include "PrismFormatDialog.hh"
#include "PrismMetaDialog.hh"
#include "PrismDataJumpCmd.hh"
#include "PrismImageCmd.hh"
#include "PrismHistogramCmd.hh"
#include "PrismPlotCmd.hh"

using namespace std;

PrismDataArea::PrismDataArea(string extname):PrismBaseObj()
{
   string ename = "";
   if (extname.length())   ename = extname.c_str(); 
   base_w = gtk_frame_new(ename.c_str());

   vbox = gtk_vbox_new(false, 1); 
   gtk_container_add (GTK_CONTAINER (base_w), vbox);

   dataBtns = new PrismDataToolBar(this);
   gtk_box_pack_start (GTK_BOX(vbox), dataBtns->base_widget(), 0, 0, 2);

   table = new PrismDataTable(this, ename);
   gtk_box_pack_start (GTK_BOX(vbox), table->base_widget(), 1, 1, 2);
   dataBtns->disable_all();  
   dataBtns->set_nav_button_sensitivity(1,1); // disable nav buttons

   // create image command object so that we can maintain default values
   // for subsequent invocations 
   _imager = new PrismImageCmd(); 
   _plotter = new PrismPlotCmd(); 
   _histoCmd = new PrismHistogramCmd(); 
   _jumpCmd = new PrismDataJumpCmd(); 
   _appendColumn = new PrismAppendColumnCmd(); 
   _appendRow = new PrismAppendRowCmd(); 
   
}

PrismDataArea::~PrismDataArea()
{
   if (table)
   {
      delete table;
      table = NULL; 
   }
   if (_imager)
   {
      delete _imager;
      _imager = NULL; 
   }
   if (_plotter)
   {
      delete _plotter;
      _plotter = NULL; 
   }
   if (_histoCmd)
   {
      delete _histoCmd;
      _histoCmd = NULL;
   }
   if (_jumpCmd)
   {
      delete _jumpCmd;
      _jumpCmd = NULL;
   }
   if (_appendColumn)
   {
      delete _appendColumn;
      _appendColumn = NULL;
   }
   if (_appendRow)
   {
      delete _appendRow;
      _appendRow = NULL;
   }
}

int PrismDataArea::set_data(PrismDataSet* dset, long row)
{
   if (!dset)  return 1;  // no data set 
   _dset = dset; 
   if (table)
   {
      long cp, np; 
      string ename = _dset->get_current_extname();
      PrismDataBlock* pb = dset->get_block(); 
      table->load_entries(pb, row);
      cp = table->get_current_page(); 
      np = table->get_npages();

      // enable buttons
      int flag = 0;
      if (pb->is_image()) flag = 2;
      dataBtns->enable_all(flag);  

      // set button sensitivity
      dataBtns->set_nav_button_sensitivity(cp, np); 
      // disable plotting for images 
      if (pb)   dataBtns->set_plot_button_sensitivity(!(pb->is_image()));
      if (pb)   dataBtns->set_hist_button_sensitivity(!(pb->is_image()));

      // disable imaging for ascii files 
      if (_dset->is_ascii())  dataBtns->set_image_button_sensitivity(false);
   }
   return 0; 
}

int PrismDataArea::show()
{
   if (table) table->show();
   if (dataBtns) dataBtns->show();
   gtk_widget_show_all(vbox);
   gtk_widget_show_all(base_w);
   return 0;
}


int PrismDataArea::next_page()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table
   if (table->next_page())  return 3;  // call failed

   long cp = table->get_current_page(); 
   long np = table->get_npages();
   dataBtns->set_nav_button_sensitivity(cp, np);   
   return 0;
}

int PrismDataArea::prev_page()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table
   if (table->prev_page())  return 3;  // call failed

   long cp = table->get_current_page(); 
   long np = table->get_npages();
   dataBtns->set_nav_button_sensitivity(cp, np);   
   return 0;
}

int PrismDataArea::first_page()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table
   if (table->first_page())  return 3;  // call failed
   dataBtns->set_nav_button_sensitivity(table->get_current_page(),
      table->get_npages());
   return 0;
}

int PrismDataArea::goto_page()
{
   if (!table)    return 1;   // no table
   if (!_jumpCmd) return 2;   // no jump cmd set 
   int rv = _jumpCmd->doit(table); 
   dataBtns->set_nav_button_sensitivity(table->get_current_page(),
      table->get_npages());
   return rv; 
}



int PrismDataArea::last_page()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table
   if (table->last_page())  return 3;  // call failed
   dataBtns->set_nav_button_sensitivity(table->get_current_page(),
      table->get_npages());
   return 0;
}

int PrismDataArea::get_meta_data()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table

   PrismDataColumn* selcols = NULL; 
   long cnt; 

   if ((selcols = table->get_selected_columns(cnt)) != NULL)
   {
      string title =  "Select Meta Data";
      string exn = _dset->get_current_extname(); 
      if (exn.length())
      {
         title.append(" for ");
         title.append(exn);
      }
      PrismMetaDialog *dialog = new PrismMetaDialog(title, selcols, cnt); 
      if (dialog)  dialog->doit(); 
      delete[] selcols; 
   }
   else
      theApp->status_message("Invalid Meta-Data request- please select one or more columns to display..."); 
   return 0;
}

int PrismDataArea::format_column()
{
   if (!_dset)  return 1;  // no data set
   if (!table)  return 2;  // no table

   PrismDataColumn* datacols = NULL;
   long cnt; 

   // datacols contains pointers to actual data columns- do not delete
   if ((datacols = table->get_data_columns(cnt)) != NULL)
   {
      string title =  "Format Display of Columns";
      string exn = _dset->get_current_extname();
      if (exn.length())
      {
	 title.append(" for ");
         title.append(exn);
      }
      PrismFormatDialog *dialog = new PrismFormatDialog(title, datacols, cnt);
      if (dialog)  if (dialog->doit() == 1)  table->reload_page();
   }
   else
      theApp->status_message("Invalid Column Display Formatting request- no columns exist."); 
   return 0; 
}


int PrismDataArea::generate_image(int dialog)
{
   if (!_dset)  return 1;  // no data set
   if (_imager)  _imager->doit(_dset, dialog); 
   return 0; 
}

int PrismDataArea::generate_histogram(int dialog)
{
   if (!_dset)  return 1;  // no data set
   if (!_histoCmd) return 2;  // no histogram cmd 

   PrismDataBlock* pb = _dset->get_block();
   if (pb->is_image())
   {
      theApp->status_message("Unable to histogram image data"); 
      return 3;  // can't histogram image data 
   }

   PrismDataColumn* selcols = NULL;
   string usecol; 
   long cnt;

   if ((selcols = table->get_selected_columns(cnt)) != NULL)
   {
      if (selcols)  usecol = selcols[0].get_name(); 
      delete[] selcols;
   }
   _histoCmd->doit(_dset, usecol, dialog); 
   return 0; 
}

int PrismDataArea::generate_plot(int dialog)
{
   if (!_dset)  return 1;  // no data set
   if (!_plotter) return 2;  // no plotter cmd 

   PrismDataBlock* pb = _dset->get_block();
   if (pb->is_image())
   {
      theApp->status_message("Unable to generate a plot from image data"); 
      return 3;  // can't plot image data 
   }

   PrismDataColumn* selcols = NULL;
   string xcol = "", ycol ="", ecol ="", rowspec; 
   long cnt;

   if ((selcols = table->get_selected_columns(cnt)) != NULL)
   {
      if (cnt > 0) xcol = selcols[0].get_name();
      if (cnt > 1) ycol = selcols[1].get_name();
      if (cnt > 2) ecol = selcols[2].get_name();
      delete[] selcols;
   } 
   rowspec = table->get_selected_rowspec(); 
   _plotter->doit(_dset, xcol, ycol, ecol, dialog, rowspec); 
   return 0; 
}

int PrismDataArea::append_column(int dialog)
{
   if (!table)  return 1;  
   if (!_dset)  return 1;  // no data set
   if (_appendColumn)  _appendColumn->doit(table); 
   return 0; 
}

int PrismDataArea::append_row(int dialog)
{
   if (!table)  return 1;  
   if (!_dset)  return 1;  // no data set
   if (_appendRow)  _appendRow->doit(table); 
   return 0; 
}


int PrismDataArea::set_selection(int val, int ascol)
{
   if (!table) return 1;

   if (ascol)
      switch (val)
      {
         case -1:   table->select_all_columns();     break;  
         case -2:   table->deselect_all_columns();   break;  
	 default:   
	    if (val < 0)  return 1;
	    table->select_column(val);     
	 break;
      }
   else 
      switch (val)
      {
         case -1:   table->select_all_rows();     break;  
         case -2:   table->deselect_all_rows();   break;  
	 default:   
	    if (val < 0)  return 1;
	    table->select_row(val);     
	 break;
      }
   return 0;
}


void PrismDataArea::set_frame_label(string val)
{
   gtk_frame_set_label(GTK_FRAME(base_w), val.c_str());
}

void PrismDataArea::disable_editing()
{
   if (dataBtns)  dataBtns->disable_editing(); 
   if (table)     table->disable_editing(); 
}

void PrismDataArea::enable_editing()
{
   if (dataBtns)  dataBtns->enable_editing(); 
   if (table)     table->enable_editing(); 
}

void PrismDataArea::set_toolbar_sensivity(int btn, bool sensitive)
{
   if (!dataBtns)  return; 
   
   if ((btn & PRISM_DATA_INFO_BTN) == PRISM_DATA_INFO_BTN)
      dataBtns->set_info_button_sensitivity(sensitive);
   if ((btn & PRISM_DATA_PLOT_BTN) == PRISM_DATA_PLOT_BTN)
      dataBtns->set_plot_button_sensitivity(sensitive);
   if ((btn & PRISM_DATA_HIST_BTN) == PRISM_DATA_HIST_BTN)
      dataBtns->set_hist_button_sensitivity(sensitive);
}


string PrismDataArea::get_selected_cols()
{
   static string filter;

   filter = "";

   if (!(_dset && table)) 
   {
      theApp->status_message("The current tab does not contain a file- unable to reopen with selected columns.");
      return filter; // no data set or table
   }
   PrismDataBlock* pb = _dset->get_block();
   if (pb->is_image())   
   {
      theApp->status_message("Reopening an image with selected columns is not supported.");
      return filter;   // column filtering only for tables
   }

   PrismDataColumn* selcols = NULL;
   long rr, cnt;

   if ((selcols = table->get_selected_columns(cnt)) != NULL)
   {
      if (cnt)
      {
         filter = "[cols "; 
         string colname; 
         string::size_type p1; 
         for (rr = 0; rr < cnt; rr++)
         {
            colname = selcols[rr].get_name();
            p1 = filter.find(colname);
            if (p1 != string::npos && filter[p1+colname.length()] == ',' &&
                (filter[p1-1] == ' ' || filter[p1-1] == ','))
            {
               // entry already in filter spec- ignore (ie. array column name)
            }
            else 
            {
               filter.append(selcols[rr].get_name());
               filter.append(",");
            }
         }
         if (filter[filter.length()-1] == ',')
            filter[filter.length()-1] = ']';
         else 
            filter = ""; 
      }
      delete[] selcols;
   }
   return filter; 
}

