// 
//  Copyright (C) 2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//



#include "PrismDataBlock.hh"
#include "PrismDataColumn.hh"
#include "PrismDataTable.hh"
#include "PrismDataToolBar.hh"
#include "PrismDataArea.hh"
#include "PrismTableColumn.hh"
#include "PrismApp.hh"

#include <sstream>
#include <iostream>


#define PRISM_MAX_EXP_SZ 100
#define PRISM_IDLE_LOAD_SZ 25
using namespace std;

PrismDataTable::PrismDataTable(PrismDataArea* parent, string exname):
   PrismBaseObj()
{
   _sid1 = _sid2 = 0; 
   _idletag = 0; 
   rows = cols = 10; 
   numpages = 1; 
   string fgcolor, bgcolor, bgodd, bgeven; 
   if (theApp->get_preference("data.rows", pagesize))  pagesize = 100;
   if (theApp->get_preference("data.select.fg", fgcolor))  fgcolor = "yellow";
   if (theApp->get_preference("data.select.bg", bgcolor))  bgcolor = "black";
   if (theApp->get_preference("data.oddrow.bg", bgodd))  bgodd = "";
   if (theApp->get_preference("data.evenrow.bg", bgeven))  bgeven = "";
   if (theApp->get_preference("data.alternate.rows", _altrow)) _altrow = 0;
   if (!(bgodd.length() && bgeven.length()))  _altrow = 0; 

   base_w = gtk_scrolled_window_new (NULL,NULL);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(base_w), 
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   sheet = gtk_sheet_new(rows, cols, exname.c_str()); 
   gdk_color_parse(bgcolor.c_str(), &_select_bg);
   gdk_color_alloc(gtk_widget_get_colormap(sheet), &_select_bg);
   gdk_color_parse(fgcolor.c_str(), &_select_fg);
   gdk_color_alloc(gtk_widget_get_colormap(sheet), &_select_fg);
   gdk_color_parse(bgodd.c_str(), &_oddrow_bg);
   gdk_color_alloc(gtk_widget_get_colormap(sheet), &_oddrow_bg);
   gdk_color_parse(bgeven.c_str(), &_evenrow_bg);
   gdk_color_alloc(gtk_widget_get_colormap(sheet), &_evenrow_bg);

   gtk_sheet_set_selection_mode(GTK_SHEET(sheet), GTK_SELECTION_NONE); 
   gtk_container_add(GTK_CONTAINER(base_w), sheet);
   _parent = parent; 
   _selected_cols = NULL;
   _selected_rows = NULL; 
   _selcval = _selrpp = _selpps = 0; 
   _valid_edit = 1;
   _loading = false;
   _floatmenu = NULL;
   _expcol = NULL; 
   _editable = false;
   curr_page = 1; 
   _data = NULL; 
   _load_last_row = _load_curr_row = 0;
}

PrismDataTable::~PrismDataTable()
{
   if (_idletag != 0)
   {
      g_source_remove(_idletag); 
      _idletag = 0; 
   }
   _load_last_row = _load_curr_row = 0;
   if (_selected_cols)  
   {
      delete[] _selected_cols;
      _selected_cols = NULL;
   }
   if (_selected_rows && _selpps)  
   {
      for (int rr = _selpps; rr--; )
         if (_selected_rows[rr])
	 {
	    delete[] _selected_rows[rr];
            _selected_rows[rr] = NULL;
	 }
      delete[] _selected_rows;
      _selected_rows = NULL;
   }
   if (!_collist.empty())
   {
      list<PrismTableColumn *>::iterator ii;
      for (ii = _collist.begin(); ii != _collist.end(); ii++)
	 if ((PrismTableColumn*)(*ii) != NULL)   delete (*ii); 
      _collist.clear(); 
   }
   if (_data)
   {
      delete[] _data;
      _data = NULL; 
   }
   if (base_w && sheet)
   {
      gtk_container_remove(GTK_CONTAINER(base_w), sheet);
      sheet = NULL; 
   }
}

int PrismDataTable::load_entries(PrismDataBlock* blk, long row)
{
   if (!blk)  return 1;   // no block;
   _load_last_row = _load_curr_row = 0;
   _block = blk; 
   rows = _block->get_nrows();
   string exname = _block->get_name(); 
   int disp_cols;
   
   cols = setup_table_columns(); 
 
   if (theApp->get_preference("data.rows", pagesize))  pagesize = 100;

   numpages = rows / pagesize;
   if (rows % pagesize) numpages++; 

   // only deal w/ a page of data at a time
   if (rows < 1 || rows > pagesize)  _disp_rows = pagesize + 1;
      else                           _disp_rows = rows + 1; 
   if (cols < 1)   disp_cols = 10; 
      else disp_cols = cols; 

   _selectMode = 0;
   if (cols > 0)
   {
      if (_selected_cols) delete[] _selected_cols; 
      _selected_cols = new int[cols];
      for (int zz = cols; zz--; ) _selected_cols[zz] = _selectMode; 
   }
   if (rows > 0 && numpages > 0)
   {
      int rr; 
      if (_selected_rows && _selpps > 0) 
      {
	 for (rr = _selpps; rr--;)
	    if (_selected_rows[rr])  
	    {
	       delete[] _selected_rows[rr]; 
	       _selected_rows[rr] = NULL; 
	    } 
         delete[] _selected_rows; 
      }
      _selected_rows = new int*[_selpps=numpages];
      for (int rr = _selpps; rr--; ) _selected_rows[rr] = NULL;
      _selected_rows[0] = new int[_selrpp = pagesize];
      for (int rr = _selrpp; rr--; ) _selected_rows[0][rr] = _selectMode;
   }

   if (sheet)
   {
#ifdef WLM_REMOVE_CBS
   // commented out since it generates warnings in ciaox build
   if (_sid1 > 0)
   {
      g_signal_handler_disconnect(sheet, _sid1);
      _sid1 = 0;
   }
   if (_sid2 > 0) 
   {
      g_signal_handler_disconnect(sheet, _sid2);
      _sid2 = 0;
   }
#endif 
      // remove old sheet from scrolled window 
      gtk_container_remove(GTK_CONTAINER(base_w), sheet);

      // reset scrollbar so new sheet is correctly oriented 
      GtkAdjustment* adj = gtk_scrolled_window_get_hadjustment(
         GTK_SCROLLED_WINDOW(base_w));
      gtk_adjustment_set_value(adj, 0.0);
      gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(base_w), adj);

      // create a new sheet to replace the existing one
      sheet = gtk_sheet_new(_disp_rows, disp_cols, exname.c_str());
      gtk_sheet_set_selection_mode(GTK_SHEET(sheet), GTK_SELECTION_NONE); 

      // set up column titles
      PrismDataColumn* columns = _block->get_columns(); 

      gtk_sheet_freeze(GTK_SHEET(sheet));

      if (columns && !_collist.empty())
      {
         string cname, units, btn_name; 
	 GtkSheetRange range = { 0,0,0,0}; 
	 PrismTableColumn *tcp; 
         PrismDataColumn* cp; 
         long pos1, pos2;
         char bufr[80];

         list<PrismTableColumn *>::iterator ii;

	 for (ii = _collist.begin(); ii != _collist.end(); ii++)
	    if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
	    {
	       pos1 = tcp->get_colmin(); 
	       pos2 = tcp->get_colmax(); 
	       if (tcp->is_image())
	       {
		  for (int cc = 0; cc <= pos2; cc++) 
		  {
                     sprintf(bufr, "%1d", cc+1);
	             gtk_sheet_column_button_add_label(GTK_SHEET(sheet), 
	                                               cc, bufr);  
	          }
	       }
	       else if (pos1 != -1) 
	       {
		  if (pos1 == pos2)   // single entry  
		  {
	             cp = tcp->get_data_column(); 
                     cname = cp->get_name(); 
                     if (cp->is_vector_array())  btn_name = cname + "()[]";
		     else if (cp->is_vector())   btn_name = cname + "()";
		     else if (cp->is_array())    btn_name = cname + "[]";
		     else if (cp->is_vector_comp())
		        btn_name = cp->get_vector_name() + "("+cname+")";
		     else                        btn_name = cname;
	             gtk_sheet_column_button_add_label(GTK_SHEET(sheet),pos1, 
	                                               btn_name.c_str());  
                     gtk_sheet_column_set_visibility(GTK_SHEET(sheet),pos1,
	                                             cp->is_visible());
	             units = cp->get_units(); 
                     gtk_sheet_set_cell(GTK_SHEET(sheet), 0, pos1,
	                GTK_JUSTIFY_LEFT, units.c_str());
		  }
		  else for (int cc = pos1; cc <= pos2; cc++) // array|vector array
		  {
	             cp = tcp->get_data_column(); 
		     cname = cp->get_name(); 
                     sprintf(bufr, "%s[%1ld]", cname.c_str(), cc - pos1);
	             gtk_sheet_column_button_add_label(GTK_SHEET(sheet), cc,
	                                               bufr);  
                     gtk_sheet_column_set_visibility(GTK_SHEET(sheet), cc,
	                                             cp->is_visible());
	             units = cp->get_units(); 
                     gtk_sheet_set_cell(GTK_SHEET(sheet), 0, cc,
	                GTK_JUSTIFY_LEFT, units.c_str());
	          }
	       }
	    } 

	 // set expanded / collapsed 
	 for (ii = _collist.begin(); ii != _collist.end(); ii++)
	    if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
            {
	       cp = tcp->get_data_column();
               if ((cp->is_vector() && !cp->is_vector_comp()) || 
                   (cp->is_array() && !cp->is_array_comp()) &&
		   (tcp->get_state() != PRISM_COLSTATE_SKIP))
	       {
	          if (theApp->check_column_expansion(cp->get_name(),
		      cp->get_dim(), cp->get_array_size()))
		     toggle_column(tcp);
		  // set collapsed vectors/arrays an uneditable
		  if ((range.col0 = range.coli = tcp->get_colmin()) != -1)
		     gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 0);
	       }
	    }

         gtk_container_add(GTK_CONTAINER(base_w), sheet);
         gtk_widget_show(sheet);
   
         // if we don't have any data, set the table uneditable, otherwise
         // set the data units row (0) uneditable
         range.col0 = range.row0 = range.rowi = 0; 
         if (_block->get_nrows() == 0 || _block->get_dispcols() == 0)
         {
            range.rowi = gtk_sheet_get_rows_count(GTK_SHEET(sheet)) -1;  
            range.coli = gtk_sheet_get_columns_count(GTK_SHEET(sheet)) -1; 
         }
         else if ((range.coli = cols) > 0)  range.coli -= 1; 
   
         gtk_sheet_row_button_add_label(GTK_SHEET(sheet), 0, "units");
         gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 0);

         // load first page of data
         if (row == 1)         load_page();
            else if (row > 0)  load_row(row);
     
         gtk_sheet_thaw(GTK_SHEET(sheet));
      }
      else
      {
         // cols array empty- 
         //create a new sheet to replace the existing one
         sheet = gtk_sheet_new(10, 10, "");
         gtk_sheet_set_selection_mode(GTK_SHEET(sheet), GTK_SELECTION_NONE); 
         gtk_container_add(GTK_CONTAINER(base_w), sheet);
         gtk_widget_show(sheet);
      }
      _sid1 = g_signal_connect(GTK_OBJECT(sheet), "set_cell",
                       (GtkSignalFunc) edit_cell_callback, this);
      _sid2 = g_signal_connect(GTK_OBJECT(sheet), "activate",
                       (GtkSignalFunc) enter_cell_callback, this);

      // track when a row/column is selected/deselected 
      g_signal_connect(GTK_OBJECT(sheet), "select-row",
                       (GtkSignalFunc)row_select_callback, this);
      g_signal_connect(GTK_OBJECT(sheet), "select-column",
                       (GtkSignalFunc)column_select_callback, this);
      g_signal_connect(GTK_OBJECT(sheet), "float-menu",
                       (GtkSignalFunc)floating_menu_callback, this);
      g_signal_connect(GTK_OBJECT(sheet), "changed",
                       (GtkSignalFunc)change_cell_callback, this);
      g_signal_connect(GTK_OBJECT(sheet), "traverse",
                       (GtkSignalFunc)traverse_callback, this);

      // clear last entered cell info from banner 
      _edstring = "";  
      _oldcellval = "";
      set_page_label();
   }
   return 0; 
}

int PrismDataTable::show()
{
   gtk_widget_show_all(GTK_WIDGET(base_w));
   return 0;
}


int PrismDataTable::change_cell_callback(GtkWidget *widget, gint row, 
   gint column, gpointer data)
{
   PrismDataTable* datTable = (PrismDataTable*)data; 

   if (datTable && datTable->is_loading())  return 1; 
   if (row > 0 && column > -1)
   {
      GtkSheet *sheet=GTK_SHEET(widget);
      const gchar* label = gtk_entry_get_text(GTK_ENTRY(sheet->sheet_entry)); 
      if (datTable && label != datTable->_oldcellval) 
      {
         string oval = datTable->_edstring; 
         if (datTable->validate_cell(column, row, label)) 
         { 
	    datTable->set_valid_edit(0);
	    return 1; 
	 }
      }
   }
   datTable->set_valid_edit(1);
   return 1;
}

int PrismDataTable::idle_callback(gpointer data)
{
   PrismDataTable* datTable = (PrismDataTable*)data;
   if (datTable)  return datTable->load_data_row();
   return 0; 
}


int PrismDataTable::edit_cell_callback(GtkWidget *widget, gint row, 
   gint column, gpointer data)
{
   GtkSheet *sheet;
   const gchar* label;
   GtkJustification justification;
   GtkSheetCellAttr attr;
   PrismDataTable* datTable = (PrismDataTable*)data; 
   sheet=GTK_SHEET(widget);

   if (datTable && datTable->is_loading())  return 1; 

   gtk_sheet_get_attributes(sheet, sheet->active_cell.row,
			           sheet->active_cell.col,
                                   &attr); 

   justification = attr.justification; 
   label = gtk_entry_get_text(GTK_ENTRY(sheet->sheet_entry)); 
   gtk_sheet_set_cell(sheet, sheet->active_cell.row,
  			     sheet->active_cell.col,
                             justification, label); 
   if (datTable && label != datTable->_oldcellval) 
   {
      string oval = datTable->_edstring; 
      datTable->_edstring = "";
      datTable->set_page_label();
      datTable->edit_cell(sheet->active_cell.col, sheet->active_cell.row, 
          label);
   }
   return 1; 
}

int PrismDataTable::traverse_callback(GtkWidget *widget, gint row,gint col,
   gint *new_row, gint *new_col, gpointer data)
{
   // this callback will force the sheet to stay in the cell of an edit 
   // until a valid edit is made (returning 0 prevents the cell exit) 
   PrismDataTable* datTable = (PrismDataTable*)data; 
   if (datTable && datTable->is_loading())  return 1; 
   if (datTable)   return datTable->traverse_cell(col, row);
   return 1; 
}


int PrismDataTable::enter_cell_callback(GtkWidget *widget, gint row, gint column, gpointer data)
{
   GtkSheet *sheet;
   const gchar* label;
   GtkJustification justification;
   GtkSheetCellAttr attr;
   PrismDataTable* datTable = (PrismDataTable*)data; 
   if (datTable && datTable->is_loading())  return 1; 
   sheet=GTK_SHEET(widget);
   gtk_sheet_get_attributes(sheet, sheet->active_cell.row,
			           sheet->active_cell.col,
                                   &attr); 
   justification = attr.justification; 
   label = gtk_entry_get_text(GTK_ENTRY(sheet->sheet_entry)); 

   char cell[256];
if (label)
   if(sheet->column[column].button.label && sheet->row[row].button.label)
      sprintf(cell,"  %s:%s = %s",sheet->column[column].button.label, 
              sheet->row[row].button.label, label); 
   else
      sprintf(cell,"  %d:%d = %s", column, row, label); 
   if (datTable && row && datTable->_valid_edit) 
   {
      datTable->_edstring = cell; 
      datTable->_oldcellval = label; 
      datTable->set_page_label();
   }
   return 1; 
}

int PrismDataTable::floating_menu_callback(GtkWidget *widget, gint column, gpointer data)
{
   PrismDataTable* datTable = (PrismDataTable*)data; 
   if (datTable)  datTable->launch_floating_menu(column); 
   return 0; 
}

int PrismDataTable::column_select_callback(GtkWidget *widget, gint column, gpointer data)
{
   PrismDataTable* datTable = (PrismDataTable*)data; 
   if (datTable)  datTable->select_column(column); 
   return 0; 
}

int PrismDataTable::row_select_callback(GtkWidget *widget, gint row, gpointer data)
{
   GtkSheet *sheet;
   PrismDataTable* datTable = (PrismDataTable*)data; 
   sheet=GTK_SHEET(widget);
   if (datTable)  datTable->select_row(row); 
   return 0;
}

int PrismDataTable::set_pagesize(long val)
{
   if (val < 1)   return 1;
   pagesize = val;
   numpages = rows / pagesize;
   if ( rows % pagesize) numpages++;
   if (curr_row > 0 && curr_row <= rows)
   {
      curr_page = curr_row / pagesize;
      if (curr_row % pagesize) curr_page++;
   }
   else   
   {
      curr_page = 1; 
      curr_row = 1;
   } 
   return 0; 
}

int PrismDataTable::load_row(long row)
{
   if ((row < 1) || (row > rows))  return 1;   // invalid row
   if (pagesize < 1)  return 2;                // invalid page size

   int rv = load_page(((row-1)/pagesize + 1));  
   if (!rv)   gtk_sheet_moveto(GTK_SHEET(sheet), (row-1)%pagesize+1, 0, 0.0, 0.0);
   return rv; 
}

int PrismDataTable::load_page(long page)
{
   long start_row, num_rows, rcnt;
   _loading = true; 
   if (page < 2)           rcnt = start_row = 1;
      else                 rcnt = start_row = (page - 1) * pagesize + 1; 
   if (page >= numpages) 
   {
      num_rows = (rows - (page -1) * pagesize);
      if (num_rows > rows)   num_rows = rows; 
   }
   else                 num_rows = _disp_rows -1;

   gtk_sheet_freeze(GTK_SHEET(sheet));

   // set up row labels
   curr_page = page;
   char sbuff[32];
   for (int rr = 1; rr <= num_rows; rr++)
   {
      sprintf(sbuff, "%1ld", rcnt++); 
      gtk_sheet_row_button_add_label(GTK_SHEET(sheet), rr, sbuff);
      gtk_sheet_row_set_visibility(GTK_SHEET(sheet), rr, 1);
   }

   if (_data)
   {
      delete[] _data;
      _data = NULL; 
   }
   if (_block->is_image())
   {
      if (cols == 0 || num_rows == 0)
      {
         gtk_sheet_thaw(GTK_SHEET(sheet));
         set_page_label();  // set label for NULL primary header 
         _loading = false; 
         return 1;
      }
      if ((_data = _block->get_image_data(start_row, num_rows)) != NULL)
      {
         int rr, cc; 
         for (cc=0; cc < cols; cc++) set_cell_bgcolor(0,cc);  // units row
	 _load_last_row = num_rows;
	 if (num_rows < PRISM_IDLE_LOAD_SZ)
            _load_curr_row = num_rows;
         else
            _load_curr_row = PRISM_IDLE_LOAD_SZ; 
         for (rr=0; rr < _load_curr_row; rr++)
            for (cc=0; cc < cols; cc++)
            {
               gtk_sheet_set_cell(GTK_SHEET(sheet),rr+1,cc,GTK_JUSTIFY_LEFT,
                                  _data[rr*cols +cc].c_str());
               set_cell_bgcolor(rr+1,cc);
            }
	 if (rr < _load_last_row)
         {
	    // add idle function to finish loading data- since gtk uses HIGH+20
	    // for redraw, we want a lower priority (lower num=higher priority)
	    if (_idletag != 0)
	    {
	       g_source_remove(_idletag); 
	       _idletag = 0; 
	    }
	    _idletag = g_idle_add_full(G_PRIORITY_HIGH_IDLE+50, &idle_callback, this,NULL); 
	 }
      }
   }
   else if (!_collist.empty())
   {
      PrismTableColumn *tcp; 
      long pos1, pos2, nc = 1;
      list<PrismTableColumn *>::iterator ii;

      for (ii = _collist.begin(); ii != _collist.end(); ii++)
         if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
         {
            pos1 = tcp->get_colmin(); 
            pos2 = tcp->get_colmax(); 
            if (pos1 != -1) 
            {
               _data = _block->get_data(table_to_data_column(pos1), 
                                        start_row, num_rows, nc);
               set_column_data_values(_data, pos1, pos2, num_rows);
	       delete[] _data; 
            } 
         }
      _data = NULL; 
   }
   curr_row = start_row;

   // hide rows beyond last
   if (num_rows != _disp_rows -1)
   {
      for (int rr = num_rows+1; rr < _disp_rows; rr++)
         gtk_sheet_row_set_visibility(GTK_SHEET(sheet), rr, 0);
   }
   gtk_sheet_thaw(GTK_SHEET(sheet));
   set_page_label(); 
   _loading = false; 
   return 0;
}

int PrismDataTable::next_page()
{
   if (curr_page == numpages)  return 1;   // on last page
   return load_page(++curr_page); 
}

int PrismDataTable::prev_page()
{
   if (curr_page == 1)  return 1;   // on first page
   return load_page(--curr_page); 
}

int PrismDataTable::first_page()
{
   return load_page(1);
}

int PrismDataTable::last_page()
{
   return load_page(numpages);
}

PrismDataColumn* PrismDataTable::get_selected_columns(long &cnt, bool override)
{
   long cc;
   PrismDataColumn* blkcols = _block->get_columns();
   PrismDataColumn* selcols = NULL;

   if (!blkcols)  return selcols; 

   if (_block->is_image())
   {
      selcols = new PrismDataColumn[cnt=1];
      selcols[0] = blkcols[0];
   }
   else if (_selcval && _selected_cols)
   {
      int ndx = 0, lastsel = _selcval + 1; 
      int* tmpndx = new int[cols];
      for (cc = 1; cc < lastsel; cc++)
      {
         for (int zz = 0; zz < cols; zz++)
	 {
	    if (_selected_cols[zz] == cc)
	    {
               tmpndx[ndx++] = zz;
	       break; 
	    }
	 }
      }
      if (ndx)
      {
         selcols = new PrismDataColumn[cnt=ndx];
         for (cc = 0; cc < ndx; cc++)
	    selcols[cc] = blkcols[table_to_data_column(tmpndx[cc])];
      }
      else 
      {
         // no columns selected- reset selected column index 
	 _selcval = 0;

         // on override- return all columns if no columns are selected
         if (override)  selcols = new PrismDataColumn[cnt=cols];
      }
      delete[] tmpndx;
   }
   else  if (override)
   {
      selcols = new PrismDataColumn[cnt=cols];
      for (cc = cnt; cc--; )   selcols[cc] = blkcols[table_to_data_column(cc)];
   }
   return selcols; 
}

PrismDataColumn* PrismDataTable::get_data_columns(long &cnt)
{
   PrismDataColumn* blkcols = NULL;
   if ((cnt = _block->get_totcols()) > 0)   blkcols = _block->get_columns(); 
   return blkcols; 
}


void PrismDataTable::set_page_label()
{
   if (!(_block && _parent))  return; 
   long r2, r1 = 1 + (curr_page -1) * pagesize;
   if (_load_curr_row != 0)  r2 = r1 + _load_curr_row -1; 
      else                   r2 = r1 + pagesize -1;
   if (r2 > rows) r2 = rows; 
   char buff[256];
   string lbl = _block->get_name();
   if (rows && numpages)
      sprintf(buff, "  %s\t\trows: (%ld-%ld)/%ld\t\tpage: %ld/%ld\t\t%s  ", lbl.c_str(), 
              r1 , r2, rows, curr_page, numpages, _edstring.c_str()); 
   else
      sprintf(buff, "  %s\t\trows: 0\t\t%s  ", lbl.c_str(), _edstring.c_str());
   _parent->set_frame_label(buff); 
}


void PrismDataTable::select_column(int column)
{
   if (_selected_cols && column <= cols && column > -1)
   {
      if (!_selected_cols[column]) 
         _selected_cols[column] = ++_selcval;
      else 
      {
         _selected_cols[column] = 0;

	 // check if we have any cols still selected 
	 bool found = false; 
	 for (int zz = cols; zz--; )
	    if (_selected_cols[zz]) found = true;  
	 if (!found)  _selcval = 0; 
      }
      set_column_bgcolor(column); 
      selected_columns_adjusted(); 
   }
}

void PrismDataTable::launch_floating_menu(int column)
{
   if (_floatmenu)  
   {
      gtk_widget_destroy(_floatmenu); 
      _floatmenu = NULL; 
   } 
   _expcol = NULL; 

   if ((_expcol = data_column_to_table_entry(column)) == NULL)  return;
   PrismDataColumn *cp = _expcol->get_data_column(); 
   if (!cp)  return; 
   if (!(cp->is_vector() || cp->is_vector_comp() || cp->is_array() || 
       cp->is_array_comp()))   return; 
   // launch new floating menu w/ appropriate option
   if ((_floatmenu = gtk_menu_new()) != NULL)
   {
      string lab, cname = cp->get_name(); 
      if (cp->is_vector_array_comp())
         lab = "Collapse Vector Array " + cp->get_vector_name();
      else if (cp->is_array_comp())
         lab = "Collapse Array " + cname;
      else if (cp->is_vector_comp())
         lab = "Collapse Vector " + cp->get_vector_name();
      else if (cp->is_vector() && cp->is_array())
         lab = "Expand Vector Array " + cname;
      else if (cp->is_vector())
         lab = "Expand Vector " + cname;
      else
         lab = "Expand Array " + cname;

      _fm_option = gtk_menu_item_new_with_label(lab.c_str());
      gtk_menu_shell_insert(GTK_MENU_SHELL(_floatmenu), _fm_option, 0);
      g_signal_connect(G_OBJECT(_fm_option), "activate",
                       G_CALLBACK(toggle_column_callback), this);
      gtk_widget_show_all(_floatmenu);
   }
   gtk_menu_popup(GTK_MENU(_floatmenu), NULL, NULL, NULL, NULL, 0,
                  gtk_get_current_event_time());
}

void PrismDataTable::select_row(int row)
{
   if (_selected_rows && row <= _selrpp && row > 0 && 
       curr_page > 0 && curr_page <= _selpps)
   {
      if (!_selected_rows[curr_page-1]) 
      {
         _selected_rows[curr_page-1] = new int[_selrpp];
         for (int zz= _selrpp; zz--; )
	   _selected_rows[curr_page-1][zz] = _selectMode;  
      }
      _selected_rows[curr_page-1][row-1] =! _selected_rows[curr_page-1][row-1];
      set_row_bgcolor(row);
   }
}


void PrismDataTable::set_column_bgcolor(int column)
{
   GtkSheetRange rng = { 0, column, _disp_rows, column};
   if (_selected_cols[column])
   {
      gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_select_bg);
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, &_select_fg);
   }
   else
   {
      if (_altrow)
      {
         GtkSheetRange r2 = { 0, column, _disp_rows, column};
         for (int rr = 0; rr < _disp_rows; rr++)
	 {
	    r2.row0 = r2.rowi = rr; 
	    if (rr % 2 == 1)
	       gtk_sheet_range_set_background(GTK_SHEET(sheet),&r2,&_oddrow_bg);
	    else 
	       gtk_sheet_range_set_background(GTK_SHEET(sheet),&r2,&_evenrow_bg);
         }
      }
      else
      {
         gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, NULL);
      }
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, NULL);
      if (_selected_rows[curr_page-1])  for (int rr = _selrpp; rr--; )
         if (_selected_rows[curr_page-1][rr] == 1)
	 {
	    // entries for selected rows need to remain selected 
	    rng.row0 = rng.rowi = rr+1;
            gtk_sheet_range_set_background(GTK_SHEET(sheet),&rng,&_select_bg);
            gtk_sheet_range_set_foreground(GTK_SHEET(sheet),&rng,&_select_fg);
	 }
   }
}

void PrismDataTable::set_row_bgcolor(int row)
{
   GtkSheetRange rng = { row, 0, row, cols-1};
   if (_selected_rows[curr_page-1] && _selected_rows[curr_page-1][row-1])
   {
      gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_select_bg);
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, &_select_fg);
   }
   else 
   {
      if (_altrow)
      {
         if (row%2==1) 
	    gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_oddrow_bg);
	 else 
	    gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_evenrow_bg);
      }
      else
      {
         gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, NULL);
      }
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, NULL);
      for (int cc = cols; cc--; ) 
         if (_selected_cols[cc])
	 {
	    // entries for selected columns need to remain selected 
	    rng.col0 = rng.coli = cc;
            gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_select_bg);
            gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, &_select_fg);
	 }
   }
}


void PrismDataTable::set_cell_bgcolor(int row, int col)
{
   GtkSheetRange rng = { row, col, row, col};
   // flag if we are in a selected column or selected row of current page 
   bool selected = (_selected_cols && _selected_cols[col]);
   if (row > 0 && _selected_rows && _selected_rows[curr_page-1] && 
       _selected_rows[curr_page-1][row-1])
      selected = 1; 
   
   if (selected)
   {
      gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &_select_bg);
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, &_select_fg);
   }
   else 
   {
      if (_altrow)
      {
         if (row%2 == 1)
	    gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng,&_oddrow_bg);
	 else 
	    gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng,&_evenrow_bg);
      }
      else
         gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, NULL);
      gtk_sheet_range_set_foreground(GTK_SHEET(sheet), &rng, NULL);
   }
}


void PrismDataTable::select_all_columns()
{
   _selcval = 0;
   if (_selected_cols)
      for (int cc = 0; cc < cols; cc++)
      {
         _selected_cols[cc] = ++_selcval;
         set_column_bgcolor(cc); 
      }
   selected_columns_adjusted(); 
}

void PrismDataTable::deselect_all_columns()
{
   _selcval = 0;
   if (_selected_cols)
      for (int cc = 0; cc < cols; cc++)
      {
         _selected_cols[cc] = 0; 
         set_column_bgcolor(cc);
      }
   selected_columns_adjusted(); 
}

void PrismDataTable::select_all_rows()
{
   if (_selected_rows && _selrpp > 0)
   {
      int pp, rr;
      for (pp = 0; pp < _selpps; pp++)
      {
         if (!_selected_rows[pp]) _selected_rows[pp] = new int[_selrpp];
	 // select each row in the current page 
         if (_selected_rows[pp])
	   for (rr = _selrpp; rr--; )   _selected_rows[pp][rr] = 1;
      }
      // highlight the rows of the current page 
      for (rr = 1; rr <= _selrpp; rr++)   set_row_bgcolor(rr);
   }
}

void PrismDataTable::deselect_all_rows()
{
   if (_selected_rows)
   {
      int pp, rr; 
      // reset flags for selected rows 
      for (pp = 0; pp < _selpps; pp++)
         if (_selected_rows[pp])
	    for (rr = _selrpp; rr--; )   _selected_rows[pp][rr] = 0;
      // clear the visible selected rows (skip units row) 
      for (rr = 1; rr <= _selrpp; rr++)   set_row_bgcolor(rr);
   }
}

int PrismDataTable::validate_cell(int col, int row, string value)
{
   if (!_block) return 1; 
   long tabrow = (curr_page - 1) * pagesize + row;
   return _block->validate_value(table_to_data_column(col), tabrow, value); 
}


int PrismDataTable::edit_cell(int col, int row, string value)
{
   if (!_block) return 1; 
   if (row < 1) return 2; 
   long tabrow = (curr_page - 1) * pagesize + row;
   int rval = 1; 
   PrismTableColumn* tcp;
   if (!(tcp = data_column_to_table_entry(col))) return 1; 
   PrismDataColumn* pdc = tcp->get_data_column();
   if (pdc) 
   {
      long apos = col - tcp->get_colmin(); 
      if (apos < -1)  apos = -1; 
      rval = _block->set_value(pdc->get_pos(), tabrow, value, apos); 
   } 
   if (rval == 0)
   {
      theApp->status_message("Changed cell from " + _oldcellval + " to " + value);
      theApp->set_modified();
   }
   else 
   {
      theApp->status_message("Failed to change cell from " + _oldcellval + " to " + value); 
   }
   return rval; 
}

int PrismDataTable::traverse_cell(int col, int row)
{
   GtkSheetRange rng = { row, col, row, col};

   if (row < 1 || col < 0)  return 1;  // units row is not editable

   if (_valid_edit)
      set_cell_bgcolor(row, col); 
   else
   {
      GdkColor color;
      gdk_color_parse("red", &color);
      gdk_color_alloc(gtk_widget_get_colormap(sheet), &color);
      gtk_sheet_range_set_background(GTK_SHEET(sheet), &rng, &color);
      if (_block) 
      {
         string msg = _block->get_errmsg() + "    (previous value=" + 
	    _oldcellval + ")";
	 theApp->status_message(msg,1);
      }
   }
   return _valid_edit;    
}


int PrismDataTable::append_row(long copyrow)
{
   if (!_block)  return 1; 
   int rv = _block->append_row(copyrow);
   if (!rv)
   {
      string bname = _block->get_name();
      char buff[256];
      // force reload to get the extension table to reflect new row count
      theApp->reload_current_tab(_block->get_extnum(), ++rows);
      if (copyrow)  
	 sprintf(buff,"Appended a copy of row %ld as row %ld of block %s",
	         copyrow, rows, bname.c_str()); 
      else
	 sprintf(buff,"Appended a zero'd out row as row %ld of block %s",
	         rows, bname.c_str()); 
      theApp->status_message(buff); 
   }
   theApp->set_modified();
   return 0; 
}

int PrismDataTable::append_column(PrismColumnInfra_p col_info)
{
   if (!col_info)  return 1;
   if (!_block)    return 2;  // no block 

   int rv = _block->append_column(col_info); 

   if (!rv)
   {
      string bname = _block->get_name();
      char buff[256];
      // force reload to get the extension table to reflect new column
      theApp->reload_current_tab(_block->get_extnum());
      sprintf(buff,"Appended a new column %s to block %s",
              col_info->name.c_str(), bname.c_str()); 
      theApp->set_modified();
   } 
   return rv; 
}



int PrismDataTable::get_extnum()
{
   if (_block)   return _block->get_extnum(); 
   return 0; 
}

int PrismDataTable::get_column_names(string* &colnames)
{
   if (!_block)  return 0;
   if (_block->get_type() == dmIMAGE)  return 0;
   return _block->get_column_names(colnames); 
}

void PrismDataTable::enable_editing()
{
   GtkSheetRange range = { 1,0,pagesize,cols-1}; 
   range.rowi = gtk_sheet_get_rows_count(GTK_SHEET(sheet)); 
   range.coli = gtk_sheet_get_columns_count(GTK_SHEET(sheet)); 

   if (!_block->is_image())
   {
      // set all rows except first as editable for all cols  
      gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 1);

      // disable edits in collapsed vectors/arrays
      PrismTableColumn *tcp;
      PrismDataColumn* cp;
      list<PrismTableColumn *>::iterator ii;
      for (ii = _collist.begin(); ii != _collist.end(); ii++)
         if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
	 {
            cp = tcp->get_data_column();
	    if ((cp->is_vector() && !cp->is_vector_comp()) || 
	        (cp->is_array() && !cp->is_array_comp()))
	    {
	       range.col0 = range.coli = tcp->get_colmin();
	       if (range.col0 > -1 && range.coli > -1)
                  gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 0);
            }
         }
   }
   else  if (_block->get_nrows() && _block->get_dispcols())
   {
      // we only set the sheet editable if the image has rows and cols
      gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 1);
   }
   _editable = true;
}

void PrismDataTable::disable_editing()
{
   GtkSheetRange range = { 0,0,pagesize,cols-1}; 
   gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 0);
   _editable = false;
}

void PrismDataTable::toggle_column(PrismTableColumn* tp)
{
   PrismTableColumn *rtcp, *tcp;
   
   if (tp)  tcp = tp;
      else  tcp = _expcol; 
   if (!tcp)  return;
   if ((rtcp = find_root_column(tcp)) == NULL)  return;

   bool done = false, vis = (tcp == rtcp); 
   long c1, c2, c3, c4;
   int id, zz; 
   gtk_sheet_freeze(GTK_SHEET(sheet));
   if (!rtcp->get_togminmax(c1, c2))
   {
      // handle child nodes
      list<PrismTableColumn *>::iterator ii = _collist.begin(); 
      while (ii != _collist.end() && !done)
      {
         id = (*ii)->get_id();
         if (id >= c1 && id <= c2)
         {
	    (*ii)->get_colminmax(c3, c4);
	    if (c3 == -1)
	    {
	       // request to expand non-inlined data
	       expand_oversized_column((*ii));
	       (*ii)->get_colminmax(c3, c4);
	    }
	    if (c3 > -1)   for (zz = c3; zz <= c4; zz++)
	       gtk_sheet_column_set_visibility(GTK_SHEET(sheet), zz, vis);
	 }
	 done = (id > c2);
         ii++;
      }
      // handle root node 
      rtcp->get_colminmax(c3, c4);
      gtk_sheet_column_set_visibility(GTK_SHEET(sheet), c3, !vis); 
   }
   gtk_sheet_thaw(GTK_SHEET(sheet));
}


void PrismDataTable::toggle_column_callback(GObject *object, gpointer user_data)
{
   PrismDataTable* tab = (PrismDataTable*)user_data;
   if (tab)  tab->toggle_column();
}


void PrismDataTable::expand_oversized_column(PrismTableColumn* col)
{
   if (col && sheet)
   {
      PrismDataColumn *pdc = col->get_data_column(); 
      if (pdc && pdc->is_array() || pdc->is_vector())
      {
         int sz = pdc->get_array_size(); 
	 int dm = pdc->get_dim(); 
	 long ncls = gtk_sheet_get_columns_count(GTK_SHEET(sheet));
         gtk_sheet_add_column(GTK_SHEET(sheet), sz*dm); 
	 string nm = pdc->get_name(); 
         GtkSheetRange range = { 1,0,pagesize,cols-1}; 
	 char buff[32];

         // we if added rows we need to resize the selected cols array 
         int zz, ccnt = cols + sz*dm;
	 int* tcols = new int[ccnt];
         if (_selected_cols && tcols)
	 {
	    for (zz = cols; zz--; )           tcols[zz] = _selected_cols[zz];
	    for (zz = cols; zz < ccnt; zz++)  tcols[zz] = _selectMode;  
	    delete[] _selected_cols;
	    _selected_cols = tcols;
	    cols = ccnt; 
         }

         for (int cc = 0; cc < sz; cc++)
	 {
	    sprintf(buff, "%s[%d])", nm.c_str(), cc);
	    gtk_sheet_column_button_add_label(GTK_SHEET(sheet), cc + ncls, buff);
	    set_cell_bgcolor(0,cc+ncls);  // units row 
	 }

         col->set_colminmax((range.col0 = ncls), (range.coli = ncls+sz*dm-1)); 

	 if (range.col0 > -1 && range.coli > -1)
	 {
            // set editability of new column
            gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, _editable);
	    // set editability of units column 
	    range.row0 = range.rowi = 0;
	    gtk_sheet_range_set_editable(GTK_SHEET(sheet), &range, 0); 
         }  

	 // for now, use load_page and reload all the page data- to improve
	 // perfomance, change this to only load the needed rows/cols
         load_page(curr_page); 
      }
   }
}


int PrismDataTable::setup_table_columns()
{
   if (!_block)  return 0; 
   _collist.clear();
   cols = _block->get_totcols(); 
   PrismDataColumn* columns = _block->get_columns(); 
   PrismTableColumn *ptc, *ptc2;
   long tmn, tmx, adm, vsz, asz, colcnt = 0; 

   if (_block->is_image())
   {
      tmn = _block->get_dispcols();
      ptc = new PrismTableColumn(&columns[0], 0, tmn-1);
      ptc->is_image(true); 
      _collist.push_back(ptc); 
      return tmn;
   }

   for (int rr = 0; rr < cols; rr++)
   {
      asz = columns[rr].get_array_size(); 
      adm = columns[rr].get_adim(); 
      vsz = columns[rr].get_dim();

      if (vsz == 1)  // scalar  
      {
         if (adm == 0 && asz == 0) // scalar 
         {
            ptc = new PrismTableColumn(&columns[rr], colcnt, colcnt); 
            _collist.push_back(ptc); 
            colcnt++;
         }
         else  // array (1 or MD)  
	 {
	    int pos1, pos2, sts; 
	    // increment rr count since we're dealing with array value here
	    ptc = new PrismTableColumn(&columns[rr++], colcnt);
            _collist.push_back(ptc); 
	    colcnt++;

	    if (asz < PRISM_MAX_EXP_SZ)
	    {
	       pos1 = colcnt; 
	       colcnt = colcnt + asz - 1;
	       pos2 = colcnt; 
	       ptc->set_state(sts = PRISM_COLSTATE_PRELOAD); 
	       colcnt++;
            }
	    else 
	    {
	       pos1 = pos2 = -1;
	       ptc->set_state(sts = PRISM_COLSTATE_SKIP); 
	       // if skipping- don't increment colcnt
            }
	    ptc2 = new PrismTableColumn(&columns[rr], pos1, pos2, sts);
	    tmn = ptc->get_id(); 
	    tmx = ptc2->get_id(); 
	    ptc->set_togminmax(tmx, tmx); 
	    ptc2->set_togminmax(tmn, tmn); 
            _collist.push_back(ptc2); 
	 }    
      }
      else 
      {
         if (adm == 0 && asz == 0)  // simple vector
         {
	    ptc = new PrismTableColumn(&columns[rr], colcnt, colcnt);
	    _collist.push_back(ptc);
	    tmn = ptc->get_id(); 
	    ptc->set_togminmax(tmn+1, tmn+vsz); 
	    colcnt++;
	    for (int zz = 1; zz <= vsz; zz++)
            {
	       rr++; 
	       ptc2 = new PrismTableColumn(&columns[rr], colcnt, colcnt);
	       _collist.push_back(ptc2);
	       ptc2->set_togminmax(tmn, tmn); 
	       colcnt++; 
            }
         }
         else 
	 {
	    int pos1, pos2; 
	    ptc = new PrismTableColumn(&columns[rr], colcnt);
	    tmn = ptc->get_id(); 
	    ptc->set_togminmax(tmn+1, tmn+vsz); 
            _collist.push_back(ptc); 
	    colcnt++;
	    for (int zz = 1; zz <= vsz; zz++)
            {
	       rr++; 
	       pos1 = colcnt;
	       pos2 = colcnt+asz -1; 
	       ptc2 = new PrismTableColumn(&columns[rr], pos1, pos2);
	       _collist.push_back(ptc2);
	       ptc2->set_togminmax(tmn, tmn); 
	       colcnt = pos2 +1; 
            }            
	 }    
      }
   }
   return colcnt; 
}



int PrismDataTable::set_column_data_values(string* data, long col1, long col2, 
   long rows)
{    
   if (data)
   {
      long zz, rr, ncols,  endc = col2 + 1;
      ncols = endc - col1; 

      for (zz = col1; zz < endc; zz++)  set_cell_bgcolor(0,zz);  // units row
      for (rr = 0; rr < rows; rr++)
      {
	 for (zz = col1; zz < endc; zz++)
	 {
            gtk_sheet_set_cell(GTK_SHEET(sheet),rr+1,zz,GTK_JUSTIFY_LEFT,
            data[rr*ncols + zz-col1].c_str());
	    set_cell_bgcolor(rr+1,zz);
	 }
      }
   }
   return 0; 
}

int PrismDataTable::table_to_data_column(int col)
{
   list<PrismTableColumn *>::iterator ii;
   PrismTableColumn *tcp;
   PrismDataColumn* cp;

   for (ii = _collist.begin(); ii != _collist.end(); ii++)
   {
      if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
         if (col >= tcp->get_colmin() && col <= tcp->get_colmax())
	    if ((cp = tcp->get_data_column()) != NULL)
               return cp->get_pos();
   }
   return -1; 
}

PrismTableColumn* PrismDataTable::data_column_to_table_entry(int col)
{
   list<PrismTableColumn *>::iterator ii;
   PrismTableColumn *tcp;

   for (ii = _collist.begin(); ii != _collist.end(); ii++)
   {
      if ((tcp = (PrismTableColumn*)(*ii)) != NULL)
         if (col >= tcp->get_colmin() && col <= tcp->get_colmax())
            return tcp;
   }
   return NULL;
}


PrismTableColumn* PrismDataTable::find_root_column(PrismTableColumn* tcp)
{
   list<PrismTableColumn *>::iterator ii;
   PrismTableColumn *rtp;
   bool done = false; 

   // if togmin is -1, the table colum,n is a scalar (not expandable) 
   if (tcp->get_togmin() == -1) return NULL; 
   for (ii = _collist.begin(); ii != _collist.end() && !done; ii++)
   {
      if (tcp == (PrismTableColumn*)(*ii))
      {
         long id, c1, c2;
         if (!tcp->get_togminmax(c1, c2))
         {
	    while (!done)
            {
               rtp = (PrismTableColumn*)(*ii); 
               id = rtp->get_id();
	       if (id <= c1 || (ii == _collist.begin()))   done = true;
                  else   ii--; 
            } 
            if (id >= c1 && id <= c2)   return rtp; // found root node
         }
      } 
   }
   // check if we are the root
   if (tcp->get_togmin() > tcp->get_id())  return tcp;  
   // didn't find the root- return NULL 
   return NULL;
}


void PrismDataTable::selected_columns_adjusted()
{
   if (_parent)   _parent->set_toolbar_sensivity(PRISM_DATA_INFO_BTN, _selcval);
}


gboolean PrismDataTable::load_data_row()
{
   if (_load_curr_row >= _load_last_row || cols == 0 || !_data)
   {
      // finished loading- return false to exit idle callback 
      _load_last_row = _load_curr_row = 0;
      return false; 
   }
   // freeze sheet to refresh after the row has been populated
   gtk_sheet_freeze(GTK_SHEET(sheet));
long zend, zz;
if (_load_last_row-_load_curr_row < PRISM_IDLE_LOAD_SZ)  zend = _load_curr_row+1;
   else   zend = _load_curr_row + PRISM_IDLE_LOAD_SZ;
for (zz= _load_curr_row; zz < zend; zz++)
   for (int cc = 0; cc < cols; cc++)
   {
      gtk_sheet_set_cell(GTK_SHEET(sheet), zz+1, cc,
         GTK_JUSTIFY_LEFT, _data[zz*cols +cc].c_str());
      set_cell_bgcolor(zz+1,cc);
   }
   gtk_sheet_thaw(GTK_SHEET(sheet));
//   _load_curr_row++;
   _load_curr_row = zend; 
   set_page_label(); 
   // return true to keep idle function active 
   return true; 
}

string PrismDataTable::get_selected_imgcol_spec()
{
   static string cfiltspec = "";
   long pp, rr, selrow = 0, selcol = 0, rowcnt = 0, colcnt = 0;

   // only build spec for images
   if (!_block)  return cfiltspec; 
   if (!_block->is_image())  return cfiltspec; 

   // check for selected column
   for (pp = 0; pp < cols; pp++)
      if (_selected_cols[pp])
      {
         if (colcnt++ == 0)  selcol = pp + 1;
      }

   // check for selected row
   if (_selected_rows && _selrpp > 0)
      for (pp = 0; pp < _selpps; pp++)
         if (_selected_rows[pp])
            for (rr = 0; rr < _selrpp; rr++)
               if (_selected_rows[pp][rr])
                  if (rowcnt++ == 0)  selrow = _selrpp * pp + rr + 1;

   if (rowcnt + colcnt == 1)
   {
      char buff[256];
      long sz = 191;
      if (rowcnt) 
      {
         sz = _block->get_dispcols(); 
         sprintf(buff,"[#1=1:%d,#2=%d]", sz, selrow);
      }
      if (colcnt)
      {
         sz = _block->get_nrows(); 
         sprintf(buff,"[#1=%d,#2=1:%d]", selcol, sz);
      }
      cfiltspec = buff; 
   }
   return cfiltspec;
}

string PrismDataTable::get_selected_rowspec()
{
   static string filtspec;
   ostringstream allrows, seq; 
   string rowfilt;
   filtspec = rowfilt = "";
   long inseq = 0, prev = -99, cur;
   if (_selected_rows && _selrpp > 0)
   {
      long pp, rr;
      for (pp = 0; pp < _selpps; pp++)
      {
         if (_selected_rows[pp])
         {
            for (rr = 0; rr < _selrpp; rr++)
            {
               if (_selected_rows[pp][rr])
               {
                  if (!inseq) 
                  {
                     if (prev != -99) seq << ",";
                     prev = _selrpp * pp + rr + 1;
                     seq << prev;
                     inseq = 1;
                  }
               }
               else if (inseq) 
               {
                  cur = _selrpp * pp + rr ;
                  if (cur != prev)
                  {
                     seq << ":" << cur;
                  }
                  inseq = 0;
               }
            }
         }
         else if (inseq)
         {
            // last row of previous page was in a sequence
            cur = _selrpp * (pp -1) + rr ;
            if (cur != prev)
            {
               seq << ":" << cur;
            }
            inseq = 0;
         }
      }
      if (inseq)
      {
         // close final sequence
         cur = rows;
         if (cur != prev)   seq << ":" << cur;
      }
      rowfilt = seq.str();
      // ignore case where all rows are selected
      if (rowfilt == "1" && rows == 1)
         rowfilt = "";
      else 
      {
         allrows << "1:" << rows;
         if (rowfilt == allrows.str())   rowfilt = "";  
      }
   }
   // if some rows are selected, built virtual filter syntax 
   if (rowfilt.length())  filtspec="[#row=" + rowfilt + "]"; 
   return filtspec; 
}

 

