
// 
//  Copyright (C) 2007-2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismApp.hh"
#include "PrismFileFrame.hh"
#include "PrismDataArea.hh"
#include "PrismDataSet.hh"
#include "PrismHeaderArea.hh"
#include "PrismExtnTable.hh"

#include <iostream>

using namespace std;

PrismFileFrame::PrismFileFrame(string filename):PrismBaseObj()
{
   _loaded = false;  
   _editlock = true;
   _currmode = 0; 
   base_w = gtk_frame_new(NULL);

   vpane = gtk_vpaned_new ();
   gtk_container_add (GTK_CONTAINER (base_w), vpane);

   hpane = gtk_hpaned_new ();
   gtk_paned_pack1(GTK_PANED(vpane), hpane, TRUE, TRUE);

   extnTable = new PrismExtnTable(this); 
   gtk_paned_pack1(GTK_PANED(hpane), extnTable->base_widget(), TRUE, TRUE);

   headerArea = new PrismHeaderArea("Header Keywords");
   gtk_paned_pack2(GTK_PANED(hpane), headerArea->base_widget(), TRUE, TRUE);

   dataArea = new PrismDataArea("Extension Data");
   gtk_paned_pack2(GTK_PANED(vpane), dataArea->base_widget(), TRUE, TRUE);

   locked = lockro = edit = deficon = edited = NULL; 
   tabdisplay = gtk_hbox_new(false, 0);
   tablab = gtk_label_new("");
   tabbtn = gtk_button_new_from_stock(GTK_STOCK_OK);
   g_signal_connect(tabbtn, "clicked", G_CALLBACK(lock_button_callback),
	          (void*)this);
   gtk_button_set_label(GTK_BUTTON(tabbtn),"");
   GtkTooltips* tips = theApp->get_tips_group();
   if (tips) 
      gtk_tooltips_set_tip(tips,tabbtn,"locks or unlocks file for editing",NULL); 
   iconfile_path ="";
   char *pth = getenv("ASCDS_INSTALL");
   if (pth)
   {
      iconfile_path = pth;
      iconfile_path.append("/doc/txt/");
   }
   string iconfile = iconfile_path; 
   iconfile.append("prism_deficon.png");
   deficon = gtk_image_new_from_file(iconfile.c_str());
   gtk_image_set_pixel_size(GTK_IMAGE(deficon), 1);
   gtk_button_set_image(GTK_BUTTON(tabbtn), deficon);
   gtk_box_pack_start(GTK_BOX(tabdisplay), tabbtn, FALSE, FALSE, 1);
   gtk_box_pack_start(GTK_BOX(tabdisplay), tablab, FALSE, FALSE, 1);
   g_object_set (gtk_settings_get_default (), "gtk-button-images", TRUE, NULL);
   gtk_widget_show_all(tabdisplay);

   // load data file, if specified 
   _data = new PrismDataSet(filename);
   if (!_data->get_status()) 
   {
      string msg = "Loaded file ";
      msg.append(filename); 

      // set extension data
      list<ExtensionInfo_t> exts;
      _data->get_extensions(exts);
      extnTable->load_entries(exts, _data->get_current_block()); 

      theApp->status_message(msg);
      _loaded = true; 
      gtk_frame_set_label(GTK_FRAME(base_w), filename.c_str());
//      gtk_label_set_text(GTK_LABEL(tablab), filename.c_str());
      set_tab_icon(filename); 
   }
   else 
   {
      set_lock_sensitivity();    // still lock if we don't have a file
      if (filename != "") 
      {
         if (!access(filename.c_str(), F_OK))
            theApp->status_message("Unable to open " + filename + ". It does not appear to be a file. Displaying a blank tab",1); 
         else 
            theApp->status_message("Warning: Unable to open " + filename + ". Displaying a blank tab", 1); 
      } 
   } 
}

PrismFileFrame::~PrismFileFrame()
{
   if (dataArea)
   {
      delete dataArea;
      dataArea = NULL;
   }
   if (headerArea)
   {
      delete headerArea;
      headerArea = NULL;
   }
   if (extnTable)
   {
      delete extnTable;
      extnTable = NULL; 
   }
   if (_data)
   {
      delete _data;
      _data = NULL;
   }
}

int PrismFileFrame::load_file(string filename) 
{
   PrismDataSet *tmpData = new PrismDataSet(filename);
   if (!tmpData->get_status()) 
   {
      if (_loaded)  delete _data;
      _data = tmpData; 
      
      string msg = "Loaded file ";
      msg.append(filename); 

      // set extension data
      list<ExtensionInfo_t> exts;
      _data->get_extensions(exts);
      extnTable->load_entries(exts, _data->get_current_block()); 

      theApp->status_message(msg);
      _loaded = true; 
      set_tab_icon(filename); 
   }
   else
   {
      string msg = "Unable to load file ";
      msg.append(filename); 
      theApp->status_message(msg, 1);
      _loaded = false; 
      return 1; 
   }
   return 0; 
}

int PrismFileFrame::open_extension(string ename)
{
   if (_data->open_extension(ename) == 0)
   {
      // update header contents 

      // update data table contents 
      dataArea->set_data(_data); 
      headerArea->set_data(_data); 
      set_lock_sensitivity(); 
   }

   return 0; 
}

int PrismFileFrame::open_extension(int num, long row)
{
   if (_data->open_extension(num) == 0)
   {
      // set extension data
      _data->browse_extensions();
      list<ExtensionInfo_t> exts;
      _data->get_extensions(exts);
      extnTable->load_entries(exts, _data->get_current_block()); 

      // update data table contents 
      dataArea->set_data(_data, row); 

      // update header contents 
      headerArea->set_data(_data); 

      // adjust toolbar based on edit lock
      set_lock_sensitivity(); 
   }
   return 0; 
}

string PrismFileFrame::get_filename(int filter)
{
   static string fname;
   fname = "";
   if (_data)  fname = _data->get_filename();

   if (filter && dataArea && fname.length())
   {
      string::size_type p1, p2;
      string filt = dataArea->get_selected_cols();
      if (_data->get_filter_block() != _data->get_current_block())
      {
         // switched extensions- remove the original filter
         if ((p1 = fname.find("[")) != string::npos)
            fname = fname.substr(0, p1); 
         // append new bock filter
         char buff[32];
         sprintf(buff,"[%1d]", _data->get_current_block());
         fname.append(buff);
         if (filt.length()) fname.append(filt);
      }
      else
      {
         // within the same extension- keep the original filter spec
         if ((p1 = fname.find("[cols ")) != string::npos)
         {
            p2 = fname.find("]", p1);
            if (p2 != string::npos)
            {
               // replace existing column filter 
               fname.replace(p1, p2-p1+1, filt); 
            }
         } 
         else if (filt.length()) fname.append(filt); 
      }
   }
   return fname; 
}

int PrismFileFrame::show()
{
   headerArea->show();
   dataArea->show();
   gtk_widget_show_all(hpane);
   gtk_widget_show_all(vpane);
   gtk_widget_show_all(base_w);
   return 0;
}

void PrismFileFrame::set_frame_title(string val)
{
   frame_title = val;
   if (base_w)  gtk_frame_set_label(GTK_FRAME(base_w), frame_title.c_str());
}

int PrismFileFrame::visualize(int type, int dialog)
{
   if (!dataArea)   return 1; 
   int rv = 1;

   if (type == 0)
      rv = dataArea->generate_plot(dialog);
   else if (type == 1)
      rv = dataArea->generate_histogram(dialog);
   else if (type == 2)
      rv = dataArea->generate_image(dialog);
   return rv; 
}

int PrismFileFrame::select_entry(int val, int as_col)
{
   if (dataArea)  return  dataArea->set_selection(val, as_col);
   return 1; 
}

GtkWidget* PrismFileFrame::get_icon()
{
   return tabdisplay; 
}

void PrismFileFrame::set_tab_icon(string fname)
{
   if (tabbtn)
   {
      GtkWidget* im = NULL; 
      if (_data)   _currmode = _data->get_editmode();
      string iconfile = iconfile_path;
      switch(_currmode)
      {
         case -1:
	    iconfile.append("prism_lockro.png");
            lockro = gtk_image_new_from_file(iconfile.c_str());
            gtk_image_set_pixel_size(GTK_IMAGE(lockro), 1);
            _editlock = true;    
	    im = lockro;
         break; 
         case 2: 
	    iconfile.append("prism_edit.png");
            edit = gtk_image_new_from_file(iconfile.c_str());
            gtk_image_set_pixel_size(GTK_IMAGE(edit), 1);
            im = edit; 
	    _editlock = false;    
         break;
         case 3:
	    iconfile.append("prism_edited.png");
            edited = gtk_image_new_from_file(iconfile.c_str());
            gtk_image_set_pixel_size(GTK_IMAGE(edited), 1);
            im = edited; 
	    _editlock = false;    
         break; 
         case 1:
	    iconfile.append("prism_locked.png");
            locked = gtk_image_new_from_file(iconfile.c_str());
            gtk_image_set_pixel_size(GTK_IMAGE(locked), 1);
            _editlock = true; 
            im = locked; 
         break;
         default:
	    iconfile.append("prism_deficon.png");
            deficon = gtk_image_new_from_file(iconfile.c_str());
            gtk_image_set_pixel_size(GTK_IMAGE(deficon), 1);
            im = deficon; 
            _editlock = true; 
         break;
      }
      gtk_button_set_image(GTK_BUTTON(tabbtn),im);
      if (fname.length())
      {
	 string txt;
	 string::size_type fpos = fname.find_last_of("["); 
	 string::size_type pos = fname.find_last_of("/", fpos); 
	 if (pos == string::npos)  txt = fname;
	    else         txt = fname.substr(pos+1); 
         gtk_label_set_text(GTK_LABEL(tablab), txt.c_str());
      }
      set_lock_sensitivity(); 
   }
}

void PrismFileFrame::lock_button_callback(GObject *object, gpointer user_data)
{
   PrismFileFrame *frm = (PrismFileFrame*)user_data;
   if (frm)  frm->set_lock_mode();
}

void PrismFileFrame::set_lock_mode(int md)
{
   if (!_data) return; 
   if (_currmode == -1) 
   {
      theApp->status_message("Unable to set edit lock for read-only file", 1);
      return;
   }

   if (md == -1)
   {
      _data->set_editmode(); 
      set_tab_icon(""); 
   }
   else if (md == 0)   // lock request
   {
      if (_currmode == 2 || _currmode == 3)  
      {
         _data->set_editmode(); 
         set_tab_icon("");
      }
      else if (_currmode == 1)
         theApp->status_message("Request to lock an already locked file ignored.");
   }
   else if (md == 1)   // unlock request
   {
      if (_currmode == 1)  
      {
         _data->set_editmode(); 
         set_tab_icon("");
      }
      else if (_currmode == 2 || _currmode == 3)
         theApp->status_message("Request to unlock an already unlocked file ignored.");
   }
}

void PrismFileFrame::set_lock_sensitivity()
{
   if (_editlock)
   {
      // disable editing for file 
      dataArea->disable_editing();
      headerArea->disable_editing();
   }
   else
   {
      // enable editing 
      dataArea->enable_editing();
      headerArea->enable_editing();
   }
}

void PrismFileFrame::set_modified()
{
   if (_data)
   {
      _data->set_modified();
      set_tab_icon("");
   }

}


