// 
//  Copyright (C) 1998,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include	"PrismIPCMessage.hh"
#include <string>


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	PrismIPCMessage
//
// DESCRIPTION:	Constructor for the PrismIPCMessage class. The string passed in
//		will be split up with a SPACE character being the delimiter.
//
// EXAMPLE:	input string:	"This is string one"
//
//			param[0]:	"This"
//			param[1]:	"is"
//			param[2]:	"string"
//			param[3]:	"one"
//
//

PrismIPCMessage::PrismIPCMessage(char *paramlist)
{

        if(paramlist == NULL)
                return;

        //
        // Do the init thing.
        //

        Init();

	//
	// Set the arguments
	//

	SetParam(paramlist);
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	SetParam
//
// DESCRIPTION:	break up the string into space separated parameters. This will
//		clear out any other argument that may have been contained
//		in the message.
//

int
PrismIPCMessage::SetParam(char *paramlist)
{
char    *p1, *p2, *tmp;

        // Free up any previous arguments
	Clear();
 
        // Make a duplicate of the paramlist
        STRDUP(tmp, paramlist);

        p1 = tmp;

        // Break up the paramlist.
        for(nEntries=0; p1 && nEntries < nSlots-1; nEntries++)
        {
		if ((*p1 == '"') && (p2 = strchr(p1+1,'"')) != NULL) {
		   // quoted string, take everything till next quote;
		   // unbalanced quote falls through to else clause
		   int l = p2 - p1 + 1;
		   args[nEntries] = (char*)malloc(l+1);
		   strncpy(args[nEntries],p1,l);
		   args[nEntries][l]=0;
		}
		else {

		   // Locate the next SPACE
                   p2 = strchr(p1, ' ');

                   // If not more SPACES, then we're done
                   if(p2 == NULL)
                        break;

                   // Write in a NULL to terminate the string
                   *p2 = 0;

		   // duplicate the string
                   STRDUP(args[nEntries], p1);
		}

                // Jump past the space/quote
                p1 = p2+1;

		// jump past all blank spaces.
		while(p1 && *p1 && *p1 == ' ')
		  	p1++;
        }

        // Grab the last one.
	if (p1 && *p1) {
	   STRDUP(args[nEntries], p1);
	   nEntries++;
	}

        // This is currently a constant, but that could change
        nSlots = PrismIPC_PARAM_LIMIT;

	free(tmp);

	return 0;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	PrismIPCMessage
//
// DESCRIPTION:	Constructor for the PrismIPCMessage class
//
//
//

PrismIPCMessage::PrismIPCMessage()
{

	//
	// Do the init thing.
	//

	Init();
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	~PrismIPCMessage
//
// DESCRIPTION:	Destructor for the PrismIPCMessage class
//
//
//

PrismIPCMessage::~PrismIPCMessage()
{
	Clear();
	free(args);
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	Init
//
// DESCRIPTION:	Do basic startup initialization of variables and anything
//		that may need to be initialized to certain values
//
//

void
PrismIPCMessage::Init()
{

        nSlots = PrismIPC_PARAM_LIMIT;
	args = (char**)calloc(PrismIPC_PARAM_LIMIT,sizeof(char*));
	nEntries = 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	Clear
//
// DESCRIPTION:	Clear all parameters within the message
//
//

int
PrismIPCMessage::Clear()
{
int	i;

	for(i=0; i<nEntries; i++)
		free(args[i]);

	nEntries = 0;

	return 0;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddPAram
//
// DESCRIPTION:	Add a Strings parameter to the message
//
//
//

int
PrismIPCMessage::AddParam(string &s)
{

  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

  	//
  	// Add the new entry
  	//

	if (s.length() > 0)
	{
		STRDUP(args[nEntries], s.c_str());
		nEntries++;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddParam
//
// DESCRIPTION:	retrun a float to the requested parameter.
//
//

int	
PrismIPCMessage::AddParam(float f)
{
strstream param;

  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

	param << f << ends;

	STRDUP( args[nEntries], param.str() );
	nEntries++;

	return 0;

} 
  		
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddParam
//
// DESCRIPTION:	Add a float parameter to the message.
//
//

int    
PrismIPCMessage::AddParam(double d)
{

  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

	ostrstream param;
	param << d << ends;

	STRDUP( args[nEntries], param.str() );
	nEntries++;

	return 0;

} 
  		
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddParam
//
// DESCRIPTION:	Add an integer parameger to the message.
//
//

int	
PrismIPCMessage::AddParam(int i)
{
  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

	strstream param;
	param << i << ends;

	STRDUP( args[nEntries], param.str() );
	nEntries++;

	return 0;
} 
  		
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddParam
//
// DESCRIPTION:	Add a long parameter to the message.
//
//

int    
PrismIPCMessage::AddParam(long l)
{
  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

	ostrstream param;
	param << l << ends;

	STRDUP( args[nEntries], param.str() );
	nEntries++;

	return 0;
} 
  		
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	AddParam
//
// DESCRIPTION:	Add a char* parameter to the message.
//
//

int	
PrismIPCMessage::AddParam(char *c)
{
  	//
  	// Can we add this parameter ??
  	//

  	if(nEntries >= nSlots)
	  	return -1;

	STRDUP(args[nEntries], c);
	nEntries++;

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetParam
//
// DESCRIPTION:	retrun a Strings class to the requested parameter.
//
//

int
PrismIPCMessage::GetParam(int index, string &s)
{

  	//
  	// make sure the index is within range
  	//

  	if(index >= nEntries && index > -1)
	  	return -1;

	//
	// Create the Strings class
	//

	s = args[index];

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetParam
//
// DESCRIPTION:	retrun a float to the requested parameter.
//
//

int	
PrismIPCMessage::GetParam(int index, float &f)
{
  	//
  	// make sure the index is within range
  	//

  	if(index >= nEntries && index > -1)
	  	return -1;

	f = atof(args[index]);

	return 0;
} 
	
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetParam
//
// DESCRIPTION:	retrun a double to the requested parameter.
//
//

int
PrismIPCMessage::GetParam(int index, double &d)
{

  	//
  	// make sure the index is within range
  	//

  	if(index >= nEntries && index > -1)
	  	return -1;

  	d = strtod(args[index], NULL);

	return 0;
} 
	
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetParam
//
// DESCRIPTION:	retrun an integer to the requested parameter.
//
//

int
PrismIPCMessage::GetParam(int index, int &i)
{

  	//
  	// make sure the index is within range
  	//

  	if(index >= nEntries && index > -1)
	  	return -1;

  	i = atoi(args[index]);

	return 0;
} 
  		
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetParam
//
// DESCRIPTION:	retrun a long to the requested parameter.
//
//

int
PrismIPCMessage::GetParam(int index, long &l)
{

  	//
  	// make sure the index is within range
  	//

  	if(index >= nEntries && index > -1)
	  	return -1;

	l = atol(args[index]);

	return 0;
} 


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetMessage
//
// DESCRIPTION:	concatenate all parameters together into a NULL terminated
//		string
//
//

char*
PrismIPCMessage::GetMessage(int index)
{
char	*mess;
int	i, length;

        if(index >= nEntries)
                return NULL;

	length = GetLength(index);

	mess = new char[length];

	if(mess == 0)
	  	return 0;

	mess[0] = 0;

	//
	// Concatenate them. Do the first argument first
	//

	strcat(mess, args[index]);

	//
	// Now do all arguments in between the first and the last because
	// we don't want a space appended to the end of the string
	//

	for(i=index+1; i<nEntries; i++)
	{
		strcat(mess, " ");
		strcat(mess, args[i]);
	}

	mess[strlen(mess)] = 0;

	return mess;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	GetLength
//
// DESCRIPTION:	Return the total number of bytes in the message
//
//
//

int
PrismIPCMessage::GetLength(int index)
{
int	length=0, i;

	//
	// No need to attempt counting if there isn't anything to count
	//

	if(nEntries == 0)
	  	return 0;

	// 
	// Count the bytes...
	// 

	for(i=index, length=0; i<nEntries; i++)
	  	length += strlen(args[i]);

	length += nEntries;


	return length;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:	operator =
//
// DESCRIPTION:	Copy constructor
//
//
//

PrismIPCMessage&
PrismIPCMessage::operator=(const PrismIPCMessage &msg)
{
int	i;


	Init(); 

	//
	// Copy everything from msg to the new PrismIPCMessage class
	//

  	nEntries = msg.nEntries;
	nSlots = msg.nSlots;

	args = (char **) calloc(1, nSlots*sizeof(char*));

	for(i=0; i<nEntries; i++)
	  	STRDUP(args[i], msg.args[i]);

  	return *this;
}

