// 
//  Copyright (C) 2007-2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismPrefs.hh"
#include <gdk/gdk.h>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include "PrismUtils.hh"
#include "CiaoGuiUtils.hh"

#define PREF_MAX_LEN  1024


PrismPrefs::PrismPrefs()
{
   initialize(); 
}

PrismPrefs::~PrismPrefs()
{
}


void PrismPrefs::initialize()
{
   _tips = true;
   _info_dialogs = false;
   _confirm_dialog = true;

   _img_dialog = true;
   _plot_dialog = true;

   _hdr_add_dialog = true;
   _hdr_delete_dialog = true;

   _hdr_filter_raw = true; 

    _hdr_name_vis  = _hdr_value_vis   = _hdr_comment_vis = 
    _hdr_type_vis  = _hdr_units_vis = _hdr_keyno_vis = true; 

   _info_desc_vis  = _info_ctype_vis  = _info_dtype_vis =
   _info_units_vis = _info_min_vis    = _info_max_vis   =
   _info_null_vis  = _info_exist_vis  = _info_coord_vis = true; 
   _info_crpix_vis = _info_crval_vis  = _info_cdelt_vis = false; 

   _v_expand = true;
   _vexp_min = _vexp_max = _vclp_min = _vclp_max = -1;
   _a_expand = false; 
   _aexp_min = _aexp_max = _aclp_min = _aclp_max = -1;
   _vclp_str = _vexp_str = _aclp_str = _aexp_str = ""; 

   _rowspp = 100; 

   _select_fg = "yellow"; 
   _select_bg = "black"; 
   _oddrow_color = ""; 
   _evenrow_color = ""; 
   _altrow = false; 

   _err_txt_color = "red";
   _err_txt_italic = _err_txt_bold = false; 

   load_preferences(); 
}

int PrismPrefs::load_preferences(string fname)
{
   char   linebuffer[PREF_MAX_LEN];
   int    errs_done = 0;

   if (fname.length())
      filename = fname;
   else
   {
      filename = getenv("HOME");
      if (filename.length()) filename.append("/");
      filename.append(".prismrc");
   }

   ifstream *infile =  new ifstream(filename.c_str(), ios::in);

   if (!infile->good())  return 1;   // file failed to open

   while (infile->getline(linebuffer, PREF_MAX_LEN))
      errs_done += parse_line(linebuffer, ':');

   delete infile;

   return 0;
}


int PrismPrefs::save_preferences(string in_fname)
{
    string fname;

    // set file name  
   if (in_fname.length() == 0)
      fname = filename;         // no filename provided, use stored filename 
   else 
      fname = in_fname;         // use provided filename

   // determine access status 
   if (fname.length() == 0) 
      return 2;                 // no file name specified;
   else if (!access(fname.c_str(), F_OK))
   {  
      if (access(fname.c_str(), W_OK) != 0)
         return 3;              // don't have write access 
   }  

   ofstream out(fname.c_str());
   if (!out)   return 4;       // file create failed
   
   out << "# prism resource file " << endl;
   if (_tips)            out << "display.tooltips : true" << endl;
      else               out << "display.tooltips : false" << endl;
   out << endl;
   out << "# data table related properties" << endl;
   if (_img_dialog)      out << "data.image.dialog: true" << endl;
      else               out << "data.image.dialog: false" << endl;
   if (_plot_dialog)     out << "data.plot.dialog: true" << endl;
      else               out << "data.plot.dialog: false" << endl;
   if (_histo_dialog)    out << "data.histo.dialog: true" << endl;
      else               out << "data.histo.dialog: false" << endl;
   out << endl;
   out << "data.rows : " << _rowspp << endl;
   out << endl;
   out << "data.select.fg : " << _select_fg << endl;
   out << "data.select.bg : " << _select_bg << endl;
   out << endl;
   out << "data.oddrow.bg  : " << _oddrow_color << endl;
   out << "data.evenrow.bg : " << _evenrow_color << endl;
   if (_altrow)          out << "data.alternate.rows: true" << endl;
      else               out << "data.alternate.rows: false" << endl;
   out << endl; 

   // save user's data formats
   if (fmt_double.size() || fmt_float.size() || fmt_long.size() || 
       fmt_short.size()  || fmt_string.size())
   {
      out << "# data.format: <column name|*>,<extension name|*>,<type>,<format>"<< endl; 
      list <FORMAT_ENTRY_P>::reverse_iterator rr;
      for (rr = fmt_double.rbegin(); rr != fmt_double.rend(); rr++)
         out << "data.format: " << (*rr)->colname << "," << (*rr)->extname << \
                ",double,"<<(*rr)->format<<endl; 
      for (rr = fmt_float.rbegin(); rr != fmt_float.rend(); rr++)
         out << "data.format: " << (*rr)->colname << "," << (*rr)->extname << \
                ",float,"<<(*rr)->format<<endl; 
      for (rr = fmt_long.rbegin(); rr != fmt_long.rend(); rr++)
         out << "data.format: " << (*rr)->colname << "," << (*rr)->extname << \
                ",long,"<<(*rr)->format<<endl; 
      for (rr = fmt_short.rbegin(); rr != fmt_short.rend(); rr++)
         out << "data.format: " << (*rr)->colname << "," << (*rr)->extname << \
                ",short,"<<(*rr)->format<<endl; 
      for (rr = fmt_string.rbegin(); rr != fmt_string.rend(); rr++)
         out << "data.format: " << (*rr)->colname << "," << (*rr)->extname << \
                ",string,"<<(*rr)->format<<endl; 
      out << endl; 
   }

   if (_a_expand)        out << "array.expansion: expand" << endl;
      else               out << "array.expansion: collapse" << endl;
   out << "array.expand.except: " << _aexp_str << endl;
   out << "array.collapse.except: " << _aclp_str << endl;
   out << endl;
   if (_v_expand)        out << "vector.expansion: expand" << endl;
      else               out << "vector.expansion: collapse" << endl;
   out << "vector.expand.except: " << _vexp_str << endl;
   out << "vector.collapse.except: " << _vclp_str << endl;
   out << endl;
   if (_confirm_dialog)  out << "edit.confirm.dialog : true" << endl;
      else               out << "edit.confirm.dialog : false" << endl;
   out << endl;
   out << "# header related properties" << endl;
   if (_hdr_add_dialog)  out << "hdr.add.dialog : true" << endl;
      else               out << "hdr.add.dialog : false" << endl;
   if (_hdr_delete_dialog)  out << "hdr.delete.dialog : true" << endl;
      else               out << "hdr.delete.dialog : false" << endl;
   out << endl;
   if (_hdr_filter_raw)  out << "hdr.filter.raw : true" << endl;
      else               out << "hdr.filter.raw : false" << endl;
   out << endl;
   if (_hdr_name_vis)    out << "hdr.display.keyname : true" << endl;
      else               out << "hdr.display.keyname : false" << endl;
   if (_hdr_value_vis)   out << "hdr.display.value : true" << endl;
      else               out << "hdr.display.value : false" << endl;
   if (_hdr_type_vis)    out << "hdr.display.type : true" << endl;
      else               out << "hdr.display.type : false" << endl;
   if (_hdr_comment_vis) out << "hdr.display.comment : true" << endl;
      else               out << "hdr.display.comment : false" << endl;
   if (_hdr_units_vis)   out << "hdr.display.units : true" << endl;
      else               out << "hdr.display.units : false" << endl;
   if (_hdr_keyno_vis)   out << "hdr.display.keyno : true" << endl;
      else               out << "hdr.display.keyno : false" << endl;
   out << endl;
   out << "# info data fields" << endl;
   if (_info_desc_vis)   out << "info.display.description : true" << endl;
      else               out << "info.display.description : false" << endl;
   if (_info_ctype_vis)  out << "info.display.columntype : true" << endl;
      else               out << "info.display.columntype : false" << endl;
   if (_info_dtype_vis)  out << "info.display.datatype : true" << endl;
      else               out << "info.display.datatype : false" << endl;
   if (_info_units_vis)  out << "info.display.units : true" << endl;
      else               out << "info.display.units : false" << endl;
   if (_info_min_vis)    out << "info.display.min : true" << endl;
      else               out << "info.display.min : false" << endl;
   if (_info_max_vis)    out << "info.display.max : true" << endl;
      else               out << "info.display.max : false" << endl;
   if (_info_null_vis)   out << "info.display.null : true" << endl;
      else               out << "info.display.null : false" << endl;
   if (_info_exist_vis)  out << "info.display.existence : true" << endl;
      else               out << "info.display.existence : false" << endl;
   if (_info_coord_vis)  out << "info.display.coords : true" << endl;
      else               out << "info.display.coords : false" << endl;
   if (_info_crpix_vis)  out << "info.display.crpix : true" << endl;
      else               out << "info.display.crpix : false" << endl;
   if (_info_crval_vis)  out << "info.display.crval : true" << endl;
      else               out << "info.display.crval : false" << endl;
   if (_info_cdelt_vis)  out << "info.display.cdelt : true" << endl;
      else               out << "info.display.cdelt : false" << endl;
   out << endl;
   out << "# status message" << endl;
   if (_err_txt_bold)    out << "error.text.bold    : true" << endl;
      else               out << "error.text.bold    : false" << endl; 
   if (_err_txt_italic)  out << "error.text.italic  : true" << endl;
      else               out << "error.text.italic  : false" << endl; 
   out << "error.text.color   : " << _err_txt_color << endl;
   out << endl;
   out.close();
   return 0;
}


int PrismPrefs::parse_line(const char *in_txt, char del) 
{
   string::size_type sep;
   const char *pos = in_txt;
   string prop, val; 

   if (!in_txt)    return 1;  // no text 

   while(isspace(*pos)) pos++;  

   if (strlen(pos) == 0)   return 0;   // blank line  
   if (pos[0] == '#')      return 0;   // comment line 

   prop = pos; 
   sep = prop.find(del); 
   if (sep == string::npos)          return 2;   // no delimiter 
   val = prop.substr(sep+1, prop.length());
   prop = prop.substr(0, sep); 

   if (!val.length())    return 3;   // no value field
   if (!prop.length())    return 4;   // no prop field

   string::size_type first = val.find_first_not_of(" \t");
   if (first == string::npos)   val = "";
      else val = val.substr(first, val.find_last_not_of(" \t")-first+1);
   first = prop.find_first_not_of(" \t");
   if (first ==  string::npos)   prop = "";
      else prop = prop.substr(first, prop.find_last_not_of(" \t")-first+1);
//   prop.erase(prop.find_last_not_of(" \t")+1);

   bool bv; 
   long lv; 

   if (prop.substr(0,4) == "hdr.")
   {
      if (prop == "hdr.display.keyname")
      {
         if (!_string_to_bool(val, bv))     _hdr_name_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.display.value")
      {
         if (!_string_to_bool(val, bv))     _hdr_value_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.display.type")
      {
         if (!_string_to_bool(val, bv))     _hdr_type_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.display.comment")
      {
         if (!_string_to_bool(val, bv))     _hdr_comment_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.display.units")
      {
         if (!_string_to_bool(val, bv))     _hdr_units_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.display.keyno")
      {
         if (!_string_to_bool(val, bv))     _hdr_keyno_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.add.dialog")
      {
         if (!_string_to_bool(val, bv))     _hdr_add_dialog = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.delete.dialog")
      {
         if (!_string_to_bool(val, bv))     _hdr_delete_dialog = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "hdr.filter.raw")
      {
         if (!_string_to_bool(val, bv))     _hdr_filter_raw = bv;
         else    return 5; // value not a bool  
      }
      else 
         return 6;  // unknown prop
   }
   else if (prop.substr(0,4) == "data")
   {
      if (prop == "data.plot.dialog")
      {
         if (!_string_to_bool(val, bv))     _plot_dialog = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "data.image.dialog")
      {
         if (!_string_to_bool(val, bv))     _img_dialog = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "data.histo.dialog")
      {
         if (!_string_to_bool(val, bv))     _histo_dialog = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "data.select.fg")
      {
	 GdkColor clr; 
	 if (gdk_color_parse(val.c_str(), &clr))  _select_fg = val;
	 else  return 7;   // invalid color specification 
      }
      else if (prop == "data.select.bg")
      {
	 GdkColor clr; 
	 if (gdk_color_parse(val.c_str(), &clr))  _select_bg = val;
	 else  return 7;   // invalid color specification 
      }
      else if (prop == "data.oddrow.bg")
      {
	 GdkColor clr; 
	 if (gdk_color_parse(val.c_str(), &clr))  _oddrow_color = val;
	 else  return 7;   // invalid color specification 
      }
      else if (prop == "data.evenrow.bg")
      {
	 GdkColor clr; 
	 if (gdk_color_parse(val.c_str(), &clr))  _evenrow_color = val;
	 else  return 7;   // invalid color specification 
      }
      else if (prop == "data.alternate.rows")
      {
         if (!_string_to_bool(val, bv))     _altrow = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "data.rows")
      {
         if (_string_to_long(val, lv)) 
            return 5; // value not a long
	 else
	 {
	    return set_rows_per_page(lv);
         }
      }
      else if (prop == "data.format")
      {
         return handle_data_format(val); 
      } 
      else 
         return 6;  // unknown prop
   }
   else if (prop.substr(0,4) == "info")
   {
      if (prop == "info.display.description")
      {
         if (!_string_to_bool(val, bv))     _info_desc_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.columntype")
      {
         if (!_string_to_bool(val, bv))     _info_ctype_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.datatype")
      {
         if (!_string_to_bool(val, bv))     _info_dtype_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.units")
      {
         if (!_string_to_bool(val, bv))     _info_units_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.min")
      {
         if (!_string_to_bool(val, bv))     _info_min_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.max")
      {
         if (!_string_to_bool(val, bv))     _info_max_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.null")
      {
         if (!_string_to_bool(val, bv))     _info_null_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.existence")
      {
         if (!_string_to_bool(val, bv))     _info_exist_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.coords")
      {
         if (!_string_to_bool(val, bv))     _info_coord_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.crpix")
      {
         if (!_string_to_bool(val, bv))     _info_crpix_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.crval")
      {
         if (!_string_to_bool(val, bv))     _info_crval_vis = bv;
         else    return 5; // value not a bool  
      }
      else if (prop == "info.display.cdelt")
      {
         if (!_string_to_bool(val, bv))     _info_cdelt_vis = bv;
         else    return 5; // value not a bool  
      }
      else
         return 6;  // unknown prop
   }
   else if (prop.substr(0,5) == "array")
   {
      if (prop == "array.expansion")
      {
         if (!_strcmp_cis(val.c_str(), "expand"))
	    _a_expand = true; 
         else if (!_strcmp_cis(val.c_str(), "collapse"))
	    _a_expand = false; 
      }
      else if (prop == "array.expand.except")
      {
         _aexp_str = val; 
         extract_exception_range(val, _aexp_items, _aexp_min, _aexp_max); 
      }
      else if (prop == "array.collapse.except")
      {
         _aclp_str = val; 
         extract_exception_range(val, _aclp_items, _aclp_min, _aclp_max); 
      }
   }
   else if (prop.substr(0,6) == "vector")
   {
      if (prop == "vector.expansion")
      {
         if (!_strcmp_cis(val.c_str(), "expand"))
	    _v_expand = true; 
         else if (!_strcmp_cis(val.c_str(), "collapse"))
	    _v_expand = false; 
      }
      else if (prop == "vector.expand.except")
      {
         _vexp_str = val; 
         extract_exception_range(val, _vexp_items, _vexp_min, _vexp_max); 
      }
      else if (prop == "vector.collapse.except")
      {
         _vclp_str = val; 
         extract_exception_range(val, _vclp_items, _vclp_min, _vclp_max); 
      }
   }
   else if (prop.substr(0,4) == "edit")
   {
      if (prop == "edit.confirm.dialog")
      {
         if (!_string_to_bool(val, bv))     _confirm_dialog = bv;
         else    return 5; // value not a bool  
      }
   }
   else if (prop.substr(0,7) == "display")
   {
      if (prop == "display.tooltips")
      {
         if (!_string_to_bool(val, bv))     _tips = bv;
         else    return 5; // value not a bool
      }
   }
   else if (prop.substr(0,5) == "error")
   {
      if (prop == "error.text.color")
      {
	 GdkColor clr; 
	 if (val.length() == 0)
	    _err_txt_color = "";
	 else if (gdk_color_parse(val.c_str(), &clr))  
	    _err_txt_color = val;
	 else  return 7;   // invalid color specification 
      }
      else if (prop == "error.text.bold")
      {
         if (!_string_to_bool(val, bv))     _err_txt_bold = bv;
         else    return 5; // value not a bool
      }
      else if (prop == "error.text.italic")
      {
         if (!_string_to_bool(val, bv))     _err_txt_italic = bv;
         else    return 5; // value not a bool
      }
   }
   return 3;   // val or prop is blank  
}


int PrismPrefs::get_preference(string prop, bool &val)
{
   if (prop.substr(0,4) == "hdr.")
   {
      if (prop == "hdr.display.keyname")
         val = _hdr_name_vis;
      else if (prop == "hdr.display.keyno")
         val = _hdr_keyno_vis;
      else if (prop == "hdr.display.value")
         val = _hdr_value_vis;
      else if (prop == "hdr.display.type")
         val = _hdr_type_vis;
      else if (prop == "hdr.display.comment")
         val =  _hdr_comment_vis;
      else if (prop == "hdr.display.units")
         val = _hdr_units_vis;
      else if (prop == "hdr.add.dialog")
         val = _hdr_add_dialog;
      else if (prop == "hdr.delete.dialog")
         val = _hdr_delete_dialog;
      else if (prop == "hdr.filter.raw")
         val = _hdr_filter_raw;
      else 
         return 1; 
   }
   else if (prop.substr(0,4) == "data")
   {
      if (prop == "data.image.dialog")
         val = _img_dialog;
      else if (prop == "data.plot.dialog")
         val = _plot_dialog;
      else if (prop == "data.histo.dialog")
         val = _histo_dialog;
      else if (prop == "data.alternate.rows")
         val = _altrow;
      else 
         return 1; 
   }
   else if (prop.substr(0,4) == "info")
   {
      if (prop == "info.display.description")
         val = _info_desc_vis;
      else if (prop == "info.display.columntype")
         val = _info_ctype_vis;
      else if (prop == "info.display.datatype")
         val = _info_dtype_vis;
      else if (prop == "info.display.units")
         val = _info_units_vis;
      else if (prop == "info.display.min")
         val = _info_min_vis;
      else if (prop == "info.display.max")
         val = _info_max_vis;
      else if (prop == "info.display.null")
         val = _info_null_vis;
      else if (prop == "info.display.existence")
         val = _info_exist_vis;
      else if (prop == "info.display.coords")
         val = _info_coord_vis;
      else if (prop == "info.display.crpix")
         val = _info_crpix_vis;
      else if (prop == "info.display.crval")
         val = _info_crval_vis;
      else if (prop == "info.display.cdelt")
         val = _info_cdelt_vis;
      else
         return 1; 
   }
   else if (prop.substr(0,4) == "edit")
   {
      if (prop == "edit.confirm.dialog")
         val = _confirm_dialog;
      else
         return 1; 
   }
   else if (prop.substr(0,5) == "error")
   {
      if (prop == "error.text.bold")
         val = _err_txt_bold;  
      else if (prop == "error.text.italic")
         val = _err_txt_italic;
      else 
         return 1; 
   }
   else 
      return 1;   // value not an expected boolean 
   return 0;
}

int PrismPrefs::get_preference(string prop, long &val)
{
   if (prop == "data.rows")
      val = _rowspp; 
   else 
      return 1;
   return 0; 
}

int PrismPrefs::get_preference(string prop, string &val)
{
   if (prop == "data.select.fg")
      val = _select_fg;
   else if (prop == "data.select.bg")
      val = _select_bg;
   else if (prop == "error.text.color")
      val = _err_txt_color;
   else if (prop == "data.oddrow.bg")
      val = _oddrow_color;
   else if (prop == "data.evenrow.bg")
      val = _evenrow_color;
   else 
      return 1;
   return 0; 
}



int PrismPrefs::set_rows_per_page(long val)
{
   if (val < PPREF_MIN_ROWS_PP)  return 1;
   if (val > PPREF_MAX_ROWS_PP) return 2; 

   _rowspp = val; 
   return 0; 
}


void PrismPrefs::extract_exception_range(string val, list<string> &lst, long &emin, long &emax)
{
   lst.clear();   // remove old entries

   if (!val.length())
   {
      emin = emax = -1;
      return;
   }

   char buff[2049]; 
   char  *word, *pos = buff;
   long mn = -1, mx = -1; 

   // tokenize string into lowercase entries
   strncpy(buff, val.c_str(), 2048); 
   for (int ii = val.length(); ii--; ) buff[ii] = tolower(buff[ii]);  
   if ((word = strtok(pos, " ")) != NULL)   lst.push_back(word);
   while (word)
      if ((word = strtok(NULL, " ")) != NULL)  lst.push_back(word);

   if (!lst.empty())
   {
      list<string>::iterator ii = lst.begin();
      for(; ii != lst.end(); ++ii)

         if ((*ii).find(':') != string::npos)      // item is a range  
	    if (sscanf((*ii).c_str(), "%ld:%ld", &mn, &mx) != 2) // min & max
	       if (sscanf((*ii).c_str(), "%ld:", &mn) != 1)     // min 
	          sscanf((*ii).c_str(), ":%ld", &mx);           // max 
   }
   emin = mn;
   emax = mx; 
}


bool PrismPrefs::check_column_expansion(string colname, int vdim, int adim)
{
   list<string>::iterator ii;
   if (vdim > 1)
   {
      // vector or vector array 
      if (_v_expand)
      {
         // check exceptions 
	 if (_vexp_min != -1 && _vexp_max != -1)
	 {
	    if (vdim >= _vexp_min && vdim <= _vexp_max) return false;  
	 }
	 if (_vexp_min != -1)
	 {
	    if (vdim >= _vexp_min) return false; 
	 }
         if (_vexp_max != -1)
	 {
	    if (vdim <= _vexp_max) return false; 
	 }
	 if (colname.length())
            for(ii = _vexp_items.begin(); ii != _vexp_items.end(); ++ii)
            {
	       if (!_strcmp_cis((*ii).c_str(), colname.c_str())) return false;   
	    }
	 // we haven't been excluded so return expand (even for vector arrays)
         return true; 
      }
      else   // collapse  
      {
         list<string>::iterator ii;

         // check exceptions 
	 if (_vclp_min != -1 && _vclp_max != -1)
	 {
	    if (vdim >= _vclp_min && vdim <= _vclp_max) return true;  
	 }
	 if (_vclp_min != -1)
	 {
	    if (vdim >= _vclp_min) return true; 
	 }
         if (_vclp_max != -1)
	 {
	    if (vdim <= _vclp_max) return true; 
	 }
	 if (colname.length())
            for(ii = _vclp_items.begin(); ii != _vclp_items.end(); ++ii)
            {
	       if (!_strcmp_cis((*ii).c_str(), colname.c_str())) return true;   
	    }
	 // if we are a vector array check to see if the array expansion is
	 // enabled otherwise we're collapsed
         if (adim != 0)  return false; 
      }
   }
   if (adim > 0)  // array or vector array (if vector set to collapse)
   {
      if (_a_expand)
      {
         // check exceptions 
	 if (_aexp_min != -1 && _aexp_max != -1)
	 {
	    if (adim >= _aexp_min && adim <= _aexp_max) return false;  
	 }
	 if (_aexp_min != -1)
	 {
	    if (adim >= _aexp_min) return false; 
	 }
         if (_aexp_max != -1)
	 {
	    if (adim <= _aexp_max) return false; 
	 }
	 if (colname.length())
            for(ii = _aexp_items.begin(); ii != _aexp_items.end(); ++ii)
            {
	       if (!_strcmp_cis((*ii).c_str(), colname.c_str())) return false;   
	    }
         return true; 
      }
      else   // collapse  
      {
         list<string>::iterator ii;

         // check exceptions 
	 if (_aclp_min != -1 && _aclp_max != -1)
	 {
	    if (adim >= _aclp_min && adim <= _aclp_max) return true;  
	 }
	 if (_aclp_min != -1)
	 {
	    if (adim >= _aclp_min) return true; 
	 }
         if (_aclp_max != -1)
	 {
	    if (adim <= _aclp_max) return true; 
	 }
	 if (colname.length())
            for(ii = _aclp_items.begin(); ii != _aclp_items.end(); ++ii)
            {
	       if (!_strcmp_cis((*ii).c_str(), colname.c_str())) return true;   
            }
         return false; 
      }
   }
   return false; 
}


void PrismPrefs::set_expand_vectors_filter(string val)
{
   if (_vexp_str != val)
   {
      _vexp_str = val; 
      extract_exception_range(val, _vexp_items, _vexp_min, _vexp_max); 
   }
}

void PrismPrefs::set_expand_arrays_filter(string val)
{
   if (_aexp_str != val)
   {
      _aexp_str = val; 
      extract_exception_range(val, _aexp_items, _aexp_min, _aexp_max); 
   }
}

void PrismPrefs::set_collapse_vectors_filter(string val)
{
   if (_vclp_str != val)
   {
      _vclp_str = val; 
      extract_exception_range(val, _vclp_items, _vclp_min, _vclp_max); 
   }
}

void PrismPrefs::set_collapse_arrays_filter(string val)
{
   if (_aclp_str != val)
   {
      _aclp_str = val; 
      extract_exception_range(val, _aclp_items, _aclp_min, _aclp_max); 
   }
}

string PrismPrefs::get_format(string cname, string ename, int type)
{
   string fmt = ""; 
   bool done = 0;
   list <FORMAT_ENTRY_P>::reverse_iterator rr;

   switch (type)
   {
      case FORMAT_DOUBLE:
         for (rr = fmt_double.rbegin(); rr != fmt_double.rend() && !done; rr++)
            if ((!_ciao_strcmp_cis((*rr)->colname,cname)||(*rr)->colname=="*") &&
                (!_ciao_strcmp_cis((*rr)->extname,ename)||(*rr)->extname=="*"))
            {
               done = 1;
               fmt = (*rr)->format;
            }
      break; 
      case FORMAT_FLOAT:
         for (rr = fmt_float.rbegin(); rr != fmt_float.rend() && !done; rr++)
            if ((!_ciao_strcmp_cis((*rr)->colname,cname)||(*rr)->colname=="*") &&
                (!_ciao_strcmp_cis((*rr)->extname,ename)||(*rr)->extname=="*"))
            {
               done = 1;
               fmt = (*rr)->format;
            }
      break; 
      case FORMAT_LONG:
         for (rr = fmt_long.rbegin(); rr != fmt_long.rend() && !done; rr++)
            if ((!_ciao_strcmp_cis((*rr)->colname,cname)||(*rr)->colname=="*") &&
                (!_ciao_strcmp_cis((*rr)->extname,ename)||(*rr)->extname=="*"))
            {
               done = 1;
               fmt = (*rr)->format;
            }
      break; 
      case FORMAT_SHORT:
         for (rr = fmt_short.rbegin(); rr != fmt_short.rend() && !done; rr++)
            if ((!_ciao_strcmp_cis((*rr)->colname,cname)||(*rr)->colname=="*") &&
                (!_ciao_strcmp_cis((*rr)->extname,ename)||(*rr)->extname=="*"))
            {
               done = 1;
               fmt = (*rr)->format;
            }
      break; 
      case FORMAT_STRING:
         for (rr = fmt_string.rbegin(); rr != fmt_string.rend() && !done; rr++)
            if ((!_ciao_strcmp_cis((*rr)->colname,cname)||(*rr)->colname=="*") &&
                (!_ciao_strcmp_cis((*rr)->extname,ename)||(*rr)->extname=="*"))
            {
               done = 1;
               fmt = (*rr)->format;
            }
      break; 
      default:
      break; 
   }
   return fmt; 
}

int PrismPrefs::handle_data_format(string val)
{
   vector<string> components = _ciao_string_split(val, ",");
   int typ, sz = components.size();
   string cname, ename, fmt;

   if (sz < 3 || sz > 4)  return 1;
   cname = _ciao_strip_whitespace(components.at(0));
   ename = _ciao_strip_whitespace(components.at(1));
   if (sz == 4)   
      fmt   = _ciao_strip_whitespace(components.at(3));
   else   
      fmt   = ""; 
   if (_ciao_strcmp_loose(components.at(2), "double") == 0) typ = 0; 
   else if (_ciao_strcmp_loose(components.at(2), "float") == 0) typ = 1; 
   else if (_ciao_strcmp_loose(components.at(2), "long") == 0) typ = 2; 
   else if (_ciao_strcmp_loose(components.at(2), "short") == 0) typ = 3; 
   else if (_ciao_strcmp_loose(components.at(2), "string") == 0) typ = 4; 
   else  return 3;

   if (_valid_format(fmt, typ) == 0) return 2; 

   FORMAT_ENTRY_P elem_p = new FORMAT_ENTRY;
   elem_p->colname = cname;
   elem_p->extname = ename;
   elem_p->format = fmt;

   switch (typ)
   {
      case 0:  fmt_double.push_back(elem_p);  break;
      case 1:  fmt_float.push_back(elem_p);  break;
      case 2:  fmt_long.push_back(elem_p);  break;
      case 3:  fmt_short.push_back(elem_p);  break;
      case 4:  fmt_string.push_back(elem_p);  break;
      default: break;
   }
   return 0; 
}



