// 
//  Copyright (C) 1998,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include   <limits.h>

#define   _CIAOGUIIPC_LIBRARY_

#include "CiaoGuiIPCComm.hh"
#include "CiaoGuiUtils.hh"

#define   CIAOGUI_IPC_FORK_EXEC

#define MAX_XPA_BUFS   30

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCComm
//
// DESCRIPTION:   Base class constructor for CiaoGuiIPC communication classes
//      
// PARAMETERS:   pname   - the name of the application.
//      targs   - the default set of arguments to use when 
//           forking/execing applications.
//      
//      

CiaoGuiIPCComm::CiaoGuiIPCComm(char* pname, char **targs)
{
   Initialize(pname, targs);
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   Initialize
//
// DESCRIPTION:   Initialization routine for all constructors.
//
//

void CiaoGuiIPCComm::Initialize(char* pname, char **targs)
{
   int   i, count;

   argv = NULL;
   pxpa = NULL;
   name = pname;
   _launched_app = 0;

   // determine the amount of time to wait before checking to see if a
   // Launch() succeeded
   waittime = 0;
   if (getenv("ASCDS_LAUNCH_WAIT"))
      waittime = atoi(getenv("ASCDS_LAUNCH_WAIT"));

   if (waittime == 0)
      waittime = COMM_WAIT_FOR_LAUNCH_SECONDS;

   // If the user passed in some argument to be used as command line 
   // arguments, then process them now. These will be the default.
   if (targs != NULL)
   {
      // allocate memory for the array of pointers.
      for(count=0; targs[count]; count++); 
      argv = (char**) malloc((count+1)*sizeof(char*));

      // cycle through and process all individual parameters...
      for(i=0; targs[i]; i++)
      {
         argv[i] = (char*) malloc(strlen(targs[i]) + 1);
         strcpy(argv[i], targs[i]);
      }

      // The last entry must be NULL
      argv[count] = NULL;
   }
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   ~CiaoGuiIPCComm
//
// DESCRIPTION:   Destructor for the CiaoGuiIPCComm class.
//
//
//


CiaoGuiIPCComm::~CiaoGuiIPCComm()
{
   // Free up te memory for the arguments, if any...
   if (argv != NULL)
   {
      for(int ii=0; argv[ii]; ii++)
           free(argv[ii]);
      free(argv);
   }
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   Poll()
//
// DESCRIPTION:   Poll for messages. Wait msec milliseconds before returning
//
//
//

void CiaoGuiIPCComm::Poll(int msec)
{
   XPAPoll(msec, 0);
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CheckParameters
//
// DESCRIPTION:   Check the parameters for the Send() and SendWithReply()
//      routines
//
//

int CiaoGuiIPCComm::CheckParameters(char *dest, CiaoGuiIPCMessage &pmess)
{
   // Can't send anything unless we have a destination.
   if (dest == NULL)   return -1;
	  
   // Can't send anything unless we have a connection.
   if (pxpa == NULL)   return -1;

   // Check that there is data to send, before trying to send...
   if (pmess.GetLength() < 1)   return -1;

   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:     Send
//
// DESCRIPTION: Send a message to another process.
//
// PARAMETERS:
//
//      dest    The name of the process you want to send the message
//      pmess   Message class (reference). This class contains the message
//              to be sent.
//      launch  optional argument. Defaults to 1, which means to launch the
//              application if it is not already running. 0 means to not
//              launch.
//
// RETURNS:     0 on success, -1 or error. If an error occurs, check the value
//              of errno to get a more detailed description of the error.
//
// ERRNO        ENOTCONN - XPA failure.
//              EDESTADDRREQ - Destination address required.
//              ENODATA - No data to be sent to destination.
//              ESRCH -  Program failed to launch.
//              plus any errno values that may originate from the system(2),
//              fork(2) and exec(2) system calls.
//
// NOTE:        The message size that can be sent through this interface is in
//              theory unlimited.
//
//

int CiaoGuiIPCComm::Send(char *destname, CiaoGuiIPCMessage &pmess, int launch,
   char *alias, char** targv)
{
   char   *dest = destname;
   char   *names[MAX_XPA_BUFS], *errs[MAX_XPA_BUFS];
   char   *buffer;
   int    status=0, length;

   // initialize xpa buffers
   for (length = MAX_XPA_BUFS; length--; )
      names[length] = errs[length] = NULL;

   // Check the parameters to make sure we can continue
   if ((status = CheckParameters(destname, pmess)) == -1)  return -1;

   buffer = pmess.GetMessage();
   length = strlen(buffer);

   // Send the message
   if (alias)   dest = alias;

   status = XPASet(pxpa, dest, NULL, 0, buffer, length,
      names, errs, MAX_XPA_BUFS);

   // Check to see if the message was sent. If not, launch application
   if (status == 0 && launch)  status = Launch(destname, targv, dest);

   // Re-send the message if necessary
   if (status == 0)
      status = XPASet(pxpa,dest,NULL,0,buffer,length,names,errs,MAX_XPA_BUFS);

   // Free the memory and exit.
   if ((status > 0) && (status < MAX_XPA_BUFS)) 
      length = status;
   else 
      length = MAX_XPA_BUFS;

   while (length--)
   {
      free (names[length]);
      free (errs[length]);
   }
   delete []buffer;
   return status;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:     SendWithReply
//
// DESCRIPTION: Send the message to dest, and then wait until we receive the
//              reply back before returning to the caller.
//
// PARAMETERS:
//
//      dest    The name of the process you want to send the message to.
//      pmess   The CiaoGuiIPCMessage class which contains the message to be sent
//   pretmes   The message that gets filled in and returned back  to the 
//      caller.
//   launch   optional argument. Defaults to 1, which means to launch the
//      application if it is not already running. 0 means to not
//      launch.
//
// RETURNS:     0 on success, -1 or error. If an error occurs, check the value
//              of errno to get a more detailed description of the error.
//
// ERRNO        ENOTCONN - XPA failure.
//              EDESTADDRREQ - Destination address required.
//              EMSGSIZE - message to long
//              ENODATA - No data to be sent to destination.
//              ESRCH -  Program failed to launch.
//                      plus any errno values that may originate from the
//                      system(2) and fork(2) system calls during launch
//         failures.
//
// NOTE:        The maximum message size that can be passed
//              through this interface is limited to ~150 bytes. This is
//              a limitation of the XPA API.
//

int CiaoGuiIPCComm::SendWithReply(char *destname, CiaoGuiIPCMessage &pmess, 
   CiaoGuiIPCMessage &pretmes, int launch, char *alias, char** targv)
{
   char   *buffer, *dest = destname;
   int    status, lens[1]={0}, length;
   char   *bufs[1]={NULL}, *names[1]={NULL}, *errs[1]={NULL};

   // Check the parameters to make sure we can continue
   if ((status = CheckParameters(destname, pmess)) == -1)  return -1;

   // Extract the message and compute the length of the messsage to send
   buffer = (char*) pmess.GetMessage();
   length = strlen(buffer);

   // SendWithReply has a byte limitation, check it now.
   if (length > CiaoGuiIPC_MAX_MESSAGE_LENGTH)
   {
      delete []buffer;
      return -1;    // Message too long
   }

   // Send the message to 'dest'
   if (alias)  dest = alias;

   status = XPAGet(pxpa,dest,buffer,NULL,bufs,lens,names,errs,1);

   // Check to see if the message was sent. If not, then launch the
   // application
   if (status == 0 && launch)  status = Launch(destname, targv, dest);

   // re-send the message if necessary
   if (status == 0)
      status = XPAGet(pxpa,dest,buffer,NULL,bufs,lens,names,errs,1);

   // Create a new message from the reply received from dest
   if (status != -1 && bufs[0] && strlen(bufs[0]) > 0)
   {
      pretmes.SetParam(bufs[0]);
      free(bufs[0]);
      free(names[0]);
   }

   // free the allocated memory
   delete []buffer;

   return status;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   Launch
//
// DESCRIPTION:   fork/exec an application. We do not worry about handling
//      SIGCHLD signals because we do not care whether or not
//      the child exits or core dumps. In fact, this application
//      may exit and the 'launched' application may continue running.
//      We just don't care.
//
//

int CiaoGuiIPCComm::Launch(char *program, char **targv, char *alias)
{
   int   status;
   char* default_argv[2], **the_argv;

   if (program == NULL || strlen(program) == 0)   return -1;

   if (targv != NULL)
      the_argv = targv;
   else if (argv != NULL)
      the_argv = argv;
   else
   {
      default_argv[0] = program;
      default_argv[1] = NULL;
      the_argv = default_argv;
   }

   string tmp = program;

   // Ensure program exists
   string path = _ciao_path_expand(tmp);
   if (!_ciao_file_executable(path)) 
   {
      cout << "Unable to access " << path << endl; 
      return -1;
   }

#if defined(CIAOGUI_IPC_FORK_EXEC)
   // Create another process.  
   int pid = fork();

   // If this is the child, then overlay this process image with
   // the application that we will be launching.
   if (pid == 0)
   {
      // fire up the new process...
      status = execvp(program,the_argv);
      // Shouldn't get here, but just in case
      string msg = "cannot launch ";
      msg.append(program);
      cout << msg.c_str() << endl;
   }
#else
   string cmd(program);
   cmd.append(" ");
   int i = 1;
   while (the_argv[i])
   {
      cmd.append(" ");
      cmd.append(the_argv[i++]);
   }
   cmd.append(" &");
   system(cmd.c_str());
#endif

   if (alias)
      status = WaitForLaunch(alias);
   else
      status = WaitForLaunch(program);

   _launched_app |= (status == 0);
 
   return status;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AlreadyRunning   
//
// DESCRIPTION:   Check to see if the given application is already up and 
//              running using XPAAccess.
//
// RETURNS:   0 if app is not running, 1 if app is already running.
//
//

int CiaoGuiIPCComm::AlreadyRunning(char* app)
{
#ifndef XPA_MAJOR_VERSION
   /* XPAAccess signature changed in V2.1; */
   return XPAAccess(app, "g");
#else
   char *names[1]={NULL}, *errs[1]={NULL};
   return XPAAccess(pxpa,app,"g",NULL, names, errs, 1);
#endif
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   WaitForLaunch
//
// DESCRIPTION:   wait for a maximum 'waittime' seconds for an application to
//      start and register itself. Sleeps with a granularity of
//      100 milliseconds second between checks.
//
// RETURNS:   0 if successful. -1 otherwise. Check the value of errno for a
//      detailed description of any errors.
//
// ERRNO:   ESRCH if the process did not come up with a specified
//         period of time
//
//

int CiaoGuiIPCComm::WaitForLaunch(char *program)
{
   for(int ii=0; ii < waittime*10; ii++)
   {
#ifndef XPA_MAJOR_VERSION
     /* XPAAccess signature changed in V2.1; */
     if (XPAAccess(program,"g"))
#else
       char  *names[1]={NULL}, *errs[1]={NULL};
       if (XPAAccess(pxpa,program,"g",NULL, names, errs, 1))
#endif
      return 0;

      usleep(100000);
   }
   return -1;
}



///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   EventLoop
//
// DESCRIPTION:   Call the XPA event loop
//
//
//

void CiaoGuiIPCComm::EventLoop()
{
   XPAMainLoop();
}
