// 
//  Copyright (C) 1998,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "CiaoGuiIPCMessage.hh"
#include <string>


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCMessage
//
// DESCRIPTION:   Constructor for the CiaoGuiIPCMessage class. The string passed in
//      will be split up with a SPACE character being the delimiter.
//
// EXAMPLE:   input string:   "This is string one"
//
//         param[0]:   "This"
//         param[1]:   "is"
//         param[2]:   "string"
//         param[3]:   "one"
//
//

CiaoGuiIPCMessage::CiaoGuiIPCMessage(char *paramlist)
{
   if(paramlist == NULL)  return;  

   Init();

   // Set the arguments
   SetParam(paramlist);
}

CiaoGuiIPCMessage::CiaoGuiIPCMessage(string paramlist)
{
   Init();

   // Set the arguments
   SetParam((char*)paramlist.c_str());
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   SetParam
//
// DESCRIPTION:   break up the string into space separated parameters. This will
//      clear out any other argument that may have been contained
//      in the message.
//

int CiaoGuiIPCMessage::SetParam(char *paramlist)
{
   char  *p1, *p2, *tmp;

   // Free up any previous arguments
   Clear();

   if ((paramlist == NULL) || (strlen(paramlist) == 0)) return -1; 
 
   // Make a duplicate of the paramlist
   STRDUP(tmp, paramlist);

   p1 = tmp;

   // Break up the paramlist.
   for(nEntries=0; p1 && nEntries < nSlots-1; nEntries++)
   {
      if ((*p1 == '"') && (p2 = strchr(p1+1,'"')) != NULL) 
      {
         // quoted string, take everything till next quote;
         // unbalanced quote falls through to else clause
         int ll = p2 - p1 + 1;
         args[nEntries] = (char*)malloc(ll+1);
         strncpy(args[nEntries],p1,ll);
         args[nEntries][ll] = 0;
      }
      else 
      {
         // Locate the next SPACE
         p2 = strchr(p1, ' ');

         // If not more SPACES, then we're done
         if(p2 == NULL)  break;

         // Write in a NULL to terminate the string
         *p2 = 0;

         // duplicate the string
         STRDUP(args[nEntries], p1);
      }

      // Jump past the space/quote
      p1 = p2+1;

      // jump past all blank spaces.
      while (p1 && *p1 && *p1 == ' ') p1++;
   }

   // Grab the last one.
   if (p1 && *p1)
   {
      STRDUP(args[nEntries], p1);
      nEntries++;
   }

   // This is currently a constant, but that could change
   nSlots = CiaoGuiIPC_PARAM_LIMIT;

   free(tmp);

   return 0;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCMessage
//
// DESCRIPTION:   Constructor for the CiaoGuiIPCMessage class
//
//
//

CiaoGuiIPCMessage::CiaoGuiIPCMessage()
{
   Init();
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   ~CiaoGuiIPCMessage
//
// DESCRIPTION:   Destructor for the CiaoGuiIPCMessage class
//
//
//

CiaoGuiIPCMessage::~CiaoGuiIPCMessage()
{
   Clear();
   free(args);
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   Init
//
// DESCRIPTION:   Do basic startup initialization of variables and anything
//      that may need to be initialized to certain values
//
//

void
CiaoGuiIPCMessage::Init()
{
   nSlots = CiaoGuiIPC_PARAM_LIMIT;
   args = (char**)calloc(CiaoGuiIPC_PARAM_LIMIT,sizeof(char*));
   nEntries = 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   Clear
//
// DESCRIPTION:   Clear all parameters within the message
//
//

int CiaoGuiIPCMessage::Clear()
{
   for(int ii=0; ii < nEntries; ii++)   free(args[ii]);
   nEntries = 0;
   return 0;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddPAram
//
// DESCRIPTION:   Add a Strings parameter to the message
//
//
//

int CiaoGuiIPCMessage::AddParam(string &s)
{
   // Can we add this parameter ??
   if (nEntries >= nSlots)  return -1;

   // Add the new entry
   if (s.length() > 0)
   {
      STRDUP(args[nEntries], s.c_str());
      nEntries++;
   }
   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddParam
//
// DESCRIPTION:   retrun a float to the requested parameter.
//
//

int CiaoGuiIPCMessage::AddParam(float ff)
{
   strstream param;

   // Can we add this parameter ??
   if (nEntries >= nSlots)   return -1;

   param << ff << ends;

   STRDUP(args[nEntries], param.str());
   nEntries++;
   return 0;
} 
        
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddParam
//
// DESCRIPTION:   Add a float parameter to the message.
//
//

int CiaoGuiIPCMessage::AddParam(double d)
{
   // Can we add this parameter ??
   if (nEntries >= nSlots)  return -1;

   ostrstream param;
   param << d << ends;

   STRDUP(args[nEntries], param.str());
   nEntries++;
   return 0;
} 
        
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddParam
//
// DESCRIPTION:   Add an integer parameger to the message.
//
//

int CiaoGuiIPCMessage::AddParam(int i)
{
   // Can we add this parameter ??
   if (nEntries >= nSlots)  return -1;

   strstream param;
   param << i << ends;

   STRDUP( args[nEntries], param.str() );
   nEntries++;
   return 0;
} 
        
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddParam
//
// DESCRIPTION:   Add a long parameter to the message.
//
//

int CiaoGuiIPCMessage::AddParam(long ll)
{
   // Can we add this parameter ??
   if (nEntries >= nSlots)  return -1;

   ostrstream param;
   param << ll << ends;

   STRDUP(args[nEntries], param.str());
   nEntries++;
   return 0;
} 
        
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddParam
//
// DESCRIPTION:   Add a char* parameter to the message.
//
//

int CiaoGuiIPCMessage::AddParam(char *cc)
{
   // Can we add this parameter ??
   if (nEntries >= nSlots)   return -1;

   STRDUP(args[nEntries], cc);
   nEntries++;
   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetParam
//
// DESCRIPTION:   retrun a Strings class to the requested parameter.
//
//

int CiaoGuiIPCMessage::GetParam(int index, string &ss)
{
   // make sure the index is within range
   if (index >= nEntries && index > -1)   return -1;

   // Create the Strings class
   ss = args[index];
   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetParam
//
// DESCRIPTION:   retrun a float to the requested parameter.
//
//

int CiaoGuiIPCMessage::GetParam(int index, float &ff)
{
   // make sure the index is within range
   if (index >= nEntries && index > -1)   return -1;

   ff = atof(args[index]);
   return 0;
} 
   
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetParam
//
// DESCRIPTION:   retrun a double to the requested parameter.
//
//

int CiaoGuiIPCMessage::GetParam(int index, double &dd)
{
   // make sure the index is within range
   if (index >= nEntries && index > -1)  return -1;
   dd = strtod(args[index], NULL);

   return 0;
} 
   
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetParam
//
// DESCRIPTION:   retrun an integer to the requested parameter.
//
//

int CiaoGuiIPCMessage::GetParam(int index, int &ii)
{
   // make sure the index is within range
   if (index >= nEntries && index > -1)   return -1;
   ii = atoi(args[index]);

   return 0;
} 
        
///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetParam
//
// DESCRIPTION:   retrun a long to the requested parameter.
//
//

int CiaoGuiIPCMessage::GetParam(int index, long &ll)
{
   // make sure the index is within range
   if (index >= nEntries && index > -1)  return -1;
   ll = atol(args[index]);
   return 0;
} 


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetMessage
//
// DESCRIPTION:   concatenate all parameters together into a NULL terminated
//      string
//
//

char* CiaoGuiIPCMessage::GetMessage(int index)
{
   char   *mess;
   int   length;

   if(index >= nEntries)   return NULL;

   length = GetLength(index);
   mess = new char[length];

   if (mess == 0)  return 0;

   mess[0] = 0;

   //
   // Concatenate them. Do the first argument first
   //

   strcat(mess, args[index]);

   //
   // Now do all arguments in between the first and the last because
   // we don't want a space appended to the end of the string
   //

   for(int ii=index+1; ii < nEntries; ii++)
   {
      strcat(mess, " ");
      strcat(mess, args[ii]);
   }
   mess[strlen(mess)] = 0;

   return mess;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   GetLength
//
// DESCRIPTION:   Return the total number of bytes in the message
//
//
//

int CiaoGuiIPCMessage::GetLength(int index)
{
   int   length=0, i;

   // No need to attempt counting if there isn't anything to count
   if(nEntries == 0)  return 0;

   // Count the bytes...
   for(i=index, length=0; i<nEntries; i++)
      length += strlen(args[i]);

   length += nEntries;

   return length;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   operator =
//
// DESCRIPTION:   Copy constructor
//
//
//

CiaoGuiIPCMessage& CiaoGuiIPCMessage::operator=(const CiaoGuiIPCMessage &msg)
{
   Init(); 

   // Copy everything from msg to the new CiaoGuiIPCMessage class
   nEntries = msg.nEntries;
   nSlots = msg.nSlots;

   args = (char **) calloc(1, nSlots*sizeof(char*));
   for(int ii = 0; ii < nEntries; ii++)    STRDUP(args[ii], msg.args[ii]);

   return *this;
}

