// 
//  Copyright (C) 1998,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//



#define   _CIAOGUIIPC_LIBRARY_

#include   "CiaoGuiIPCServer.hh"

static   CiaoGuiIPCServer   *pgthis=NULL;
CiaoGuiIPCCallbackTablePtr  CiaoGuiIPCServer::cbTable;

// static members 
int CiaoGuiIPCServer::_refcount    = 0; 
int CiaoGuiIPCServer::_initialized = 0; 

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCServer::CiaoGuiIPCServer
//
// DESCRIPTION:   Constructor for the CiaoGuiIPCServer class
//
// NOTE: The constructor no longer initializes the CiaoGuiIPCServer. The argument
//       pname is the name of the application. If the argument initialize
//       is set to zero, the user must  explicitly call InitializeServer()
//       to initialize the xpa acess point.
//

CiaoGuiIPCServer::CiaoGuiIPCServer(char* pname, int initialize):CiaoGuiIPCComm(pname) 
{
   _refcount++; 
   if (!pgthis) 
      pgthis = this;    // Save a copy of 'this'

   if (initialize)   InitializeServer(); 
}



///////////////////////////////////////////////////////////////////////////////
// ROUTINE:     CiaoGuiIPCServer::InitializeServer
//
// DESCRIPTION: routine to initialize the CiaoGuiIPCServer - setting up default
//              access name. The first argument app is the name of the 
//              application. The second argument is the name to assign the
//              access point. If the second argument is NULL, the application
//              name is used as the access point name.   
//
//

void CiaoGuiIPCServer::InitializeServer(char* nam) 
{
   if (!pxpa)   // has the xpa handle been set yet? 
   {
      _useDefaultName = 1;
      if (nam && (nam[0] != '\0'))
      {
         _accessName = nam; 
         _useDefaultName = 0;
      }
      else
      {
         _accessName = name; 
      } 
      Init();   // actual routine to perform initialization 
   }
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:     CiaoGuiIPCServer::SetupServer
//
// DESCRIPTION: Initializes the xpa access point and add standard callbacks
//              to server
//
//

void CiaoGuiIPCServer::SetupServer(const char* classnam, const char* nam)
{
   pxpa = XPANew((char*)classnam, (char*)nam, NULL, &CiaoGuiIPCServer::sendCallback,
                 NULL, NULL, &CiaoGuiIPCServer::receiveCallback, NULL, NULL);
   AddCallback( "exit", &CiaoGuiIPCServer::exitMsgCB,
                "exit           : exits application");
   AddCallback( "quit", &CiaoGuiIPCServer::exitMsgCB, 
                "quit           : exits application");
   _accessName = nam;
   _initialized = 1;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCServer::Init
//
// DESCRIPTION:   Common Initializer for the CiaoGuiIPCServer class
//
//
//

void CiaoGuiIPCServer::Init()
{
   if(!_initialized)
   {
      string xpanam = name; 
      cbTable = 0;

      if (!_useDefaultName && (_accessName.length()))
         xpanam = _accessName.c_str(); 
      else
         xpanam = name;

      SetupServer(name.c_str(), _accessName.c_str());

   }  // end if(!_initialized)
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CiaoGuiIPCServer::sendCallback
//
// DESCRIPTION:   This routine gets called when an application makes a request
//      for the server process to do something. A reply message is
//      sent back to the client.
//
//

int CiaoGuiIPCServer::sendCallback(void *receivedata, void *calldata, char *paramlist,
   char **buf, int *len)
{
   CiaoGuiIPCMessage pmess(paramlist), *pretmes=NULL;
   (void) receivedata;   // to avoid compiler warnings 
   (void) calldata;      // to avoid compiler warnings 

   // Can't do anything without a paramlist
   if(paramlist == NULL)  return 0;

   if (!strcmp( paramlist, "" ))
   {
      // If an empty paramlist is supplied, then return info
      // of all the available xpa command line options.
      pretmes = pgthis->CmdOptions();
   }
   else
   {
      // Call the appropiate callback function
      pretmes = pgthis->DoCallback(pmess);
   }

   // initialize the length to ZERO, and if we received a CiaoGuiIPCMessage
   // back from the callback function, then set up the return parameters
   // correctly so XPA can send the data back to the client process
   *len = 0;

   if(pretmes && pretmes->GetMessage())
   {
      *buf = (char*) calloc(1, strlen(pretmes->GetMessage()) + 1);
      strcpy(*buf, pretmes->GetMessage());
      *len = strlen(*buf);
      delete pretmes;
   }
   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   receiveCallback
//
// DESCRIPTION:   This function is called when a client process makes a request
//      of the server process and the client is not expecting any data
//      to be returnd back to the client from the server.
//
//

int CiaoGuiIPCServer::receiveCallback(void *receivedata, void *calldata, 
   char *paramlist, char *buf, int len)
{
   CiaoGuiIPCMessage *pmsg=NULL;
   char *ptr;
   (void) receivedata;   // to avoid compiler warnings 
   (void) calldata;      // to avoid compiler warnings 
   (void) len;           // to avoid compiler warnings 

   if (paramlist == NULL || strlen(paramlist) == 0)
      ptr = buf;
   else
      ptr = paramlist;

   CiaoGuiIPCMessage   pmess(ptr);

   if(paramlist != NULL)
      pmsg = pgthis->DoCallback(pmess); 

   // DoCallback could return a pointer to an CiaoGuiIPCMessage class, even
   // though a message is not being returned from this routine. It is
   // our responsibility to make sure the class is freed before we
   // return.
   if(pmsg) delete pmsg;

   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   ~CiaoGuiIPCServer
//
// DESCRIPTION:   Destructor for the CiaoGuiIPCServer class
//
//
//

CiaoGuiIPCServer::~CiaoGuiIPCServer()
{
   // Only really destruct when the reference count has fallen to zero
   if(--_refcount == 0)
   {
      // Free up the XPA connection
      if (pxpa)  XPAFree(pxpa);

      // Free the callback table
      CiaoGuiIPCCallbackTablePtr tmp,tcb = cbTable;

      while(tcb)
      {
         tmp = tcb;
         tcb = tcb->next;
         free(tmp);
      }
      cbTable = NULL;
   }
   else
   {
      CiaoGuiIPCCallbackTablePtr   curr, prev=0;
      curr = cbTable;

      while(curr)
      {
         // If the names match, then delete this entry
         if (!strcmp(curr->name, name.c_str()))
         {
            // If prev == NULL, then we're deleting 
            // the head of the list
            if(prev != NULL)
               prev->next = curr->next;
            else
               prev = curr->next;
            free(curr);
            curr = prev;
         }
         prev = curr;
         curr = curr->next;
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   AddCallback
//
// DESCRIPTION:   Add a callback to our internal linked list so the caller
//      can be notified of specific messages.
//
// PARAMETERS:
//
//   pname      The name of the message
//   appcb      The address of a routine to be called when a message
//         arrives that has a keyword of 'pname'
//
// NOTES:   Q: Why do we keep our own callback table?
//      
//      A: Becasue the purpose of the CiaoGuiIPC classes is to encapsulate 
//      some form of CiaoGuiIPC mechanism from the applications. In theory,
//      we can remove XPA rfom underneath the CiaoGuiIPC classes and replace
//      it with someting else and the applications will not have to
//      change a single line of code (in theory).
//
//      Even though XPA itself could callback the applications
//      directly, this enables us to keep the applications clean
//      and free of XPA (or whatever) specific routines. XPA knows
//      of exactly one 2 callbacks within any application that uses
//      the CiaoGuiIPC classes. And those are:
//      
//         CiaoGuiIPCServer::sendCallback
//         CiaoGuiIPCServer::receiveCallback
//      
//      These 2 routines in turn then execute a callback into the
//      application space.
//      
//      
//      

int CiaoGuiIPCServer::AddCallback(char *pname, 
   CiaoGuiIPCMessage* (*appcb)(CiaoGuiIPCMessage&), string doc_comment)
{
   CiaoGuiIPCCallbackTablePtr  tcb, newcb;

   // Absolutely need a callback address
   if(appcb == NULL)   return -1;

   // Create an entry for this new callback
   newcb = (CiaoGuiIPCCallbackTablePtr)
           calloc(1, sizeof(CiaoGuiIPCCallbackTable));

   if(newcb == NULL)   return -1;

   if(pname != NULL)   strcpy(newcb->key, pname);

   strcpy(newcb->name, name.c_str());

   newcb->cb = appcb;

   // Add this keyword and address to our internal callback table
   if(cbTable == NULL)
   {
      // First time through.
      cbTable = newcb;
   }
   else
   {
      // Not the first time through. Add the new entry to the end of the list
      tcb = cbTable;

      // get to the end
      while(tcb->next)   tcb = tcb->next;

      // append the new entry
      tcb->next = newcb;
   }
   if (doc_comment.length())  _optList.push_back(doc_comment);
   return 0;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   exitMsgCB
//
// DESCRIPTION:   
//
//

CiaoGuiIPCMessage* CiaoGuiIPCServer::exitMsgCB(CiaoGuiIPCMessage&)
{
   pgthis->Exit();
   return 0;
}

///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   CmdOptions   
//
// DESCRIPTION:   
//
//

CiaoGuiIPCMessage* CiaoGuiIPCServer::CmdOptions()
{
   CiaoGuiIPCMessage* cmdMsg = new CiaoGuiIPCMessage();

   // generate response from AddCallback comments 
   list<string>::iterator ii; 
   string sv;
   cmdMsg->AddParam( "\n" );
   for (ii = _optList.begin(); ii != _optList.end(); ii++)
   {
      sv = (*ii) + "\n"; 
      cmdMsg->AddParam(sv);
   }
   cmdMsg->AddParam( "\n" );
   return cmdMsg;
}


///////////////////////////////////////////////////////////////////////////////
// ROUTINE:   DoCallback
//
// DESCRIPTION:   
//
//
//

CiaoGuiIPCMessage* CiaoGuiIPCServer::DoCallback(CiaoGuiIPCMessage &pmess)
{
   CiaoGuiIPCCallbackTablePtr   pcbt;
   string         key;
   CiaoGuiIPCMessage      *pretmes=NULL;

   // Now determine which callback routine to call
   pcbt = cbTable;
   pmess.GetParam(0, key);

   while(pcbt)
   {
      // If this is a 'global' message callback, then just call
      // the routine and get it over with.

      if(strlen(pcbt->key) == 0)
      {
         pretmes = (*pcbt->cb)(pmess);
         break;
      }

      // It's not global, so do the comparison before calling...
      if(!strcmp(pcbt->key, key.c_str()))
      {
         pretmes = (*pcbt->cb)(pmess);
         break;
      }
      pcbt = pcbt->next;
   }
   return pretmes;
}

void CiaoGuiIPCServer::GtkLoopAdd()
{
   XPAGtkAddInput(NULL);
}


