// 
//  Copyright (C) 2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <iostream>
#include <vector>
#include <string>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>


using namespace std; 

// utility to tokenize a string into a vector of strings based on a
// string delimeter. 
vector<string> _ciao_string_split(string text, const string &delim=" ")
{
   vector<string> tokens;
   string::size_type end;
   string::size_type delim_len = delim.length();
   tokens.clear();
   while (text.length())
   {
      if ((end = text.find(delim)) == string::npos)  end = text.length() + 1;
      tokens.push_back(text.substr(0, end));
      text.replace(0, end + delim_len, "");
   }
   return tokens;
} 

// returns 0 if two char* arrays values are equal (ignoring case) 
int _ciao_strcmp_cis(const char *str1, const char *str2)
{
   int  ii, status = 1;
   if (str1 && str2)
   {
      if ((ii = strlen(str1)) == (int)strlen(str2))
      {
         status = 0;
         while (ii-- && !status)
         if (tolower(str1[ii]) != tolower(str2[ii]))   status = 1;
      }
   }
   return status;
}

// returns 0 if two strings are equal (ignoring case) 
int _ciao_strcmp_cis(const string str1, const string str2)
{
   int  status = 1;
   string::size_type ii;
   if ((ii = str1.length()) == (int)str2.length())
      status = _ciao_strcmp_cis(str1.c_str(), str2.c_str());
   return status;
}


// returns a string where all occurrance os oldstr are replaced by newstr
string _ciao_substring_replace(const string str, const string oldsub, const string newsub)
{
   string aa = str; 
   string::size_type pos=aa.find(oldsub);
   string::size_type olen = oldsub.length(); 
   while (pos != string::npos)
   {
      aa.replace(pos, olen, newsub);
      pos=aa.find(oldsub);
   }
   return aa; 
}

// returns string without  1-preceding, 2-trailing or 3-preceding/trial whitespace
string _ciao_strip_whitespace(const string str, int mode=3)
{
   string aa = str;
   string::size_type first = aa.find_first_not_of(" \n\t\r\v\f");
   string::size_type last = aa.find_last_not_of(" \n\t\r\v\f");
   if (first == string::npos ||((mode&0x01) != 0x01))  first = 0;
   if (last == string::npos || ((mode&0x02) != 0x02))  last = str.length();
   aa = str.substr(first, last-first+1);
   return aa;
}

// loose string compare- ignores case and preceeding/trailing whitespace
int _ciao_strcmp_loose(const string str1, const string str2)
{
   string s1 = _ciao_strip_whitespace(str1);
   string s2 = _ciao_strip_whitespace(str2);
   return _ciao_strcmp_cis(s1, s2);
}


bool _ciao_file_writable(string fname)
{
   struct stat   sbuf;
   mode_t        mode;

   if (!stat(fname.c_str(), &sbuf))
      mode = sbuf.st_mode;
   else
      return 0;

   //Get the REAL user information to determine r/w privileges
   uid_t uid = getuid();
   gid_t gid = getgid();

   // Get the EFFECTIVE user information to determine r/w privileges
   uid_t euid = geteuid();
   gid_t egid = getegid();

   return ((mode & S_IWOTH) ||
       ((mode & S_IWUSR) && (uid == sbuf.st_uid || euid == sbuf.st_uid)) ||
       ((mode & S_IWGRP) && (gid == sbuf.st_gid || egid == sbuf.st_gid)));
}


// returns 0 if successful 
int _ciao_file_copy(string infile, string outfile, mode_t mode)
{
   int     rval = 0;
   struct stat buf, buf2;
 
   // blank file specified 
   if (!(infile.length() && outfile.length()))
      return -1;    // blank file name specified 

   if ((stat(infile.c_str(),&buf) == 0) && (stat(outfile.c_str(),&buf2) ==0))
      if (buf.st_ino == buf2.st_ino)   return -2;   // same file 

   if (!access(outfile.c_str(),F_OK) && !access(outfile.c_str(),W_OK))
      return -3;    // file alread exists but is read only 
   
    string cmd = "cp ";
    cmd.append(infile);
    cmd.append(" ");
    cmd.append(outfile);
    rval = system(cmd.c_str());
   if (rval == 0)   chmod(outfile.c_str(), mode);  // set file permissions
   return rval;
}

bool _ciao_file_executable(string fname)
{
   struct stat   sbuf;
   mode_t        mode;

   if (!stat(fname.c_str(), &sbuf))
      mode = sbuf.st_mode;
   else
      return 0;

   //Get the REAL user information to determine r/w privileges
   uid_t uid = getuid();
   gid_t gid = getgid();

   // Get the EFFECTIVE user information to determine r/w privileges
   uid_t euid = geteuid();
   gid_t egid = getegid();

   return ((mode & S_IXOTH) ||
       ((mode & S_IXUSR) && (uid == sbuf.st_uid || euid == sbuf.st_uid)) ||
       ((mode & S_IXGRP) && (gid == sbuf.st_gid || egid == sbuf.st_gid)));
}


// expand out env vars w/in supplied path. Blindy substitute so don't worry about
// unknown env vars or multiple entries with env var (ie PATH may have ".:/bin")  
string _ciao_path_expand(string str)
{
   string env, sub, path;
   char* val; 
   string::size_type start = 0, stop;

   path = str; 
   while ((start = path.find('$', start)) != string::npos)
   {
      if ((path[start+1]=='{') && ((stop=path.find('}', start))!= string::npos))
      {
         sub = path.substr(start+2, stop-2-start);
         if ((val = getenv(sub.c_str())) != NULL)  env = val; 
            else env = "";
         path.replace(start, stop-start+1, env);
      }
      else
      {
         if ((stop = path.find('/', start+1)) != string::npos)
	    sub = path.substr(start+1, stop-1-start);
         else
	 {
	    stop = path.length();
	    sub = path.substr(start+1, stop-start);
	 }
         if ((val = getenv(sub.c_str())) != NULL) env = val;  
            else env = "";
         path.replace(start, stop-start, env); 
      }
      start++;
   }
   return path;
}

// return true if path is a directory, false otherwise 
bool _ciao_is_directory(string path)
{
   struct stat   sbuf;
   mode_t        mode;

   if (!stat(path.c_str(), &sbuf))
      mode = sbuf.st_mode;
   else
      return 0;
   return (S_ISDIR(mode) && !(S_ISREG(mode) || S_ISCHR(mode) || 
           S_ISBLK(mode) || S_ISFIFO(mode))); 
}


