// 
//  Copyright (C) 2003,2007,2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


//      This file contains source code for the CiaoToolList class. This
//      class is utilized by the gui infrastructure to maintain information
//      stored in the ciao.ans file to populate the Analysis Pulldown menu.
//      Its main purpose is to parse a file into a list of ciao analysis 
//      menu components (CiaoToolEntry class). 

#include <fstream>
#include <iostream>
#include <list>
#include <string>
      
#include "CiaoToolList.hh"
#include "CiaoGuiUtils.hh"

#define CTL_MAX_BUF_SZ   1024

// constants
const int CiaoToolList::CTL_NO_ERR         = 0x000;
const int CiaoToolList::CTL_NO_FILE_ERR    = 0x001;
const int CiaoToolList::CTL_OPEN_ERR       = 0x002;
const int CiaoToolList::CTL_BAD_CAT_ERR    = 0x004;
const int CiaoToolList::CTL_CAT_ALLOC_ERR  = 0x008;
const int CiaoToolList::CTL_CMD_ALLOC_ERR  = 0x010;
const int CiaoToolList::CTL_HMENU_ERR      = 0x020;
const int CiaoToolList::CTL_HELP_ALLOC_ERR = 0x040;
const int CiaoToolList::CTL_SEP_ALLOC_ERR  = 0x080;
const int CiaoToolList::CTL_EMPTY_BUF_ERR  = 0x100;
const int CiaoToolList::CTL_BAD_TYPE_ERR   = 0x200;
const int CiaoToolList::CTL_PARSE_SEQ_ERR  = 0x300;

// Constructor 
CiaoToolList::CiaoToolList(string name)
{
   _fname = name;
   _depth = 0;
   _state = INITIAL; 
   _hmenuCnt = 0; 
   _parsed_OK = parseList(); 
}

// Destructor 
CiaoToolList::~CiaoToolList()
{
   list<CiaoToolEntry*>::iterator it;
   for (it=_list.begin(); it != _list.end(); it++)
      delete (CiaoToolEntry*)(*it);
   _list.clear();
}

int CiaoToolList::parseCategory(char *buf)
{
   int rval = CTL_NO_ERR; 
   if (buf)
   {
      char empty[1] = {'\0'}; 
      char *tip, *ent = buf; 
      while (isspace(*ent)) ent++;   // remove leading whitespace

      if (ent)
      {
         if ((tip = strchr(ent, '#')) != NULL)
         {
            *tip = '\0';
            tip++;
            while (isspace(*tip)) tip++;
            if (((tip[0] == 't') || (tip[0] == 'T')) &&
                ((tip[1] == 'i') || (tip[1] == 'I')) &&
                ((tip[2] == 'p') || (tip[2] == 'P')))
            {
               tip = &tip[3];
               while (isspace(*tip)) tip++;
            }
            else   
               tip = empty;
         }
         if (!tip)  tip = empty; 
         CiaoToolEntry* te = new CiaoToolEntry(ent, tip, _depth++);   
         if (te)   
            _list.push_back(te);
         else   
            rval = CTL_CAT_ALLOC_ERR; 
      }
      else   
         rval = CTL_BAD_CAT_ERR; 
   }
   else   
      rval = CTL_EMPTY_BUF_ERR;
   return (rval); 
}

int CiaoToolList::parseEntry()
{
   int rval = CTL_NO_ERR;

   if (_label.length())
   {
      char name[256];
      char empty[1] = {'\0'}; 
      char *tip; 
  
      strcpy(name, _label.c_str());
      if ((tip = strchr(name, '#')) != NULL)
      {
         *tip = '\0'; 
         tip++;
         while (isspace(*tip)) tip++; 
         if (((tip[0] == 't') || (tip[0] == 'T')) && 
             ((tip[1] == 'i') || (tip[1] == 'I')) && 
             ((tip[2] == 'p') || (tip[2] == 'P')))  
         {
            tip = &tip[3]; 
            while (isspace(*tip)) tip++; 
         } 
         else   
            tip = empty;
      }
      if (!tip)  tip = empty;
      CiaoToolEntry* te = new CiaoToolEntry(name, tip, _templates, _type, 
         _command, _depth);
      if (te)   
         _list.push_back(te);
      else    
         rval = CTL_CMD_ALLOC_ERR; 
      _label.clear();  
   }
   return (rval);
}


int CiaoToolList::parseHelp()
{
   int rval = CTL_NO_ERR;

   CiaoToolEntry* te = new CiaoToolEntry(_label, _help, 1, _depth);
   if (te)   
      _list.push_back(te);
   else   
      rval = CTL_HELP_ALLOC_ERR; 
   return (rval); 
}


int CiaoToolList::addSeparator()
{
   int rval = CTL_NO_ERR;

   CiaoToolEntry* te = new CiaoToolEntry(_depth);
   if (te)   
      _list.push_back(te);
   else   
      rval = CTL_SEP_ALLOC_ERR;

   return (rval);
}


// destructively modifies buffer
void CiaoToolList::removeComment(char *buf)
{
   if (buf)
   {
      char *val = buf;
      if (*buf == '#')
      {
         // first char is character marker
         *val = '\0'; 
      }
      else if ((*buf != '\0') && (*buf != '\n')) 
      {
         val++;  
         while ((*val != '\0' && *val != '\n') &&
              ((*val != '#') || ((*val == '#') && (*(val-1) == '\\')))) val++;
         if ((*val == '#') && (*(val-1) != '\\')) *val = '\0';  
      } 
   }
}

int CiaoToolList::parseLine(char *buf)
{
   int rval = CTL_NO_ERR;
   static int line = 1; 

   if (buf)
   {
      char  tmp[CTL_MAX_BUF_SZ];
      char  *word, *ps = tmp;

      strcpy(tmp, buf);
      while (isspace(*ps)) ps++;
      word = ps;

      if ((word[0] != '\0') && (word[0] != '\n') && (word[0] != '#'))
      {
         if (_state == COMMAND)
         {
            if ((strncmp(word,"# tip",5)==0) || (strncmp(word,"# TIP",5)==0))
            {
               _state = TIP; 
               _misc = word; 
            }
            else   
               _state = INITIAL;
            rval |= parseEntry(); 
         }
         if (_state == INITIAL)
         {
            if (!strncmp(word,"help",4))
            {
               // start help description
               _help.clear();
               _label = &word[4];
	       _label = _ciao_strip_whitespace(_label, 1); 
               _state = HELP; 
            }
            else if (!strncmp(word,"hmenu",5))
            {
               // start category description 
               rval = parseCategory(&word[6]); 
               _hmenuCnt++; 
            }
            else if (!strncmp(word,"endhmenu",7))
            {
               // finished category - adjust depth 
               _depth--; 
               _hmenuCnt--; 
            }
            else if (!strncmp(word,"param ",6))
            {
               _state = PARAM; 
            }
            else if (!strncmp(word,"---",3))
            {
               // insert separator 
               rval = addSeparator(); 
            }
            else 
            {
               // parse tool entry 
               _state = TOOLNAME; 
               _label = word; 
            } 
         }
         else if (_state == PARAM)   // in param declaration 
         { 
            if ((strncmp(word,"end",3)==0) || (strncmp(word,"endparam",8)==0))
            {
               // finished param declaration
               _state = INITIAL;
            }
         }
         else if (_state == TOOLNAME)
         {
            // parse second line entry 
            _templates = word; 
            _state = TEMPLATES; 
         }
         else if (_state == TEMPLATES)
         {
            // parse third line entry 
            if (strstr(word, "menu") || strstr(word, "bind"))
            {
               _type = word; 
               _state = CMDTYPE; 
            }
            else rval |= CTL_BAD_TYPE_ERR;
         }
         else if (_state == CMDTYPE)
         {
            // parse fourth line 
            _command = word; 
            _state = COMMAND; 
         }
         else if (_state == HELP)
         {
            // finished help
            if (strncmp(word,"endhelp",7))
            {
               _help.append(word);
               _help.append("\n");
            }
            else
            {
               rval = parseHelp(); 
               _label.clear(); 
               _state = INITIAL;
            }
         }
    
         if (_state == TIP)   _state = INITIAL; 
      }
      line++; 
   }
   if (_hmenuCnt < 0)   rval = CTL_HMENU_ERR; 
   return (rval);
}

int CiaoToolList::parseList()
{
   int rval = CTL_NO_ERR; 

   if (_fname.length())
   {
      char   buf[CTL_MAX_BUF_SZ];
      ifstream *infile =  new ifstream(_fname.c_str(), ios::in);
      if (infile->good())    
      {
         while (infile->getline(buf, CTL_MAX_BUF_SZ))   
            rval |= parseLine(buf);
         if ((_state == COMMAND) && _label.length())  
            rval |= parseEntry(); 
         if (!(_state == COMMAND || _state == INITIAL || _state == TIP)) 
            rval |= CTL_PARSE_SEQ_ERR;
         if (_hmenuCnt)  
            rval |= CTL_HMENU_ERR;  
      }
      else
         rval = CTL_OPEN_ERR; 
      delete infile; 
   }
   else 
      rval = CTL_NO_FILE_ERR; 

   return rval; 
}


// enable/disable entries based on file format
void CiaoToolList::SupportByFormat(char* type)
{
   if ((_parsed_OK == 0) && type)
   {
      CiaoToolEntry *te_p;
      list<CiaoToolEntry*>::iterator it;

      for (it = _list.begin(); it != _list.end(); it++)
      {
         if ((te_p = (CiaoToolEntry*)(*it)) != NULL)
            te_p->SupportByFormat(type);
      }
   }
}

